(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Equal.
Require Export Marks.

Set Implicit Arguments.
Unset Strict Implicit.

Section IInvReduction.

  Context {sg : sigma}.

  Section GenericInversion.

    Variable R : red_rule.
    Hypothesis R_refl : forall (e : env) (t : term), R e t t.
    Hypothesis R_trans :
      forall (e : env) (x y z : term), R e x y -> R e y z -> R e x z.
    Hypothesis R_stb_var :
      forall (e : env) (x y : name) (t t' a b : term),
        R e t t' -> R (e; x : t') a b -> R (e; y : t) a b.

    Inductive hn_inv_pred : env -> term -> term -> Prop :=
      | redhi_srt : forall (e : env) (s : sort), hn_inv_pred e %%s %%s
      | redhi_rel : forall (e : env) (n : nat), hn_inv_pred e #n #n
      | redhi_cst0 :
        forall (e : env) (c : cci_op), hn_inv_pred e (Cst0 c) (Cst0 c)
      | redhi_cst1 :
        forall (e : env) (c : cci_op) (x y : term),
          R e x y -> hn_inv_pred e (Cst1 c x) (Cst1 c y)
      | redhi_prd :
        forall (e : env) (x : name) (A A' B B' : term),
          R e A A' ->
          R (e; x : A) B B' ->
          hn_inv_pred e (Forall x : A, B) (Forall x : A', B')
      | redhi_lam :
        forall (e : env) (x : name) (A A' M M' : term),
          R e A A' ->
          R (e; x : A) M M' ->
          hn_inv_pred e (Fun x : A => M) (Fun x : A' => M')
      | redhi_sum :
        forall (e : env) (A A' B B' : term),
          R e A A' -> R e B B' -> hn_inv_pred e (A ** B) (A' ** B')
      | redhi_pair :
        forall (e : env) (M M' N N' : term),
          R e M M' -> R e N N' -> hn_inv_pred e (M,,N) (M',,N')
    .

    Lemma hn_inv_refl : forall (e : env) (x : term), hn_inv_pred e x x.
    Proof.
      simple destruct x; constructor; auto.
    Qed.

    Lemma hn_inv_trans :
      forall (e : env) (x y z : term),
        hn_inv_pred e x y -> hn_inv_pred e y z -> hn_inv_pred e x z.
    Proof.
      simple destruct 1; intros.
      inversion_clear H0; constructor.

      inversion_clear H0; constructor.

      inversion_clear H0; constructor.

      inversion_clear H1; constructor.
      apply R_trans with y0; trivial.

      inversion_clear H2; constructor.
      apply R_trans with A'; trivial.

      apply R_trans with B'; trivial.
      apply R_stb_var with x0 A'; trivial.

      inversion_clear H2; constructor.
      apply R_trans with A'; trivial.

      apply R_trans with M'; trivial.
      apply R_stb_var with x0 A'; trivial.

      inversion_clear H2; constructor.
      apply R_trans with A'; trivial.

      apply R_trans with B'; trivial.

      inversion_clear H2; constructor.
      apply R_trans with M'; trivial.

      apply R_trans with N'; trivial.
    Qed.

  End GenericInversion.

  (* Inversion of redn_term *)

  Lemma stb_var : forall (e : env) (x y : name) (t t' a b : term),
    red redn_term e t t' -> red redn_term (e; x : t') a b ->
      red redn_term (e; y : t) a b.
  Proof.
    intros.
    red in |- *.
    apply
      (stable_rt _ _ (ctxt_stable _ _ (R_stable_redn (R:=red redn_term))))
      with (e; x : t'); auto.
    unfold add_var in |- *; constructor.
    split; simpl in |- *; auto.
  Qed.

  Definition redn_term_hn_inv := hn_inv_pred (red redn_term).

  Lemma inv_redn_trans :
    forall (e : env) (x y z : term),
      redn_term_hn_inv e x y -> redn_term_hn_inv e y z -> redn_term_hn_inv e x z.
  Proof.
    red in |- *; intros.
    apply hn_inv_trans with y; auto.
    exact (red_trans redn_term).

    exact stb_var.
  Qed.


  Lemma inv_redn_red :
    forall (e : env) (x y : term),
      redn_term_hn_inv e x y -> red redn_term e x y.
  Proof.
    simple destruct 1; intros; auto.
    apply red_trans with (Fun x0 : A => M'); auto.
  Qed.

  Hint Resolve inv_redn_red: core.

  (* One step head reduction *)
  Inductive head_redn_term (e : env) (x t : term) : Prop :=
    Red_hd_intro : forall y z : term,
      redn_term_hn_inv e x y ->
      redn_term e y z -> red redn_term e z t -> head_redn_term e x t.

  Lemma head_redn_not_hn :
    forall (e : env) (M N : term),
      head_redn_term e M N -> ~ head_normal redn_term e M.
  Proof.
    simple destruct 1; intros.
    unfold head_normal, normal, not in |- *; intros.
    apply H3 with y z; auto.
  Qed.

  Lemma redn_head :
    forall (e : env) (x y : term), redn_term e x y -> head_redn_term e x y.
  Proof.
    intros.
    split with x y; auto.
    red in |- *.
    apply hn_inv_refl; auto.
  Qed.

  Lemma inv_redn_hn0 :
    forall (e : env) (x y : term),
      red redn_term e x y -> head_redn_term e x y \/ redn_term_hn_inv e x y.
  Proof.
    simple induction 1.
    simple destruct 1; intros; try (right; constructor; auto; fail).
    left.
    apply redn_head.
    trivial.

    right.
    red in |- *.
    apply hn_inv_refl; auto.

    intros.
    inversion_clear H1.
    left.
    inversion_clear H4.
    split with y1 z0; auto.
    apply red_trans with y0; auto.

    inversion_clear H3.
    left.
    inversion_clear H1.
    split with y1 z0; auto.
    apply inv_redn_trans with y0; trivial.

    right.
    apply inv_redn_trans with y0; trivial.
  Qed.

  Lemma inv_redn_hn :
    forall (e : env) (x y : term),
      red redn_term e x y -> head_normal redn_term e x -> redn_term_hn_inv e x y.
  Proof.
    intros.
    elim inv_redn_hn0 with (1 := H); intros; trivial.
    elim head_redn_not_hn with (1 := H1); trivial.
  Qed.

  Definition hn_pair e t P :=
    head_normal redn_term e t /\
    match t with
      | (u,,v) => P u v
      | _ => True
    end.

  Definition hn_constr e t P :=
    head_normal redn_term e t /\
    match t with
    | Cstr{c // a} => P c a
    | _ => True
    end.

  Fixpoint hn_proj (e : env) (t : term) (P : term -> Prop) {struct t} : Prop :=
    head_normal redn_term e t /\
    match t with
      | u..R => hn_proj e u P
      | _ => P t
    end.

  Definition hn_mutfix e t :=
    hn_proj e t
      (fun mf =>
        head_normal redn_term e mf /\
        match mf with
          | Fix {uf} => hn_pair e uf (fun _ _ => False)
          | _ => True
        end).

  Definition hn_fix e f :=
    head_normal redn_term e f /\
    match f with
      | tlf..L => hn_mutfix e tlf
      | _ => True
    end.

  Definition hn_subterm e t :=
    match t with
      | #n => forall u, ~ delta e t u
      | Cst1 (Const _) _ => forall u, ~ delta_glob sg e t u
      | u..L => hn_pair e u (fun _ _ => False)
      | u..R => hn_pair e u (fun _ _ => False)
      | Cst1 (MutCase b pat) u =>
        hn_pair e u
        (fun c u' =>
          hn_pair e u'
          (fun p br =>
            hn_constr e c
            (fun cn cu =>
              hn_pair e cu (fun _ _ => forall n, ~ match_pattern pat cn n))))
      | u..( f) =>
        head_normal redn_term e u /\
        match u with
          | Cst1 (Struct l) m => hn_pair e m (fun _ _ => ~ In f l)
          | _ => True
        end
      | Cst1 App u =>
        hn_pair e u
        (fun f arg =>
          head_normal redn_term e f /\
          match f with
            | Fun _ : _ => _ => False
            | Cst1 App u' =>
              hn_pair e u'
              (fun f' _ => hn_constr e arg (fun _ _ => hn_fix e f'))
            | _ => True
          end)
      | _ => True
    end.

  Hint Unfold hn_pair hn_constr hn_proj hn_mutfix hn_fix hn_subterm: core.

  Lemma hn_proj_remove :
    forall (P : term -> Prop) e n t, hn_proj e (t .^ n) P -> hn_proj e t P.
  Proof.
    simple induction n; simpl in |- *.
    trivial.

    intros n0 Hr t.
    replace (t..R .^ n0) with (t .^ n0) ..R.
    simpl in |- *.
    simple induction 1; auto.
    generalize t.
    elim n0; simpl in |- *; intros; auto.
  Qed.

  Lemma hn_proj_red1 :
    forall (P : term -> Prop) e a b,
      ctxt redn_term e a b ->
      (forall a b, ctxt redn_term e a b -> P a -> P b) ->
      hn_proj e a P -> hn_proj e b P.
  Proof.
    simple induction 1; simpl in |- *; intros;
      try (destruct H3; split; [ eapply hn_red_hn; eauto | eauto ]).
    assert (hn_t : head_normal redn_term e0 t).
    destruct t as [s| n| c| c t| n t t0| n t t0| t t0| t t0]; destruct H2;
      trivial.

    elim hn_t with t u; auto.

    destruct c; eauto.
  Qed.

  Lemma hn_subterm_normal : forall e t, hn_subterm e t -> normal redn_term e t.
  Proof.
    red in |- *; red in |- *; intros e t hnty u H.
    generalize hnty; clear hnty.
    repeat
      (inversion_clear H;
        [ inversion_clear H0 in H0; simpl in |- *; unfold not in |- *; intro hnty;
          unfold hn_pair, hn_fix, hn_constr in hnty; decompose [and] hnty;
            eauto
        | rename H0 into H
        ]).
    apply H4; elim H; auto.

    assert (hn_fixp := hn_proj_remove H7).
    clear H7.
    destruct H; unfold hn_proj, hn_pair in hn_fixp.
    tauto.

    contradiction.
  Qed.

  Lemma hn_subterm_red : forall (e : env) (a b : term),
    red redn_term e a b -> hn_subterm e a -> hn_subterm e b.
  Proof.
    simple induction 1; auto.
    simple destruct 1; simpl in |- *; intros; trivial.
    elim hn_subterm_normal with (2 := H1); trivial.

    destruct c; trivial.
    (* apply *)
    destruct H2; split.
    eapply hn_red_hn; eauto.

    destruct H1; simpl in |- *; auto.
    elim H2 with t u; auto.

    destruct H3; split.
    eapply hn_red_hn; eauto.

    destruct H1; simpl in |- *; auto.
    elim H3 with t u; auto.

    destruct c; auto.
    destruct H4; split.
    eapply hn_red_hn; eauto.

    destruct H1; auto.
    elim H4 with t u; auto.

    destruct H5; split.
    eapply hn_red_hn; eauto.

    destruct n as [s| n| c| c t| n t t0| n t t0| t t0| t t0]; simpl in |- *; auto.
    destruct c; auto.
    destruct H6; split.
    eapply hn_red_hn; eauto.

    destruct H1; auto.
    elim H6 with t0 u; auto.

    destruct c; auto.
    apply hn_proj_red1 with (1 := H1) (3 := H7); clear H1 H7; intros.
    destruct H7; split.
    eapply hn_red_hn; eauto.

    destruct H1; auto.
    elim H7 with t1 u; auto.

    destruct c; auto.
    destruct H8; split.
    eapply hn_red_hn; eauto.

    destruct H1; auto.
    elim H8 with t1 u; auto.

    destruct H3; split.
    eapply hn_red_hn; eauto.

    destruct m as [s| n0| c| c t| n0 t t0| n0 t t0| t t0| t t0]; auto.
    destruct c; auto.
    destruct H4; split.
    eapply hn_red_hn; eauto.

    destruct t as [s| n0| c| c t| n0 t t0| n0 t t0| t t0| t t0]; auto.
    destruct H5; split.
    eapply hn_red_hn; eauto.

    destruct H1; auto.
    elim H5 with t1 u; auto.

    (* proj1 *)
    destruct H2; split.
    eapply hn_red_hn; eauto.

    destruct H1; simpl in |- *; auto.
    elim H2 with t u; auto.

    (* proj2 *)
    destruct H2; split.
    eapply hn_red_hn; eauto.

    destruct H1; simpl in |- *; auto.
    elim H2 with t u; auto.

    (* projr *)
    destruct H2; split.
    eapply hn_red_hn; eauto.

    destruct H1; simpl in |- *; auto.
    elim H2 with t u; auto.

    destruct c; auto.
    destruct H3; split.
    eapply hn_red_hn; eauto.

    destruct H1; auto.
    elim H3 with t u; auto.

    (* delta-g *)
    red in |- *; intros u red_cst.
    inversion_clear red_cst.
    apply H2 with (d[0:=t]).
    eapply deltag_intro; eauto.

    (* Case *)
    destruct H2; split.
    eapply hn_red_hn; eauto.

    destruct H1; simpl in |- *; auto.
    elim H2 with t u; auto.

    destruct H3; split; trivial.
    destruct n as [s| n| c| c t| n t t0| n t t0| t t0| t t0]; trivial.
    destruct H4; split.
    eapply hn_red_hn; eauto.

    destruct H1; simpl in |- *; auto.
    elim H4 with t1 u; auto.

    destruct c; auto.
    destruct H5; split.
    eapply hn_red_hn; eauto.

    destruct H1; auto.
    elim H5 with t1 u; auto.

    (* Case' *)
    destruct H3; split.
    eapply hn_red_hn; eauto.

    destruct H1; auto.
    elim H3 with t u; auto.
  Qed.

  Lemma hn_subterm_hn :
    forall (e : env) (t : term), hn_subterm e t -> head_normal redn_term e t.
  Proof.
    red in |- *; intros.
    apply hn_subterm_normal.
    apply hn_subterm_red with t; trivial.
  Qed.

  Hint Resolve hn_subterm_hn: core.

  Lemma mt_mark_normal : forall e : env, normal (ctxt redn_term) e ##.
  Proof.
    red in |- *; red in |- *; intros.
    inversion_clear H.
    elim hn_subterm_normal with e ## u; simpl in |- *; trivial.
  Qed.

  Lemma nil_mark_normal : forall e : env, normal (ctxt redn_term) e [].
  Proof.
    red in |- *; red in |- *; intros.
    inversion_clear H.
    elim hn_subterm_normal with e (Nil_mark:term) u; simpl in |- *; trivial.
  Qed.

  Lemma one_normal : forall e : env, normal (ctxt redn_term) e one.
  Proof.
    red in |- *; red in |- *; intros.
    inversion_clear H.
    elim hn_subterm_normal with e one u; simpl in |- *; trivial.
  Qed.

  Lemma red_pair :
    forall (P : term -> Prop) (e : env) (a b : term),
      (forall a' b' : term,
        red redn_term e a a' -> red redn_term e b b' -> P (a',,b')) ->
      forall m : term, red redn_term e (a,,b) m -> P m.
  Proof.
    intros.
    cut (redn_term_hn_inv e (a,,b) m); intros.
    inversion_clear H1; auto.

    apply inv_redn_hn; auto.
  Qed.

  Lemma inv_red_prd_r :
    forall (e : env) (x1 x2 : name) (a1 a2 b1 b2 : term),
      red redn_term e (Forall x1 : a1, b1) (Forall x2 : a2, b2) ->
        red redn_term (e; x1 : a1) b1 b2.
  Proof.
    intros.
    cut (redn_term_hn_inv e (Forall x1 : a1, b1) (Forall x2 : a2, b2)); intros;
      auto.
    inversion_clear H0; trivial.

    apply inv_redn_hn; auto.
  Qed.

  Axiom confluent_redn : confluent (red redn_term).

  Axiom commut_alpha_red1 :
    forall e : env, commut _ (redn_term e) (fun y x => red alpha e x y).

  Axiom commut_alpha_redn :
    forall e : env, commut _ (red redn_term e) (fun y x => red alpha e x y).

  Lemma inv_alpha_cons_sym : forall (e : env) (a b : term),
    red alpha e a b -> forall m ml : term,
      a = m && ml ->
      exists2 m' : _,
      red alpha e m m' & (exists2 ml' : _, red alpha e ml ml' & b = m' && ml').
  Proof.
    simple induction 1; intros.
    rewrite H1 in H0.
    clear H1.
    inversion_clear H0.
    inversion_clear H1.

    inversion_clear H1.
    inversion_clear H0.

    split with m'; auto.
    split with ml; auto.

    split with m; auto.
    split with n'; auto.

    split with m; auto.
    split with ml; auto.

    elim H1 with (1 := H4); intros.
    inversion_clear H6.
    elim H3 with (1 := H8); intros.
    inversion_clear H9.
    split with x2.
    apply red_trans with x0; trivial.

    split with x3; trivial.
    apply red_trans with x1; trivial.
  Qed.

  Lemma commut_alpha_incl :
    forall (e : env) (k k' : le_marks),
      commut term (fun x y => incl_seq k x k' y) (fun x y => red alpha e y x).
  Proof.
    red in |- *.
    simple induction 1; intros.
    split with z; auto.
    apply Incl_refl.

    elim H1 with (1 := H4); intros.
    elim H3 with (1 := H5); intros.
    split with x1; trivial.
    apply Incl_trans with m2 x0; trivial.

    split with []; auto.
    apply Incl_mt.

    split with z; trivial.
    apply Incl_str.

    elim inv_alpha_cons_sym with (1 := H0) (m := m) (ml := ml); intros; trivial.
    inversion_clear H2.
    rewrite H4.
    split with x1; trivial.
    apply Incl_remove.

    split with (## && z); auto.
    apply Incl_add_mt.

    elim inv_alpha_cons_sym with (1 := H2) (m := m) (ml := ml); intros; trivial.
    inversion_clear H4.
    rewrite H6.
    elim H1 with (1 := H5); intros.
    split with (x0 && x2); auto.
    apply Incl_deep_le; trivial.

    elim inv_alpha_cons_sym with (1 := H2) (m := m) (ml := ml); intros; trivial.
    inversion_clear H4.
    rewrite H6.
    elim H1 with (1 := H5); intros.
    split with (x0 && x2); auto.
    apply Incl_deep_lt; trivial.
  Qed.

  (* Used in InvSubtyping *)
  Lemma hn_alpha : forall (e : env) (t u : term),
    red alpha e t u -> head_normal redn_term e t -> head_normal redn_term e u.
  Proof.
    simple induction 1; intros; auto.
    generalize H1.
    elim H0; intros; auto.
    generalize H3.
    inversion_clear H2; auto.
    red in |- *; intros.
    elim commut_alpha_redn with (1 := H5) (z := Cst1 c t0); intros.
    red in H4.
    red in |- *; red in |- *; intros.
    elim commut_alpha_red1 with (1 := H8) (2 := H6); intros.
    elim H4 with (1 := H7) (2 := H10).

    apply alpha_sym; auto.
  Qed.

End IInvReduction.

Hint Resolve @inv_redn_red: core.
Hint Unfold hn_pair hn_constr hn_proj hn_mutfix hn_fix hn_subterm: core.
Hint Resolve @hn_subterm_hn: core.
