(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Signature.

Set Implicit Arguments.
Unset Strict Implicit.

(* N-th projection *)
Fixpoint proj (t : term) (n : nat) {struct n} : term :=
  match n with
    | O => t
    | S k => proj t..R k
  end.

Notation "t .^ n" := (proj t n) (at level 30).

Lemma proj_comm : forall n m, m..R .^ n = (m .^ n) ..R.
Proof.
  simple induction n; simpl in |- *; intros; auto.
Qed.
Hint Rewrite proj_comm : core.

Lemma proj_injective : forall n t1 t2, t1 .^ n = t2 .^ n -> t1 = t2.
Proof.
  induction n; simpl; intros; auto.
  do 2 rewrite proj_comm in H.
  inversion H.
  apply IHn; auto.
Qed.

Lemma lift_proj : forall n k i t, ^(n//k)(t .^ i) = ^(n//k)t .^ i.
Proof.
  simple induction i; simpl in |- *; intros; auto.
  rewrite H; trivial.
Qed.

Lemma subst_proj : forall k i u t, (t .^ i)[k:=u] = t[k:=u] .^ i.
Proof.
  simple induction i; simpl in |- *; intros; auto.
  rewrite H; trivial.
Qed.

(* Inclure alpha et delta dans le PTS d'office ?
   (seules re'ductions n'affectant pas les constantes) *)

Section AlphaConversion.

  (* Alpha-conversion: renaming of bound variables *)

  Inductive alpha (e : env) : term -> term -> Prop :=
    | Alpha_lam : forall x y a m, alpha e (Fun x : a => m) (Fun y : a => m)
    | Alpha_prd : forall x y a m, alpha e (Forall x : a, m) (Forall y : a, m)
  .

  Lemma R_lift_alpha : R_lift alpha.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *; constructor.
  Qed.

  Lemma R_subst_alpha : R_subst alpha.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *; constructor.
  Qed.

  Lemma stable_alpha : forall R, R_stable R alpha.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *; constructor.
  Qed.

End AlphaConversion.

Section DeltaReduction.

  (* Delta-reduction: expanding a De Bruijn *)

  Inductive delta (e : env) : term -> term -> Prop :=
    delta_intro : forall n x t m,
      item_rel e n (mkdecl x t (Some m)) -> delta e #n ^(S n)m.

  Lemma R_lift_delta : R_lift delta.
  Proof.
    red in |- *.
    simple destruct 1; intros.
    apply lift_split; simpl in |- *; intros.
    replace k with (S n + (k - S n)); auto with arith.
    rewrite <- permute_lift.
    apply delta_intro with x (^(1//k - S n)t0).
    change (item_rel f n (liftd_rec 1 (mkdecl x t0 (Some m)) (k - S n))) in |- *.
    apply (ins_item_lt _ _ _ _ _ H1); trivial.

    unfold lift in |- *.
    rewrite simpl_lift_rec; simpl in |- *; auto with arith.
    fold (^(S (S n))m) in |- *.
    apply delta_intro with x t0; auto.
    apply ins_item_ge with (1 := H1); trivial.
  Qed.

  Lemma R_subst_delta : R_subst_weak delta.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    generalize H1.
    clear H1.
    pattern n0, (#n0[n:=m]) in |- *.
    apply subst_split; intros.
    replace n with (S n0 + (n - S n0)); auto with arith.
    rewrite commut_lift_subst.
    apply rt_step.
    apply delta_intro with x (t0[n - S n0:=m]).
    change (item_rel f n0 (substd_rec m (mkdecl x t0 (Some m0)) (n - S n0)))
      in |- *.
    apply nth_sub_inf with (1 := H0); trivial.

    rewrite simpl_subst; auto with arith.
    generalize (sub_decl_eq_ok _ _ _ _ _ _ H0).
    elim (inj_item_rel e n (mkdecl x t0 (Some m0)) d);
      trivial.
    unfold val_ok in |- *.
    simpl in |- *.
    simple induction 1.
    auto.

    apply (sub_decl_eq_def _ _ _ _ _ _ H0).

    rewrite simpl_subst; auto with arith.
    apply rt_step.
    apply delta_intro with x t0.
    apply (nth_sub_sup _ _ _ _ _ _ H0); trivial.
  Qed.

  Lemma stb_decl_rel :
    forall R, R_stable (decl_rel (fun e t1 t2 => t1 = t2) R) delta.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    elim rel_list_item with (1 := H1) (2 := H0); intros.
    split with x t0; trivial.

    elim item_trunc with (1 := H1); intros.
    elim H with (1 := H2).
    intros (y, (rel_ty, rel_body), itm) tr.
    simpl in rel_body.
    split with (NAME y) (TYPE y).
    generalize itm rel_body.
    case y.
    intros yn yt yb.
    case yb; simpl in |- *; intros.
    elim rel_body0; trivial.

    elim rel_body0.
  Qed.


  Lemma stable_delta : forall R, R_stable (rel_typ R) delta.
  Proof.
    intros.
    apply stable_incl with (decl_rel (fun e t1 t2 => t1 = t2) R).
    apply stb_decl_rel.

    red in |- *; simple destruct 1; constructor; simpl in |- *; trivial;
      rewrite H1; case (BODY y); trivial.
  Qed.

End DeltaReduction.

Section DeltaGlobal.

  Variable sg : sigma.

  (* Delta-expansion of a global constant *)

  Inductive delta_glob (e : env) : term -> term -> Prop :=
    deltag_intro : forall d m x c,
      sign_glob_spec sg x (GlobConst c) ->
      BODY (c_decl c) = Some d -> delta_glob e (Cst1 (Const x) m) d[0:=m].

  Lemma R_lift_delta_glob : R_lift delta_glob.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    simpl in |- *.
    rewrite distr_lift_subst.
    split with c; trivial.
    rewrite closed_lift_rec; trivial.
    apply closed_weak with 1; auto with arith.
    apply closbd with (c_decl c); trivial.
  Qed.

  Lemma R_subst_delta_glob : R_subst delta_glob.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    simpl in |- *.
    rewrite distr_subst.
    split with c; trivial.
    rewrite closed_subst_rec; trivial.
    apply closed_weak with 1; auto with arith.
    apply closbd with (c_decl c); trivial.
  Qed.

  Lemma stable_delta_glob : forall R, R_stable R delta_glob.
  Proof.
    red in |- *.
    simple destruct 1; intros.
    split with c; trivial.
  Qed.

End DeltaGlobal.

Section BetaReduction.

  (* Beta-reduction *)

  Inductive beta (e : env) : term -> term -> Prop :=
    beta_intro : forall x n t m, beta e ((Fun x : t => m) @ n) m[0:=n].

  Lemma R_lift_beta : R_lift beta.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    simpl in |- *.
    rewrite distr_lift_subst.
    constructor.
  Qed.

  Lemma R_subst_beta : R_subst beta.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    simpl in |- *.
    rewrite distr_subst.
    constructor.
  Qed.

  Lemma stable_beta : forall R, R_stable R beta.
  Proof.
    red in |- *.
    simple destruct 1; constructor.
  Qed.

End BetaReduction.

Section Projections.

  (* Projections of sigma-types *)

  Inductive proj1 (e : env) : term -> term -> Prop :=
    Proj1_intro : forall u v, proj1 e (u,,v) ..L u.

  Lemma R_lift_proj1 : R_lift proj1.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *; constructor.
  Qed.

  Lemma R_subst_proj1 : R_subst proj1.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *; constructor.
  Qed.

  Lemma stable_proj1 : forall R, R_stable R proj1.
  Proof.
    red in |- *; intros.
    inversion_clear H; constructor.
  Qed.

  Inductive proj2 (e : env) : term -> term -> Prop :=
    Proj2_intro : forall u v, proj2 e (u,,v) ..R v.

  Lemma R_lift_proj2 : R_lift proj2.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *; constructor.
  Qed.

  Lemma R_subst_proj2 : R_subst proj2.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *; constructor.
  Qed.

  Lemma stable_proj2 : forall R, R_stable R proj2.
  Proof.
    red in |- *; intros.
    inversion_clear H; constructor.
  Qed.

  Inductive projr (e : env) : term -> term -> Prop :=
    Projr_intro : forall x f t m n,
      first_item x f n -> projr e {<f : t := m>} ..( x) (m .^ n) ..L.

  Lemma R_lift_projr : R_lift projr.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *.
    rewrite lift_proj; simpl in |- *.
    constructor; trivial.
  Qed.

  Lemma R_subst_projr : R_subst projr.
  Proof.
    red in |- *; intros.
    inversion_clear H; simpl in |- *.
    rewrite subst_proj; simpl in |- *.
    constructor; trivial.
  Qed.

  Lemma stable_projr : forall R, R_stable R projr.
  Proof.
    red in |- *; intros.
    inversion_clear H; constructor; trivial.
  Qed.

End Projections.

Section IotaReduction.

  (* Fixpoint reduction guarded by a constructor *)

  Inductive fix_free : term -> term -> Prop :=
    fix_free_intro : forall ty fixs,
      fix_free (Fix {(ty,,fixs)}) ((fixs @ ##) @ (Fix {(ty,,fixs)})).


  Inductive iota_fix (e : env) : term -> term -> Prop :=
    iota_fix_intro : forall n c cargs arg fx efix,
      fix_free fx efix ->
      iota_fix e (((fx .^ n) ..L @ arg) @ Cstr{c // cargs})
      (((efix .^ n) ..L @ arg) @ Cstr{c // cargs}).

  Lemma R_lift_iota_fix : R_lift iota_fix.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    simpl in |- *.
    rewrite lift_proj.
    rewrite lift_proj.
    simpl in |- *.
    constructor.
    inversion_clear H1; simpl in |- *; constructor.
  Qed.

  Lemma R_subst_iota_fix : R_subst iota_fix.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    simpl in |- *.
    rewrite subst_proj.
    rewrite subst_proj.
    simpl in |- *.
    constructor.
    inversion_clear H1; simpl in |- *; constructor.
  Qed.

  Lemma stable_iota_fix : forall R, R_stable R iota_fix.
  Proof.
    red in |- *; intros.
    inversion_clear H; constructor; trivial.
  Qed.

  (* Case reduction: pattern-matching *)

  Definition match_pattern p c n := first_item c (p:patterns) n.

  Inductive iota_case (e : env) : term -> term -> Prop :=
    iota_case_intro : forall c pats n p a pred br b,
      match_pattern pats c n ->
      iota_case e
      Match Cstr{c // p,,a} => pred with b — pats => br end
      ((br .^ n) ..L @ a).

  Lemma R_lift_iota_case : R_lift iota_case.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    simpl in |- *.
    rewrite lift_proj.
    simpl in |- *.
    constructor; auto.
  Qed.

  Lemma R_subst_iota_case : R_subst iota_case.
  Proof.
    red in |- *; intros.
    inversion_clear H.
    simpl in |- *.
    rewrite subst_proj.
    simpl in |- *.
    constructor; auto.
  Qed.

  Lemma stable_iota_case : forall R, R_stable R iota_case.
  Proof.
    red in |- *; intros.
    inversion_clear H; constructor; auto.
  Qed.

End IotaReduction.

Hint Resolve Proj1_intro Proj2_intro Projr_intro beta_intro delta_intro
  deltag_intro iota_fix_intro fix_free_intro iota_case_intro: core.

Hint Resolve R_lift_alpha R_subst_alpha stable_alpha: core.
Hint Resolve R_lift_delta R_subst_delta stable_delta: core.
Hint Resolve R_lift_delta_glob R_subst_delta_glob stable_delta_glob: core.
Hint Resolve R_lift_beta R_subst_beta stable_beta: core.
Hint Resolve R_lift_iota_fix R_subst_iota_fix stable_iota_fix: core.
Hint Resolve R_lift_iota_case R_subst_iota_case stable_iota_case: core.
Hint Resolve R_lift_proj1 R_subst_proj1 stable_proj1: core.
Hint Resolve R_lift_proj2 R_subst_proj2 stable_proj2: core.
Hint Resolve R_lift_projr R_subst_projr stable_projr: core.
