(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export CicCore.

Set Implicit Arguments.
Unset Strict Implicit.

Section ListSearchMap.

  Variable A : Set.
  Variable B : Set.
  Variable C : A -> Prop.
  Variable l : list A.
  Variable f : forall x, In x l -> B + {C x}.

  Lemma search_map : B + {(forall x, In x l -> C x)}.
  Proof.
    generalize f.
    clear f.
    elim l; intros.
    right; intros.
    inversion H.

    elim f with a; intros; auto.
    elim H; intros; auto.
    right; intros.
    generalize b.
    inversion_clear H0; auto.
    subst; intro; assumption.
  Qed.

End ListSearchMap.

Unset Elimination Schemes.

Record constructor : Set := mk_cstr {
  c_name : name;
  c_arg : cterm 1;
  c_inst : cterm 2
}.

Record one_ind : Set := mk_one_ind {
  i_name : name;
  i_param : cterm 0;
  i_arg : cterm 1;
  i_ckind : sort;
  i_kind : sort;
  i_constrs : option (list constructor)
}.

Definition cstr_one_ind is :=
  match i_constrs is with
    | Some lc => map c_name lc
    | _ => nil
  end.

Definition glob_symb_one_ind is := i_name is :: cstr_one_ind is.

Inductive constr_in_ind : name -> one_ind -> constructor -> Prop :=
  Ok_cstr_ind : forall is lc c,
    i_constrs is = Some lc -> In c lc -> constr_in_ind (c_name c) is c.

Definition one_ind_unique is :=
  forall x c1 c2, constr_in_ind x is c1 -> constr_in_ind x is c2 -> c1 = c2.

Definition ind_pack := list one_ind.

Definition inds_of_pack := map i_name.
Definition cstr_ind_pack := flat_map cstr_one_ind.

Record constant_obj := mk_cobj {
  c_par : cterm 0; c_decl : cdecl 1
}.

Definition glob_symb_cst co := NAME (c_decl co) :: nil.

Definition const_val c d := BODY (c_decl c) = Some d.

Inductive sig_item :=
  | MutInductive : ind_pack -> sig_item
  | Constant : constant_obj -> sig_item
.

Lemma constr_in_pack : forall ip is,
  In is (ip:ind_pack) ->
  forall x c, constr_in_ind x is c -> In (c_name c) (cstr_ind_pack ip).
Proof.
  induction ip; intros.
  inversion H; fail.
  inversion H; subst; clear H.

  inversion_clear H0.
  unfold cstr_ind_pack in |- *; simpl in |- *.
  unfold cstr_one_ind at 1 in |- *; rewrite H.
  apply in_or_app; left.
  apply in_map; assumption.

  unfold cstr_ind_pack in |- *; simpl in |- *.
  elim (cstr_one_ind a); simpl in |- *; intros; auto.
  eapply IHip; eassumption.
Qed.

Definition abs_is is :=
  mk_one_ind (i_name is) (i_param is) (i_arg is)
  (i_ckind is) (i_kind is) None.

Definition abstract_ind : ind_pack -> ind_pack := map abs_is.

Lemma abstract_idem :
  forall ip, abstract_ind (abstract_ind ip) = abstract_ind ip.
Proof.
  simple induction ip; simpl in |- *; intros; auto.
  unfold abstract_ind in |- *; simpl in |- *; fold (abstract_ind l) in |- *.
  fold (abstract_ind (abstract_ind l)) in |- *.
  rewrite H; trivial.
Qed.

Lemma abstract_no_cstr :
  forall ip is x c, In is (abstract_ind ip) ->
    constr_in_ind x is c -> False.
Proof.
  intros ip is x c Habs Hconc.
  induction ip.
  inversion Habs; fail.
  inversion Habs; subst; clear Habs.
  inversion Hconc; subst; clear Hconc.
  discriminate.
  auto; fail.
Qed.

Inductive global_spec : Set :=
  | GlobConst : constant_obj -> global_spec
  | GlobInd : list name -> one_ind -> global_spec
  | GlobCstr : one_ind -> constructor -> global_spec
.

Inductive item_glob_spec : sig_item -> name -> global_spec -> Prop :=
  | Okglob_cst :
    forall co,
      item_glob_spec (Constant co) (NAME (c_decl co)) (GlobConst co)
  | Okglob_ind :
    forall (ip : ind_pack) is,
      In is ip ->
      item_glob_spec (MutInductive ip) (i_name is)
      (GlobInd (inds_of_pack ip) is)
  | Okglob_cstr :
    forall x ip is c,
      In is ip ->
      constr_in_ind x is c ->
      item_glob_spec (MutInductive ip) x (GlobCstr is c)
.

Class sigma := sg_contents : list sig_item.

Definition sign_glob_spec sg x g :=
  exists2 si : _, In si (sg:sigma) & item_glob_spec si x g.

Lemma glob_spec_cons :
  forall sg x g si, sign_glob_spec sg x g -> sign_glob_spec (si :: sg) x g.
Proof.
  simple destruct 1; intros.
  split with x0; auto.
Qed.

Section Isg.

  Context {sg : sigma}.

  Lemma glob_ind_eq_name :
    forall ind inds is,
      sign_glob_spec sg ind (GlobInd inds is) -> ind = i_name is.
  Proof.
    intros ind inds is (itm, insig, gs).
    inversion_clear gs; reflexivity.
  Qed.

  Lemma glob_cstr_eq_name :
    forall x is c, sign_glob_spec sg x (GlobCstr is c) -> x = c_name c.
  Proof.
    intros x is c (itm, insig, gs).
    inversion_clear gs.
    inversion_clear H0; reflexivity.
  Qed.

  Lemma glob_cst_ind :
    forall x is c,
      sign_glob_spec sg x (GlobCstr is c) ->
      exists inds : _, sign_glob_spec sg (i_name is) (GlobInd inds is).
  Proof.
    intros x is c (it, in_sg, gs_it).
    generalize in_sg.
    inversion_clear gs_it; intros.
    split with (inds_of_pack ip).
    split with (MutInductive ip); trivial.
    constructor; trivial.
  Qed.

  Lemma glob_ind_cst :
    forall ind x inds is c,
      sign_glob_spec sg ind (GlobInd inds is) ->
      constr_in_ind x is c -> sign_glob_spec sg x (GlobCstr is c).
  Proof.
    intros ind x inds is c (it, in_sg, gs_it) inc.
    split with it; trivial.
    inversion_clear gs_it.
    constructor; trivial.
  Qed.

  Lemma glob_ind_cst2 :
    forall x is c, sign_glob_spec sg x (GlobCstr is c) -> constr_in_ind x is c.
  Proof.
    intros x is c (it, in_sg, gs_it).
    inversion_clear gs_it; trivial.
  Qed.

  Lemma inv_incl_cst_tl :
    forall x ip c,
      sign_glob_spec (MutInductive ip :: sg) x (GlobConst c) ->
      sign_glob_spec sg x (GlobConst c).
  Proof.
    intros.
    inversion_clear H.
    generalize H0.
    inversion_clear H1; intros.
    inversion_clear H1.
    split with (Constant c); discriminate.
    econstructor; [eassumption| ].
    constructor; assumption.
  Qed.

  Lemma inv_incl_ind_tl :
    forall x c inds is,
      sign_glob_spec (Constant c :: sg) x (GlobInd inds is) ->
      sign_glob_spec sg x (GlobInd inds is).
  Proof.
    intros.
    inversion_clear H.
    generalize H0.
    inversion_clear H1; intros.
    inversion_clear H1.
    split with (MutInductive ip); discriminate.
    econstructor; [eassumption| ].
    constructor; assumption.
  Qed.

  Lemma inv_incl_cstr_tl :
    forall x co is c,
      sign_glob_spec (Constant co :: sg) x (GlobCstr is c) ->
      sign_glob_spec sg x (GlobCstr is c).
  Proof.
    intros.
    inversion_clear H.
    generalize H0.
    inversion_clear H1; intros.
    inversion_clear H1.
    split with (MutInductive ip); discriminate.
    econstructor; [eassumption| ].
    constructor; assumption.
  Qed.

  Lemma inv_cons_induc :
    forall ip x g,
      sign_glob_spec (MutInductive ip :: sg) x g ->
      sign_glob_spec sg x g \/
      (exists2 is : _,
        In is ip &
        g = GlobInd (inds_of_pack ip) is \/
        (exists2 c : _, constr_in_ind x is c & g = GlobCstr is c)).
  Proof.
    simple destruct g; intros.
    left.
    apply (inv_incl_cst_tl H).

    inversion_clear H.
    generalize H0.
    inversion_clear H1; intros.
    generalize H.
    inversion_clear H1; intros.
    injection H2; intros; subst; clear H2.
    right.
    split with o; auto.

    left.
    split with (MutInductive ip0); auto.
    constructor; trivial.

    inversion_clear H.
    generalize H0.
    inversion_clear H1; intros.
    generalize H.
    inversion_clear H1; intros.
    injection H3; intros; subst; clear H3.
    right.
    split with o; auto.
    right.
    split with c; auto.

    left.
    split with (MutInductive ip0); auto.
    constructor; trivial.
  Qed.


  Lemma inv_incl_cst :
    forall x c co,
      sign_glob_spec (Constant c :: sg) x (GlobConst co) ->
      sign_glob_spec sg x (GlobConst co) \/ c = co.
  Proof.
    intros.
    inversion_clear H.
    generalize H0.
    inversion_clear H1; intros.
    inversion_clear H1; auto.
    inversion H; subst; clear H.
    right; reflexivity.
    left.
    split with (Constant co); auto.
    constructor; trivial.
  Qed.

  Lemma inv_incl_ind :
    forall ip ind inds is,
      sign_glob_spec (MutInductive ip :: sg) ind (GlobInd inds is) ->
      sign_glob_spec sg ind (GlobInd inds is) \/ In is ip.
  Proof.
    intros.
    inversion_clear H.
    generalize H0.
    inversion_clear H1; intros.
    generalize H.
    inversion_clear H1; auto.
    inversion H2; subst; clear H2.
    right; assumption.
    left.
    split with (MutInductive ip0); auto.
    constructor; trivial.
  Qed.

  Lemma inv_incl_cstr :
    forall ip x is c,
      sign_glob_spec (MutInductive ip :: sg) x (GlobCstr is c) ->
      sign_glob_spec sg x (GlobCstr is c) \/ In is ip /\ constr_in_ind x is c.
  Proof.
    intros.
    inversion_clear H.
    generalize H0.
    inversion_clear H1; intros.
    generalize H.
    inversion_clear H1; auto.
    inversion H3; subst; clear H3.
    auto; fail.
    left.
    split with (MutInductive ip0); auto.
    constructor; trivial.
  Qed.

  Lemma incl_inv_cst :
    forall c x g,
      sign_glob_spec (Constant c :: sg) x g ->
      sign_glob_spec sg x g \/ g = GlobConst c.
  Proof.
    intros.
    inversion_clear H.
    generalize H1.
    inversion_clear H0; intros.
    subst.
    right.
    inversion_clear H0; trivial.

    left.
    split with x0; trivial.
  Qed.

  Lemma incl_inv_ind :
    forall ip x g,
      sign_glob_spec (MutInductive (abstract_ind ip) :: sg) x g ->
      sign_glob_spec sg x g \/ (exists is : _, g = GlobInd (inds_of_pack ip) is).
  Proof.
    intros.
    inversion_clear H.
    generalize H1.
    inversion_clear H0; intros.
    subst.
    right.
    inversion_clear H0.
    split with is.
    replace (inds_of_pack (abstract_ind ip)) with (inds_of_pack ip); trivial.
    unfold inds_of_pack in |- *.
    elim ip; simpl in |- *; intros; auto.
    rewrite H0; trivial.

    elim abstract_no_cstr with (1 := H) (2 := H2).

    left.
    split with x0; trivial.
  Qed.

  Lemma unique_constr_names :
    forall c1 c2 x is1 is2 ip,
      list_unique (cstr_ind_pack ip) ->
      In is1 ip ->
      In is2 ip ->
      constr_in_ind x is1 c1 ->
      constr_in_ind x is2 c2 -> GlobCstr is1 c1 = GlobCstr is2 c2.
  Proof.
    unfold cstr_ind_pack in |- *.
    intros c1 c2 x is1 is2 ip nu in1.
    generalize nu x is2 c1 c2.
    clear c1 c2 is2 x nu.
    induction ip as [ |is ip]; intros nu x is2 c1 c2 Hin His1 His2.
    destruct in1; fail.
    destruct in1 as [in1|in1].
    subst is.
    generalize His2; clear His2.
    destruct Hin as [Hin|Hin]; intros.
    subst is2; simpl in nu.
    inversion His1 as [is lc c Hlc Hc]; subst; clear His1.
    unfold cstr_one_ind at 1 in nu.
    rewrite Hlc in nu.
    inversion His2 as [is' lc' c' Hlc' Hc' Heq]; subst; clear His2.
    rewrite Hlc' in Hlc; injection Hlc; intros; subst lc'.
    elim (unique_map (unique_app_l nu) Hc' Hc Heq); trivial.

    simpl in nu.
    elim unique_app_diff with (1 := nu) (x := x).
    inversion His1 as [is lc c Hlc Hc]; subst; clear His1.
    unfold cstr_one_ind in |- *.
    rewrite Hlc.
    apply in_map; assumption.

    inversion His2 as [is' lc' c' Hlc' Hc'].
    apply constr_in_pack with (2 := His2); trivial.

    destruct Hin as [Hin|Hin].
    subst is2.
    simpl in nu.
    elim unique_app_diff with (1 := nu) (x := x).
    inversion His2 as [is' lc' c' Hlc' Hc']; subst; clear His2.
    unfold cstr_one_ind in |- *.
    rewrite Hlc'.
    apply in_map; assumption.

    inversion His1 as [is' lc' c' Hlc' Hc'].
    apply constr_in_pack with (2 := His1); trivial.

    apply IHip with x; auto.
    simpl in nu.
    apply unique_app_r with (1 := nu).
  Qed.

  (* Decidability of signature lookup *)

  Lemma search_const :
    forall c,
      {co : _ | sign_glob_spec sg c (GlobConst co)} +
      {(forall co, ~ sign_glob_spec sg c (GlobConst co))}.
  Proof.
    (*
    Realizer [c]
      (Fix search_const
        { search_const [sign:sigma]: ? :=
            Cases sign of
              nil => (inright constant_obj)
            | (cons (Constant co) tsg) =>
                if (eq_name (NAME (c_decl co)) c) then (inleft ? co)
                else (search_const tsg)
            | (cons si tsg) => (search_const tsg)
            end }
       sg).
    *)
    intros c.
    elim sg; clear sg.
    right; red in |- *; intros co Hco.
    destruct Hco as [si Hin Hitem].
    destruct Hin; fail.

    intros [ip|co] sg Hsg.
    destruct Hsg as [[co Hco]|Hsg].
    left; split with co; auto.
    destruct Hco as [si Hin Hitem].
    split with si; auto.

    right; red in |- *; intros co Hco.
    destruct Hco as [si Hin Hitem].
    destruct Hin as [Hin|Hin].
    subst si.
    inversion_clear Hitem.

    elim Hsg with co.
    split with si; trivial.

    case (NAME (c_decl co) == c); intros Heq.
    left.
    split with co.
    split with (Constant co); auto.
    elim Heq.
    constructor.

    destruct Hsg as [[co' Hco']|Hsg].
    left; split with co'; auto.
    destruct Hco' as [si Hin Hitem].
    split with si; auto.

    right; red in |- *; intros co' Hco'.
    destruct Hco' as [si Hin Hitem].
    destruct Hin as [Hin|Hin].
    subst si.
    apply Heq.
    inversion_clear Hitem.
    trivial.

    elim Hsg with co'.
    split with si; trivial.
  Qed.

  Lemma search_constr :
    forall x lc,
      {c : _ | In c lc &  x = c_name c} +
      {forall c, In c lc -> x = c_name c -> False}.
  Proof.
    (*
    Realizer Fix srch_c { srch_c [x:name; lc:(list constructor)]
                       : (sumor constructor) :=
            Cases lc of
              nil => (inright constructor)
            | (cons c cl) =>
                 if (eq_name x (c_name c)) then (inleft ? c)
                 else (srch_c x cl)
            end}.
    *)
    induction lc as [ | c lc ].
    { right; intros c Hc Hx; inversion Hc; fail. }
    {
      destruct (eq_dec x (c_name c)) as [ He | He ].
      {
        left; exists c; [ left; reflexivity | ].
        assumption.
      }
      {
        destruct IHlc as [ IH | IH ].
        {
          destruct IH as [c' Hc_In Hc_eq].
          left; exists c'; [ right | ]; assumption.
        }
        {
          right; intros c' Hc_In Hc_eq.
          destruct Hc_In as [ Hc_In | Hc_In ].
          - subst c'; contradict He; assumption.
          - eapply IH; eassumption.
        }
      }
    }
  Qed.

  Lemma search_constr_ind :
    forall x is,
      {c : _ | constr_in_ind x is c} +
      {forall c, ~ constr_in_ind x is c}.
  Proof.
    intros x (i, p, a, sc, s, [lc| ]).
    destruct (search_constr x lc) as [[c Hin Hname]|Hlc].
    left.
    split with c.
    rewrite Hname.
    split with lc; auto.

    right; red in |- *; intros c Hc.
    specialize (Hlc c).
    inversion Hc as [is lc' c' Hconstrs Hin]; subst.
    simpl in *.
    injection Hconstrs; intros; subst lc'.
    auto; fail.

    right; red in |- *; intros c Hc.
    inversion Hc as [is lc c' Hconstrs Hin]; subst.
    simpl in *.
    discriminate.
  Qed.

  Lemma search_glob_one_ind :
    forall x ip is,
      In is ip ->
      {g : _ | item_glob_spec (MutInductive ip) x g} +
      {x <> i_name is /\ (forall c, ~ constr_in_ind x is c)}.
  Proof.
    (*
    Realizer [x:name; ip:ind_pack; is:one_ind]
      if (eq_name x (i_name is)) then (inleft ? (GlobInd (inds_of_pack ip) is))
      else Cases (search_constr_ind x is) of
             (inleft c) => (inleft ? (GlobCstr is c))
           | inright => (inright global_spec)
           end.
    *)
    intros.
    elim (eq_dec x (i_name is)); intros.
    left.
    split with (GlobInd (inds_of_pack ip) is).
    rewrite a.
    constructor.
    trivial.

    elim (search_constr_ind x is); intros.
    left.
    elim a; intros.
    split with (GlobCstr is x0).
    constructor; trivial.

    right; auto.
  Qed.

  Lemma search_glob_ind_pack :
    forall x ip,
      {g : _ | item_glob_spec (MutInductive ip) x g} +
      {(forall g, ~ item_glob_spec (MutInductive ip) x g)}.
  Proof.
    (*
    Realizer Fix srch_ip
      { srch_ip [x:name; ip:ind_pack] : (sumor global_spec) :=
         Cases ip of
           nil => (inright global_spec)
         | (cons is ipl) =>
             Cases (search_glob_one_ind x ip is) of
               (inleft gs) => (inleft ? gs)
             | inright => (srch_ip x ipl)
             end
         end}.
    *)
    intros.
    elim search_map
      with
        (A := one_ind)
        (B := {g : _ | item_glob_spec (MutInductive ip) x g})
        (C := fun is => x <> i_name is /\ (forall c, ~ constr_in_ind x is c))
        (l := ip); intros; auto.
    right; red in |- *; intros.
    inversion H.
    elim b with is; intros; auto.

    elim b with is; intros; auto.
    red in H6; eauto.

    apply search_glob_one_ind; trivial.
  Qed.

  Lemma search_glob_item :
    forall x it,
      {g : _ | item_glob_spec it x g} + {(forall g, ~ item_glob_spec it x g)}.
  Proof.
    (*
    Realizer [x:name; it:sig_item]
      Cases it of
        (MutInductive ip) => (search_glob_ind_pack x ip)
      | (Constant co) =>
          if (eq_name x (NAME (c_decl co))) then (inleft ? (GlobConst co))
          else (inright global_spec)
      end.
    *)
    simple destruct it; intros.
    apply search_glob_ind_pack.

    elim (eq_dec x (NAME (c_decl c))); intros.
    left.
    split with (GlobConst c).
    rewrite a.
    constructor.

    right; red in |- *; intros.
    apply b.
    inversion_clear H; trivial.
  Qed.

  Lemma search_global :
    forall x,
      {g : _ | sign_glob_spec sg x g} + {(forall g, ~ sign_glob_spec sg x g)}.
  Proof.
    intros.
    elim search_map
      with
        (A := sig_item)
        (B := {g : _ | sign_glob_spec sg x g})
        (C := fun si => forall g, ~ item_glob_spec si x g)
        (l := sg); intros; auto.
    right; red in |- *; intros.
    inversion_clear H.
    elim b with (2 := H1); trivial.

    elim search_glob_item with x x0; intros; auto.
    left.
    inversion_clear a.
    split with x1.
    split with x0; trivial.
  Qed.

End Isg.
