(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Set Implicit Arguments.
Unset Strict Implicit.
Require Import General.
Require Export Relations.

Section SortsOfV6.

  (* Definition des composantes du calcul *)

  Inductive calc : Set :=
    | Pos : calc
    | Neg : calc.

  Inductive srt_v6 : Set :=
    | Sprop : calc -> srt_v6
    | Stype : nat -> srt_v6.

  Inductive axiom_v6 : srt_v6 -> srt_v6 -> Prop :=
    | ax_prop : forall c n, axiom_v6 (Sprop c) (Stype n)
    | ax_type : forall n m, n < m -> axiom_v6 (Stype n) (Stype m).

  Inductive pairs_v6 : srt_v6 -> srt_v6 -> srt_v6 -> Prop :=
    | pair_set_l : forall c, pairs_v6 (Sprop Pos) (Sprop c) (Sprop Pos)
    | pair_set_r : forall c, pairs_v6 (Sprop c) (Sprop Pos) (Sprop Pos)
    | pair_prop : pairs_v6 (Sprop Neg) (Sprop Neg) (Sprop Neg)
    | pair_type_l :
        forall n m c, n <= m -> pairs_v6 (Stype n) (Sprop c) (Stype m)
    | pair_type_r :
        forall n m c, n <= m -> pairs_v6 (Sprop c) (Stype n) (Stype m)
    | pair_type :
        forall n m p,
        n <= p -> m <= p -> pairs_v6 (Stype n) (Stype m) (Stype p).


  Inductive rules_v6 : srt_v6 -> srt_v6 -> srt_v6 -> Prop :=
    | rule_prop_l : forall c s, rules_v6 (Sprop c) s s
    | rule_prop_r : forall c s, rules_v6 s (Sprop c) (Sprop c)
    | rule_type :
        forall n m p,
        n <= p -> m <= p -> rules_v6 (Stype n) (Stype m) (Stype p).

  Inductive univ_v6 : srt_v6 -> srt_v6 -> Prop :=
    | univ_prop : forall s, univ_v6 (Sprop Neg) s
    | univ_prop_set : forall c n, univ_v6 (Sprop c) (Stype n)
    | univ_type : forall n m, n <= m -> univ_v6 (Stype n) (Stype m).

  Definition univ := clos_refl_trans _ univ_v6.


  Inductive elim_v6 : srt_v6 -> srt_v6 -> srt_v6 -> Prop :=
    | elim_prop : forall s, elim_v6 s (Sprop Neg) (Sprop Neg)
    | elim_set_pred : forall c s, elim_v6 (Sprop c) (Sprop Pos) s
    | elim_set_impr : forall n c, elim_v6 (Stype n) (Sprop Pos) (Sprop c)
    | elim_type_pred1 : forall c n s, elim_v6 (Sprop c) (Stype n) s
    | elim_type_pred2 :
        forall n1 n2 s, n1 <= n2 -> elim_v6 (Stype n1) (Stype n2) s.


  Hint Resolve ax_prop ax_type rule_prop_l rule_prop_r rule_type pair_set_l
    pair_set_r pair_prop pair_type_l pair_type_r pair_type univ_prop
    univ_prop_set univ_type elim_prop elim_set_pred elim_set_impr
    elim_type_pred1 elim_type_pred2: core.

  Hint Unfold univ: core.

  (* Inversion et Decidabilite de l'inclusion entre sortes *)

  Let univ_trans : forall x y z, univ x y -> univ y z -> univ x z.
Proof rt_trans srt_v6 univ_v6.


  Inductive inv_univ : srt_v6 -> srt_v6 -> Prop :=
    | iu_prop_set : forall s, inv_univ (Sprop Neg) s
    | iu_set : inv_univ (Sprop Pos) (Sprop Pos)
    | iu_pr_ty : forall c n, inv_univ (Sprop c) (Stype n)
    | iu_type : forall n m, n <= m -> inv_univ (Stype n) (Stype m).

  Hint Resolve iu_prop_set iu_set iu_pr_ty iu_type: core.


  Lemma inv_univ_trans :
   forall x y z, inv_univ x y -> inv_univ y z -> inv_univ x z.
simple induction 1; intros; auto.
inversion_clear H0; auto.

inversion_clear H1.
apply iu_type.
apply le_trans with m; auto.
Qed.


  Lemma univ_inv :
   forall s s', univ s s' -> forall P : Prop, (inv_univ s s' -> P) -> P.
simple induction 1.
simple induction 1; auto.

simple destruct x; auto.
simple destruct c; auto.

intros.
apply H4.
apply inv_univ_trans with y.
apply H1; auto.

apply H3; auto.
Qed.



  Lemma calc_dec : forall c c' : calc, decide (c = c').
simple destruct c; simple destruct c'; (right; discriminate) || auto.
Defined.


  Lemma v6_sort_dec : forall s s' : srt_v6, decide (s = s').
simple destruct s; simple destruct s'; intros; try (right; discriminate).
elim calc_dec with c c0; intros.
left; elim a; auto.

right; red in |- *; intros H; apply b; injection H; auto.

elim eq_nat_dec with n n0; intros.
left; elim a; auto.

right; red in |- *; intros H; apply b; injection H; auto.
Defined.


  Lemma univ_v6_dec : forall s s', decide (univ s s').
intros [[| ]| u1].
(* Set *)
intros [[| ]| u2];
 [ left; red in |- *; apply rt_refl | right | left; repeat constructor ].
red in |- *; intro univ_set_prop; apply univ_inv with (1 := univ_set_prop);
 intro ui_set_prop; inversion ui_set_prop.

(* Prop *)
left; repeat constructor.

(* Type *)
intros [c2| u2].
right; red in |- *; intro univ_type_prop;
 apply univ_inv with (1 := univ_type_prop); intro ui_type_prop;
 inversion ui_type_prop.
case (le_gt_dec u1 u2).
intros le_u1_u2; left; repeat constructor; trivial.

intro gt_u1_u2; right; red in |- *; intro univ_ty;
 apply univ_inv with (1 := univ_ty); intro ui_type; 
 inversion ui_type; omega.
Defined.
(*Realizer [s,s':srt_v6]
  Cases s s' of
    (Sprop Neg) _           => true
  | (Sprop Pos) (Sprop Neg) => false
  | (Sprop _)   _           => true
  | (Stype n)   (Stype n')  => (le_gt_dec n n')
  | _           _           => false
  end.
*)

  (* Inference des axiomes et regles *)

  Lemma v6_inf_axiom : forall s, {sp : _ | ppal (axiom_v6 s) univ sp}.
Proof.
intros [c| n].
exists (Stype 0).
split; intros; auto.
inversion_clear H; auto with arith.

exists (Stype (S n)).
split; intros; auto.
inversion_clear H; auto with arith.
Defined.
(*Realizer [s:srt_v6]
  Cases s of
    (Sprop c) => (Stype O)
  | (Stype n) => (Stype (S n))
  end.
*)

  Lemma v6_not_topped : forall s, in_dom_rel axiom_v6 s.
Proof.
intros.
elim (v6_inf_axiom s); intros.
split with x.
apply (pp_ok p).
Qed.


  Lemma v6_elim_typed :
   forall sc si se, elim_v6 sc si se -> in_dom_rel axiom_v6 se.
Proof.
intros.
apply v6_not_topped.
Qed.

  Lemma v6_hierarchy :
   forall s1 s2,
   univ s1 s2 -> in_dom_rel axiom_v6 s2 -> in_dom_rel axiom_v6 s1.
Proof.
intros.
apply v6_not_topped.
Qed.



  Lemma v6_inf_pair :
   forall x1 x2,
   {x3 : _ | pairs_v6 x1 x2 x3 & 
   forall s1 s2 s3,
   pairs_v6 s1 s2 s3 -> univ x1 s1 -> univ x2 s2 -> univ x3 s3}.
Proof.
intros [[| ]| n1].
(* Set *)
intros [c2| n2].
exists (Sprop Pos); auto; intros.
apply univ_inv with (1 := H0); intros.
generalize H.
inversion_clear H2; intros.
inversion_clear H2; auto.

inversion_clear H2; auto.

exists (Stype n2); auto; intros.
apply univ_inv with (1 := H1); intros.
generalize H.
inversion_clear H2; intros.
inversion_clear H2.
cut (n2 <= m0); auto.
omega.

cut (n2 <= p); auto; omega.

(* Prop *)
intros [[| ]| n2].
exists (Sprop Pos); auto; intros.
apply univ_inv with (1 := H1); intros.
generalize H.
inversion_clear H2; intros.
inversion_clear H2; auto.

inversion_clear H2; auto.

exists (Sprop Neg); auto; intros.

exists (Stype n2); auto; intros.
apply univ_inv with (1 := H1); intros.
generalize H.
inversion_clear H2; intros.
inversion_clear H2.
cut (n2 <= m0); auto.
omega.

cut (n2 <= p); auto; omega.

(* Type *)
intros [c2| n2].
exists (Stype n1); auto; intros.
apply univ_inv with (1 := H0); intros.
generalize H.
inversion_clear H2; intros.
inversion_clear H2; auto.
cut (n1 <= m0); auto; omega.

cut (n1 <= p); auto; omega.

exists (Stype (max_nat n1 n2)); intros.
constructor.
unfold max_nat in |- *.
elim (le_gt_dec n1 n2); auto.

unfold max_nat in |- *.
elim (le_gt_dec n1 n2); auto with arith.

apply univ_inv with (1 := H0); intros.
apply univ_inv with (1 := H1); intros.
generalize H.
inversion_clear H2.
inversion_clear H3; intros.
inversion_clear H3.
cut (max_nat n1 n2 <= p); auto.
apply least_upper_bound_max_nat; omega.
Defined.
(*Realizer [x1,x2:srt_v6]
  Cases x1 x2 of
    (Sprop Neg) (Sprop Neg) => (Sprop Neg)
  | (Sprop c1)  (Sprop c2)  => (Sprop Pos)
  | (Stype n)   (Sprop c)   => (Stype n)
  | (Sprop c)   (Stype n)   => (Stype n)
  | (Stype n1)  (Stype n2)  => (Stype (max_nat n1 n2))
  end.
*)

  Lemma v6_all_pair : forall s1 s2, exists s3 : _, pairs_v6 s1 s2 s3.
Proof.
intros.
elim v6_inf_pair with s1 s2; intros.
split with x; trivial.
Qed.



  Lemma v6_inf_rule :
   forall x1 x2,
   {x3 : _ | rules_v6 x1 x2 x3 & 
   forall s1 s2 s3,
   rules_v6 s1 s2 s3 -> univ x1 s1 -> univ x2 s2 -> univ x3 s3}.
(*Realizer [x1,x2:srt_v6]
  Cases x1 x2 of
    (Sprop c)   _        => x2
  | (Stype n) (Sprop c') => (Sprop c')
  | (Stype n) (Stype n') => (Stype (max_nat n n'))
  end.
*)
Proof.
intros [c| n].
(* Prop & Set *)
intros x2; exists x2; intros; auto.
apply univ_inv with (1 := H1); intros; auto.
generalize H.
inversion_clear H2; intros; auto.
inversion_clear H2; auto.

inversion_clear H2; auto.

inversion_clear H2; eauto with arith.

(* Type *)
intros [c'| n'].
exists (Sprop c'); intros; auto.
apply univ_inv with (1 := H1); intros; auto.
generalize H.
inversion_clear H2; intros; auto.

inversion_clear H2; auto.

inversion_clear H2; auto.

exists (Stype (max_nat n n')); intros.
unfold max_nat in |- *.
elim (le_gt_dec n n'); auto with arith.

apply univ_inv with (1 := H0); intros; auto.
apply univ_inv with (1 := H1); intros; auto.
generalize H.
inversion_clear H2.
inversion_clear H3; intros.
inversion_clear H3.
cut (max_nat n n' <= p); auto.
apply least_upper_bound_max_nat; omega.
Defined.
(*Realizer [x1,x2:srt_v6]
  Cases x1 x2 of
    (Sprop c)   _        => x2
  | (Stype n) (Sprop c') => (Sprop c')
  | (Stype n) (Stype n') => (Stype (max_nat n n'))
  end.
*)


  Lemma v6_full : forall s1 s2, exists s3 : _, rules_v6 s1 s2 s3.
Proof.
intros.
elim v6_inf_rule with s1 s2; intros.
split with x; trivial.
Qed.


  Lemma elim_v6_dec : forall sc si se, decide (elim_v6 sc si se).
(*Realizer [sc,si,se:srt_v6]
  Cases si sc se of
    (Sprop Neg) _          _         => (v6_sort_dec se (Sprop Neg))
  | (Sprop Pos) (Sprop c)  _         => true
  | (Sprop Pos) (Stype n)  (Sprop c) => true
  | (Stype n)   (Sprop c)  _         => true
  | (Stype n)   (Stype n') _         => (le_gt_dec n' n)
  | _           _          _         => false
  end.
*)
intros sc [[| ]| n] se.
(* si = Set *)
case sc.
intros c.
left; auto.

case se.
left; auto.

right; red in |- *; intros abs_elim; inversion abs_elim.

(* si = Prop *)
elim (v6_sort_dec se (Sprop Neg)).
intro eq_prop; left; rewrite eq_prop; auto.

intro neq_prop; right; red in |- *; intro abs_elim; apply neq_prop;
 inversion abs_elim; trivial.

(* Type *)
case sc.
intro c'.
left; auto.

intro n'.
elim (le_gt_dec n' n).
intro predic; left; auto.

intro impredic; right; red in |- *; intro abs_elim; inversion_clear abs_elim;
 omega.
Defined.

  Lemma elim_v6_sat :
   forall sc si s1 s2, univ s1 s2 -> elim_v6 sc si s2 -> elim_v6 sc si s1.
Proof.
intros.
generalize H.
inversion_clear H0; intros.
apply univ_inv with (1 := H0); intros.
inversion_clear H1; trivial.

trivial.

apply univ_inv with (1 := H0); intros.
inversion_clear H1; trivial.

trivial.

auto.
Qed.


End SortsOfV6.


(* Uniform interface of sorts *)
Require Export GenericSort.

  Definition sort_of_gen (gs : gen_sort) :=
    match gs with
    | Gprop => Some (Sprop Neg)
    | Gset => Some (Sprop Pos)
    | Gtype n => Some (Stype n)
    | _ => None
    end.

  Definition gen_of_sort (s : srt_v6) :=
    match s with
    | Sprop Neg => Gprop
    | Sprop Pos => Gset
    | Stype n => Gtype n
    end.