(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Signature.
Require Export Marks.

Set Implicit Arguments.
Unset Strict Implicit.

(* Defining the typing rules of fixpoints *)

Unset Elimination Schemes.

Record fix_info : Set := mk_fi {
  fi_parname : name;
  fi_party : term;
  fi_argname : name;
  fi_indty : name;
  fi_ipar : term;
  fi_iarg : term;
  fi_lem : le_marks;
  fi_marks : term;
  fi_fixty : term
}.

Definition lift_finfo n fi k :=
  mk_fi (fi_parname fi) ^(n//k)(fi_party fi) (fi_argname fi)
  (fi_indty fi) ^(n//S k)(fi_ipar fi) ^(n//S k)(fi_iarg fi)
  (fi_lem fi) ^(n//S k)(fi_marks fi) ^(n//S (S k))(fi_fixty fi).

Definition subst_finfo s fi k :=
  mk_fi (fi_parname fi) (fi_party fi)[k:=s] (fi_argname fi)
  (fi_indty fi) (fi_ipar fi)[S k:=s] (fi_iarg fi)[S k:=s]
  (fi_lem fi) (fi_marks fi)[S k:=s] (fi_fixty fi)[S (S k):=s].

Definition fix_scheme (b : bool) fd :=
  Forall (fi_parname fd) : ^(1)(fi_party fd),
  (Forall (fi_argname fd)
    : Ind{fi_indty fd; if b then S (fi_lem fd) else 0
      // ^(1//1)(fi_ipar fd),,
      (^(1//1)(fi_iarg fd),,#1 && ^(1//1)(fi_marks fd))},
    ^(1//2)(fi_fixty fd)).

Definition fix_ty fdi :=
  Forall (fi_parname fdi) : fi_party fdi,
  (Forall (fi_argname fdi)
    : Ind{fi_indty fdi; fi_lem fdi
      // fi_ipar fdi,,(fi_iarg fdi,,fi_marks fdi)},
    fi_fixty fdi).

Definition fix_types fds := mkvect (map fix_ty fds).

Definition fix_bodies m fds :=
  Forall m : Mark,
  (Forall x_ : mkvect (map (fix_scheme false) fds),
    ^(1)(mkvect (map (fix_scheme true) fds))).

Lemma fix_scheme_lift : forall b n fd k,
  ^(n//S k)(fix_scheme b fd) = fix_scheme b (lift_finfo n fd k).
Proof.
  unfold fix_scheme, lift in |- *.
  simpl in |- *; intros.
  repeat rewrite <- permute_lift_rec1; auto with arith.
Qed.

Lemma fix_scheme_subst : forall b s fd k,
  (fix_scheme b fd)[S k:=s] = fix_scheme b (subst_finfo s fd k).
Proof.
  unfold fix_scheme, lift in |- *.
  simpl in |- *; intros.
  repeat rewrite <- commut_lift_subst_rec1; auto with arith.
Qed.

Lemma fix_types_lift : forall fds n k,
  ^(n//k)(fix_types fds) = fix_types (map (fun fi => lift_finfo n fi k) fds).
Proof.
  intros.
  unfold fix_types in |- *; simpl in |- *.
  rewrite mkvect_lift.
  rewrite map_map.
  rewrite map_map.
  apply f_equal with (f := mkvect).
  apply map_ext; intros.
  reflexivity.
Qed.

Lemma fix_types_subst : forall s fds k,
  (fix_types fds)[k:=s] = fix_types (map (fun fi => subst_finfo s fi k) fds).
Proof.
  intros.
  unfold fix_types in |- *; simpl in |- *.
  rewrite mkvect_subst.
  rewrite map_map.
  rewrite map_map.
  apply f_equal with (f := mkvect).
  apply map_ext; intros.
  reflexivity.
Qed.

Lemma fix_bodies_lift : forall m fds n k,
  ^(n//k)(fix_bodies m fds) =
  fix_bodies m (map (fun fi => lift_finfo n fi k) fds).
Proof.
  intros.
  unfold fix_bodies, lift in |- *; simpl in |- *.
  rewrite <- permute_lift_rec1; auto with arith.
  rewrite mkvect_lift.
  rewrite mkvect_lift.
  repeat rewrite map_map.
  elim map_ext with
    (f := fun x => ^(n//S k)(fix_scheme false x))
    (g := fun x => fix_scheme false (lift_finfo n x k));
    intros; auto.
  elim map_ext with
    (f := fun x => ^(n//S k)(fix_scheme true x))
    (g := fun x => fix_scheme true (lift_finfo n x k));
    intros; auto.
  apply fix_scheme_lift.

  apply fix_scheme_lift.
Qed.

Lemma fix_bodies_subst : forall u m fds k,
  (fix_bodies m fds)[k:=u] =
  fix_bodies m (map (fun fi => subst_finfo u fi k) fds).
Proof.
  intros.
  unfold fix_bodies, lift in |- *; simpl in |- *.
  rewrite <- commut_lift_subst_rec1; auto with arith.
  rewrite mkvect_subst.
  rewrite mkvect_subst.
  repeat rewrite map_map.
  elim map_ext with
    (f := fun x => (fix_scheme false x)[S k:=u])
    (g := fun x => fix_scheme false (subst_finfo u x k));
    intros; auto.
  elim map_ext with
    (f := fun x => (fix_scheme true x)[S k:=u])
    (g := fun x => fix_scheme true (subst_finfo u x k));
    intros; auto.
  apply fix_scheme_subst.

  apply fix_scheme_subst.
Qed.

(* Two basic conditions entailing subject reduction
   for fixpoint expansions *)

(* Ensures that the expanded fixpoint (after erasing the mark) has a
   smaller type than the fixpoint *)
Lemma fix_body_incl_types : forall e fds,
  incl_mark e (mkvect (map (fix_scheme true) fds))[0:=##] (fix_types fds).
Proof.
  intros.
  rewrite mkvect_subst.
  rewrite map_map.
  unfold fix_scheme, fix_types in |- *.
  apply mkvect_mark; intros.
  unfold lift in |- *.
  simpl in |- *.
  repeat rewrite simpl_subst_rec; auto with arith.
  repeat rewrite lift_rec0.
  red in |- *.
  unfold fix_ty in |- *.
  apply Clos_prd; auto.
  apply Clos_prd; auto.
  apply incl_add_mt_mark.
Qed.

(* Ensures that the fixpoint can be applied to the bodies after erasing
   the mark *)
Lemma fix_exp_incl : forall e fds,
  incl_mark e (fix_types fds) (mkvect (map (fix_scheme false) fds))[0:=##].
Proof.
  intros.
  rewrite mkvect_subst.
  rewrite map_map.
  unfold fix_types, fix_ty, fix_scheme in |- *; simpl in |- *.
  apply mkvect_mark; intros.
  simpl in |- *.
  repeat rewrite simpl_subst_rec; auto with arith.
  repeat rewrite lift_rec0.
  rewrite simpl_subst; auto.
  rewrite lift0.
  unfold lift in |- *.
  red in |- *.
  apply Clos_prd; auto.
  apply Clos_prd; auto.
  apply incl_str_marks.
Qed.

(* Type checking decidability *)

Lemma fix_ty_inj : forall fd1 fd2, fix_ty fd1 = fix_ty fd2 -> fd1 = fd2.
Proof.
  unfold fix_ty in |- *.
  intros.
  simplify_eq H.
  case fd1; case fd2; simpl in |- *; repeat simple destruct 1; trivial.
Qed.

Lemma fix_types_inj :
  forall fd1 fd2, fix_types fd1 = fix_types fd2 -> fd1 = fd2.
Proof.
  unfold fix_types in |- *.
  simple induction fd1; simple destruct fd2; simpl in |- *; intros.
  auto.

  discriminate H.

  discriminate H0.

  elim H with l0.
  elim fix_ty_inj with a f; trivial.
  change
    (match (fix_ty a ** mkvect (map fix_ty l)) with X ** _ => X | t => t end =
      match (fix_ty f ** mkvect (map fix_ty l0)) with X ** _ => X | t => t end) in |- *.
  elim H0; trivial.

  change
    (match (fix_ty a ** mkvect (map fix_ty l)) with _ ** X => X | t => t end =
      match (fix_ty f ** mkvect (map fix_ty l0)) with _ ** X => X | t => t end) in |- *.
  elim H0; trivial.
Qed.

Lemma is_fix_ty_dec : forall t, dec_find (fun fd => t = fix_ty fd).
Proof.
  refine (fun t =>
    match t as t0 return dec_find (fun fd => t0 = fix_ty fd) with
      | Prd x pty (Prd y (Cst1 (MutInd i m) (Pair0 p (Pair0 a l))) ty) => inleft _ _
      | _ => inright _ _
    end); try (intros; discriminate).
  exists (mk_fi x pty y i p a m l ty); reflexivity.
Defined.

Lemma is_fix_tyl_dec :
  forall l, dec_find (fun fds => mkvect l = fix_types fds).
Proof.
  (*Realizer Fix fty
    {fty [l:?]: ? :=
       Cases l of
         nil => (inleft ? (nil fix_info))
       | (cons u tl) =>
           Cases (is_fix_ty_dec u) (fty tl) of
             (inleft f1) (inleft fdtl) => (inleft ? (cons ? f1 fdtl))
           | _ _ => (inright (list fix_info))
           end
       end}.*)
  fix fix_tyl 1.
  intros [| u tl]; simpl in |- *.
  left; exists (nil (A:=fix_info)); reflexivity.

  case (is_fix_ty_dec u).
  intros (fi, is_fty); subst u.
  case (fix_tyl tl).
  intros (fds, is_ftyl); rewrite is_ftyl; clear is_ftyl tl.
  left; exists (fi :: fds); reflexivity.

  unfold fix_types in |- *.
  intros not_ftyl; red in not_ftyl.
  right; red in |- *; intros [| h l]; simpl in |- *; intros eq_v;
    try discriminate.
  injection eq_v; intros eq_vect; rewrite eq_vect in not_ftyl; eauto.

  unfold fix_types in |- *.
  intros not_ftyl; red in not_ftyl.
  right; red in |- *; intros [| h l]; simpl in |- *; intros eq_v;
    try discriminate.
  injection eq_v; intros; subst u; eauto.
Defined.

Lemma is_fix_types_dec : forall t, dec_find (fun fds => t = fix_types fds).
Proof.
  (*Realizer [t]
    Cases (is_vect_dec t) of
      (inleft lt) => (is_fix_tyl_dec lt)
    | _ => (inright (list fix_info))
    end.*)
  intros t.
  case (is_vect_dec t).
  intros (v, eq_t); subst t.
  apply is_fix_tyl_dec.

  intros not_vect; red in not_vect.
  unfold fix_types in |- *.
  right; red in |- *; intros fds eq_t; subst t; eauto.
Defined.
