(**************************************************************************)
(*                      Certified Extraction in Coq                       *)
(*                                                                        *)
(*                            Stéphane Glondu                             *)
(*                                                                        *)
(*                           © 2007-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export MoreCic.

Set Implicit Arguments.
Unset Strict Implicit.

(***** Règles de réduction des termes logiques extraits *****)

Section DummyReduction.

  (* dummyfied rules from redn_term *)

  Inductive dummy_proj1 (e : env) : term -> term -> Prop :=
      dummy_proj1_intro :
        dummy_proj1 e (##..L) ##.

  Inductive dummy_proj2 (e : env) : term -> term -> Prop :=
      dummy_proj2_intro :
        dummy_proj2 e (##..R) ##.

  Inductive dummy_projr (e : env) : term -> term -> Prop :=
      dummy_projr_intro : forall x,
        dummy_projr e (##..(x)) ##.

  Inductive dummy_beta (e : env) : term -> term -> Prop :=
      dummy_beta_intro : forall n,
        dummy_beta e (## @ n) ##.

  Inductive dummy_iota_fix (e : env) : term -> term -> Prop :=
      dummy_iota_fix_intro :
        forall n arg fx efix,
        fix_free fx efix ->
        dummy_iota_fix e (((fx .^ n) ..L @ arg) @ ##)
          (((efix .^ n) ..L @ arg) @ ##).

  Inductive dummy_iota_case (e : env) : term -> term -> Prop :=
      dummy_iota_case_intro : forall pred patts br,
        dummy_iota_case e (Match ## => pred with true — patts => br end) ##.

  (* proof-irrelevant pattern-matching *)

  Inductive dummy_iota_case_pi (e : env) : term -> term -> Prop :=
      dummy_iota_case_pi_intro :
        forall C pred br,
        dummy_iota_case_pi e
          (Match ## => pred with false — C::nil => br end)
          (br..L @ ##).

  Definition dummy_redn_term {sg : sigma} :=
    reunion_list
      (Cs_rl redn_term
        (Cs_rl dummy_proj1
          (Cs_rl dummy_proj2
            (Cs_rl dummy_projr
              (Cs_rl dummy_beta
                (Cs_rl dummy_iota_fix
                  (Cs_rl dummy_iota_case
                    (Cs_rl dummy_iota_case_pi
                      Mt_rl)))))))).

End DummyReduction.


Section ExtractionProperties.

  Context {sg : sigma}.
  Context {Hsg : wf_sign sg}.
  Let sign_ok := wf_sign_ok Hsg.
  Existing Instance sign_ok.


(***** Définition du prédicat d'extraction *****)

Inductive Pe : env -> term -> term -> Prop :=
  | Pe_Prop : forall e t T,
      sg – e ⊢ t : T ->
      sg – e ⊢ T : Srt (Sprop Neg) ->
      Pe e t ##
  | Pe_Srt : forall e s, Pe e (Srt s) ##
  | Pe_Rel : forall e i, Pe e (Rel i) (Rel i)
  | Pe_Cst0 : forall e c, Pe e (Cst0 c) ##
  | Pe_Proj1 : forall e t t', Pe e t t' -> Pe e (t..L) (t'..L)
  | Pe_Proj2 : forall e t t', Pe e t t' -> Pe e (t..R) (t'..R)
  | Pe_App : forall e t t' u u', Pe e t t' -> Pe e u u' -> Pe e (t@u) (t'@u')
  | Pe_Const : forall e t t' x, Pe e t t' -> Pe e (Cst1 (Const x) t) (Cst1 (Const x) t')
  | Pe_Struct : forall e t t' l T, Pe e t t' -> Pe e {<l : T := t>} {<l : ## := t'>}
  | Pe_Record : forall e l T, Pe e {<l : T>} ##
  | Pe_MutInd : forall e n l t, Pe e Ind{n; l // t} ##
  | Pe_Projr : forall e t t' x, Pe e t t' -> Pe e t..(x) t'..(x)
  | Pe_Fix : forall e t t' T, Pe e t t' -> Pe e Fix{T,,t} Fix{##,,t'}
  | Pe_Case : forall e c c' pr br br' pats b,
      Pe e c c' -> Pe e br br' ->
      Pe e (Match c => pr with b — pats => br end) (Match c' => ## with b — pats => br' end)
  | Pe_Constr : forall e p p' a a' C, Pe e p p' -> Pe e a a' -> Pe e Cstr{C//p,,a} Cstr{C//p',,a'}
  | Pe_Prd : forall e t T x, Pe e (Forall x : T, t) ##
  | Pe_Lam : forall e t t' T x,
      Pe (e; x : T) t t' ->
      Pe e (Fun x : T => t) (Fun x : ## => t')
  | Pe_Sum : forall e t u, Pe e (t ** u) ##
  | Pe_Pair : forall e t t' u u', Pe e t t' -> Pe e u u' -> Pe e (t,,u) (t',,u').

Hint Constructors Pe.

Ltac invPe :=
  lazymatch goal with
    | _ : Pe _ _ (_ .^ _) |- _ => fail
    | H : Pe _ _ (_ _) |- _ => inversion H; subst; clear H
    | H : Pe _ _ (_ _ _) |- _ => inversion H; subst; clear H
    | H : Pe _ _ (_ _ _ _) |- _ => inversion H; subst; clear H
    | H : Pe _ _ (_ _ _ _ _) |- _ => inversion H; subst; clear H
  end.


(***** Lemmes divers *****)

Lemma inv_Pe_proj : forall n e t t0',
  Pe e t (t0' .^ n) -> exists t0, t = t0 .^ n /\ Pe e t0 t0'.
Proof.
  induction n; simpl; intros.
  eauto.
  rewrite proj_comm in H.
  inversion H; subst.
  destruct (IHn _ _ _ H3) as (t1, (Ht1, Ht1')).
  exists t1; split.
  rewrite proj_comm; f_equal; auto.
  auto.
Qed.

Lemma Pe_proj : forall n e t t',
  Pe e t t' -> Pe e (t .^ n) (t' .^ n).
Proof.
  induction n; simpl; intros; auto.
Qed.
Hint Resolve Pe_proj.


Lemma inv_Pe_logical_ind : forall e t I n X,
  sg – e ⊢ t : Ind{I ; n // X} -> Pe e t ## ->
  sg – e ⊢ Ind{I ; n // X} : %%(Sprop Neg).
Proof.
  intros e t I n X HI Ht.
  assert (He : sg – ⊢ e) by eauto using typ_wf.

  (* par cas sur la sorte de l'inductif *)

  let t :=
    InvTypHyp HI; intros; simpl in *;
    try inv mem_sign0; try inv mem_sign1; repeat contradict_le
  in
  inversion Ht; subst; [ |t|t|t|t|t|t].

  (* le seul cas non contradictoire, Prop *)
  (* on met T en forme hnf *)
  rename T into T'.
  destruct (whnf_cci (type_correctness _ _ _ H)) as (T, HT).
  assert (H102 : sg – e ⊢ t : T)
    by (eapply subj_red_ty; eauto).
  assert (H103 : sg – e ⊢ T : %%(Sprop Neg))
    by (eapply subject_reduction; eauto).
  clear H H0 HT T'.
  (* on prend le type principal de t (tau) *)
  destruct (inf_ppal_type _ t He) as [(Hppal, ?)|Ha]; subst.
  destruct Hppal as (?, tau, Hppal); simpl in *; subst.
  hnf in Hppal.
  destruct Hppal as (Htau, Hppal).
  assert (Hp1 := Hppal _ HI); simpl in Hp1.
  assert (Hp2 := Hppal _ H102); simpl in Hp2.
  (* on met ce type principal en forme hnf (τ) dans les inégalités *)
  destruct (whnf_cci (type_correctness _ _ _ Htau)) as (τ, Hτ).
  assert (H100 : sg – e ⊢ τ <= Ind { I ; n // X }).
    apply cci_trans with tau; auto.
  assert (H101 : sg – e ⊢ τ <= T).
    apply cci_trans with tau; auto.
  (* on détermine la forme de τ *)
  inversion H100 using hn_ind_inv_left.
    eapply type_correctness.
    apply subj_red_ty with tau; eauto.
    auto.
  clear Hp1 Hp2 Hτ Hppal Htau tau.
  intros; subst.
  inversion H101 using hn_ind_inv_right.
    econstructor; eassumption.
    auto.
  intros; subst.
  (* on détermine la forme de X *)
  assert (H104 : wf_type e Ind { I ; n // X })
    by (eapply type_correctness; eassumption).
  invc H104.
  InvTypHyp H; intros.
  inv in_sign1.
  (* on identifie toutes les sortes mises en jeu *)
  InvTypHyp H103; intros.
  inv in_sign1.
  assert (H105 := ok_gs_inj H10 H17).
  injection H105; intros; subst inds0 is0; clear H105.
  simpl in H8.
  assert (H105 := inv_sort H8).
  assert (H106 : i_kind is = Sprop Neg).
    simple eapply Prop_is_minimal; apply H8.
  rewrite <- H106.
  (* le but devient facile à prouver... *)
  inversion H3 using inv_pair_sum; intros; auto.
  inversion H9 using inv_pair_sum; intros; auto.
  eapply typ_induc0; eauto.
  (* cas où t n'est pas typable *)
  edestruct Ha; eassumption.
  (* ouf ! *)
Qed.

Lemma logical_case_form : forall e t C pr br T,
  closed 0 t -> gclosed t ->
  sg – e ⊢ Match t => pr with false — C::nil => br end : T ->
  Pe e t ## ->
  exists p, exists a,
  R_rt (wctxt redn_term) e t Cstr{C//p,,a} /\ Pe e a ##.
Proof.
  intros.
  (* t must be of an inductive type I *)
  InvTypHyp H1; intros.
  inversion H3; subst; clear H3.
  inversion H16; subst; clear H16.
  InvTypHyp H6; intros; clear H6.
  change (sg – e ⊢ a ** b <=
    Ind  {i_name is; m // par,, arg,, l} **
    (case_predicate_form is se)[0:=par] ** (mkvect tybr)[0:=par]
  ) in H13.
  apply inv_sum_l in H13.
  inversion H4 using inversion_wf; simpl; intros.
  InvTypHyp H6; intros; clear H6.
  assert (Ht : sg – e ⊢ t : Ind{i_name is ; m // par,, arg,, l}).
    eapply Typ_conv; eassumption.
  rename H14 into His.
  clear a b s s0 s2 s3 H4 H5 H10 H11 H12 H13 H16 H17 H18.
  set (I := i_name is) in *.
  (* I has a single constructor, which is C *)
  inversion H9; subst; clear H9.
  inversion H11; subst; clear H11.
  inversion H6; subst; clear H6.
  rename H4 into HC_in_is.
  inversion H3; subst; clear H3.
  inversion HC_in_is; subst.
  assert (lc = x0) by (rewrite H3 in H4; congruence); subst x0.
  rename H3 into Hconstrs.
  rename H9 into Hx_in_is.
  rename H6 into HC_is_uniq.
  destruct H8 as [?|[_ Hckind]]; [discriminate| ].
  assert (HC := HC_in_is).
  eapply glob_ind_cst in HC; [ |eassumption].
  clear H4 H1 H5 H7 pr y.
  set (C := c_name x) in *.
  (* I is logical *)
  assert (Hsort_of_I : sg – e ⊢ Ind {I; m // par,, arg,, l} : %%(Sprop Neg)).
    eapply inv_Pe_logical_ind; eassumption.
  clear s1 His.
  (* t reduces to Cstr{C // par,, a'} *)
  inversion Ht using closed_ind_form_cstr; try assumption; [ ]; intros.
  assert (is0 = is) by (eapply inj_ind_is; eassumption); subst is0.
  assert (inds0 = inds) by (eapply inj_ind_il; eassumption); subst inds0.
  apply glob_ind_cst2 in H3.
  inversion H3; subst; clear H3.
  assert (lc0 = lc) by congruence; subst lc0.
  assert (Hcx : c_name c = c_name x).
    apply HC_is_uniq in H6.
    inversion H6; subst.
    symmetry; assumption.
    inversion H3; fail.
  rewrite Hcx in *; clear c H6 Hcx.
  do 3 esplit; [eassumption|].
  (* a' is logical *)
  eapply subjred_rt_wctxt in Ht; [ |eassumption].
  rename H15 into HI.
  clear H1 H2 H4 H5.
  InvTypHyp Ht; intros.
  inversion H1; subst; clear H1.
  assert (HH : GlobCstr is x = GlobCstr is0 c) by (eapply ok_gs_inj; eassumption).
  injection HH; intros; subst is0 c; clear HH.
  InvTypHyp H4; intros.
  change (sg – e ⊢ a ** b <= i_param is ** (c_arg x)[0:=par]) in H8.
  assert (H20 := H8).
  apply inv_sum_r in H8.
  apply inv_sum_l in H20.
  inversion H2; subst; clear H2.
  InvTypHyp H14; intros.
  eapply Typ_conv in H6; [ |eassumption|eassumption].
  eapply Typ_conv in H1; [ |eassumption|eassumption].
  rename H6 into Ha', H1 into Hpar.
  clear - Ha' Hpar Hckind HC HC_in_is Hsg.
  (* something should definitely be done in Cic about the following... *)
  assert (Harg : sg – (add_var nil x_ (i_param is)) ⊢ c_arg x : %%(i_ckind is)).
    clear Ha' Hpar.
    destruct HC as [si Hsi1 Hsi2].
    inversion Hsi2 as [ | |? ip ? ? His _]; subst; clear Hsi2.
    apply in_split in Hsi1.
    destruct Hsi1 as [l1 [l2 Happ]]; subst.
    induction l1 as [ |d l1]; simpl in *.
      inversion Hsg as [ | |? ? _ _ [ok _]]; subst.
      specialize (ok _ _ _ His HC_in_is).
      destruct ok as [ok _ _ _].
      eapply typ_incl; [eapply incl_ind_abstr| ]; eassumption.
    eapply typ_incl; [eapply InclSign.incl_cons| ]; apply IHl1.
    generalize (l1 ++ MutInductive ip :: l2) Hsg; clear.
    induction l; intro H; inversion H; subst; auto; fail.
  rewrite Hckind in Harg; clear Hckind.
  eapply Pe_Prop; [eassumption| ].
  change (%%(Sprop Neg)) with ((%%so)[0:=par]).
  eapply @substitution with (x:=x_); [ |eassumption].
  apply typ_wf in Hpar.
  clear - Harg Hpar; induction e.
  assumption.
  change (%%so) with (^(1//1)(%%so)).
  replace (c_arg x : term) with (^(1//1)(c_arg x)) by apply closed1_lift.
  eapply typ_weak.
  - eassumption.
  - eapply IHe; inversion Hpar; assumption.
  - eapply ins_tl'.
    econstructor.
    unfold liftd_rec; simpl; f_equal.
    symmetry; apply closed0_lift.
  Existential 1 := Hsg.
Qed.

(***** Lemmes inspirés du typage *****)

Lemma Pe_weak : forall (g : env) d e t t',
  sg – ⊢ (g; d) ->
  Pe e t t' ->
  forall f k, ins_in_env g d k e f -> Pe f ^(1//k)t ^(1//k)t'.
Proof.
  induction 2; simpl; intros; auto.
  induction H2.
  assert (sg – d :: g ⊢ ^(1)t : ^(1)T)
    by (eapply thinning; auto).
  assert (sg – d :: g ⊢ ^(1)T : ^(1)%%(Sprop Neg))
    by (eapply thinning; auto).
  unfold lift in *; simpl in *.
  econstructor; eauto.
  assert (sg – liftd_rec 1 t0 n :: g0 ⊢ ^(1//S n)t : ^(1//S n)T)
    by (eapply typ_weak; eauto).
  assert (sg – liftd_rec 1 t0 n :: g0 ⊢ ^(1//S n)T : ^(1//S n)%%(Sprop Neg))
    by (eapply typ_weak; eauto).
  simpl in *.
  econstructor; eauto.
Qed.

Lemma Pe_thinning : forall t' t e f n,
  trunc n f e -> sg – ⊢ f -> Pe e t t' -> Pe f ^(n)t ^(n)t'.
Proof.
  induction 1; intros.
  autorewrite with core; auto.
  rewrite simpl_lift.
  pattern (^(S k)t').
  rewrite simpl_lift.
  unfold lift at 1 3.
  eapply Pe_weak; eauto.
  apply IHtrunc; auto.
  inversion H0; auto.
Qed.

Lemma Pe_sub : forall e t t',
   Pe e t t'->
   forall T, sg – e ⊢ t : T ->
   forall u u' d g,
   sg – g ⊢ u : TYPE d ->
   Pe g u u' ->
   forall n f, sub_in_env g u d n e f -> Pe f t[n:=u] t'[n:=u'].
Proof.
  induction 1; simpl; intros; eauto; try (inv_typ; constructor; eauto).
  (* Prop *)
  assert (sg – f ⊢ t[n:=u] : T[n:=u])
    by (eapply typ_sub; eauto).
  assert (sg – f ⊢ T[n:=u]: %%(Sprop Neg)[n:=u])
    by (eapply typ_sub; eauto).
  simpl in *; econstructor; eauto.
  (* Rel *)
  destruct (lt_eq_lt_dec n i); auto.
  destruct s; auto; subst.
  eapply Pe_thinning; eauto.
  eapply sub_env_trunc; eauto.
  eapply sub_in_env_wf; [eassumption| |eassumption].
  eapply typ_wf; eauto.
Qed.


(***** Lemme principal *****)

Lemma Pe_redn_correct : forall e t t' u' T,
  closed O t ->
  gclosed t ->
  sg – e ⊢ t : T ->
  Pe e t t' ->
  dummy_redn_term e t' u' ->
  exists u, R_t (wctxt redn_term) e t u /\ Pe e u u'.
Proof.
  destruct 5; destruct H3.

  (* delta *)
  inv delta; invPe; inv closed; inv lt.
  inv reunion_list.

  (* delta_glob *)
  inv delta_glob; invPe; inv gclosed.
  destruct (H7 x); auto.
  inv reunion_list.

  (* proj1 *)
  inv proj1; repeat invPe.
  econstructor; split.
  left; apply WCtx_rule; do 2 right; left; constructor.
  auto.
  inv reunion_list.

  (* proj2 *)
  inv proj2; repeat invPe.
  econstructor; split.
  left; apply WCtx_rule; do 3 right; left; constructor.
  auto.
  inv reunion_list.

  (* projr *)
  inv projr; repeat invPe.
  econstructor; split.
  left; apply WCtx_rule; do 4 right; left; constructor; eauto.
  auto.
  inv reunion_list.

  (* beta *)
  inv beta; repeat invPe.
  econstructor; split.
  left; apply WCtx_rule; do 5 right; left; constructor.
  inv_typ.
  assert (sg – e ⊢ u : T0) by (eapply typ_app_arg; eauto).
  apply Pe_sub with (e; x:T0) u1 (mkdecl x T0 None) e; eauto.
  unfold add_var; constructor.
  unfold val_ok; simpl; auto.
  inv reunion_list.

  (* iota_fix *)
  inv iota_fix; repeat invPe.
  destruct (inv_Pe_proj H6) as (x, (Hx, Hx')); subst.
  inv fix_free; invc Hx'.
  econstructor; split.
  left; apply WCtx_rule; do 6 right; left; do 2 constructor.
  repeat constructor; auto.
  inv reunion_list.

  (* iota_case *)
  inv iota_case; repeat invPe.
  econstructor; split.
  left; apply WCtx_rule; do 7 right; left; constructor; eauto.
  repeat constructor; auto.
  contradiction.

  (* dummy_proj1 *)
  (* needs a generalization of logical_case_form *)
  admit.
  inv reunion_list.

  (* dummy_proj2 *)
  (* should be the same as dummy_proj1 *)
  admit.
  inv reunion_list.

  (* dummy_projr *)
  (* should be quite similar to dummy_projx *)
  admit.
  inv reunion_list.

  (* dummy_beta *)
  inv dummy_beta.
  invPe.
  (* t0 should be lambda with fixapp... *)
  admit.
  inv reunion_list.

  (* dummy_iota_fix *)
  inv dummy_iota_fix.
  (* On retrouve la forme de t *)
  invPe.
  invc H8.
  invc H7.
  rename fx into fx'; rename H9 into Hu_Pe.
  destruct (inv_Pe_proj H6) as (fx, (?, Hfx_Pe)); subst.
  inv fix_free; invc Hfx_Pe.
  destruct (typ_fix_arg H1) as [I [n_u [p_u [a_u [l_u Hu_typ]]]]].
  assert (H100 : exists C, exists a',
    R_rt (wctxt redn_term) e u Cstr {C // p_u,, a'})
    by (unfold redn_term; eapply closed_ind_form; auto_closed; eauto).
  destruct H100 as [C [a' Hu_red]].
  econstructor; split.
  hnf; eapply step_clos_trans.
  (* On réduit d'abord le terme déstructuré *)
  unfold redn_term.
  apply rt_wctxt_stable.
  apply WCtx_cst.
  apply WCtx_pair_r.
  apply WCtx_rule.
  eauto.
  (* On effectue ensuite une iota_fix-réduction *)
  left; apply WCtx_rule; do 6 right; left; do 2 constructor.
  (* Le réduit s'extrait toujours comme on veut *)
  repeat constructor; auto.
  econstructor.
  eapply subjred_rt_wctxt; eauto.
  eapply inv_Pe_logical_ind; eauto.
  inv reunion_list.

  (* dummy_iota_case *)
  inv dummy_iota_case.
  invPe.
  (* c reduces to constructor, triggering iota_case *)
  admit.
  inv reunion_list.

  (* dummy_iota_case_pi *)
  inversion H3 as [C pr' br']; subst; clear H3.
  (* On retrouve la forme de t *)
  invPe.
  assert (Hc_closed : closed 0 c) by auto_closed.
  assert (Hc_gclosed : gclosed c) by auto_closed.
  destruct (logical_case_form Hc_closed Hc_gclosed H1 H7) as (p, (a, (Hc, Ha))).
  set (t := Match c => pr with false — C::nil => br end) in *.
  set (t' := Match (Cstr  {C // p,, a}) => pr with false — C::nil => br end) in *.
  econstructor; split.
  hnf; apply step_clos_trans with t'.
  (* On réduit d'abord le terme déstructuré *)
  unfold t, t', redn_term.
  apply rt_wctxt_stable.
  apply WCtx_cst.
  apply WCtx_pair_l.
  apply WCtx_rule.
  auto.
  (* On effectue ensuite une iota_case-réduction *)
  unfold t'.
  left; apply WCtx_rule; do 7 right; left; constructor; hnf; eauto.
  (* Le réduit s'extrait toujours comme on veut *)
  simpl; eauto.

  contradiction.
  Existential 1 := Hsg.
Qed.


(***** Réductions impossibles *****)

Ltac contradict_redn_simpl :=
  match goal with
    | H : redn_term _ _ _ |- _  => invc H; [inv delta|]
    | H : redn_term _ _ _ _ |- _  => invc H; [inv delta|]
    | H : dummy_redn_term _ _ _ |- _  => invc H
    | H : reunion_list (Cs_rl ?a _) _ _ _ |- _ => invc H; [inv a|]
    | H : reunion_list (Cs_rl ?a _) _ _ _ |- _ => invc H; [inv a|]
    | H : wctxt _ _ ## _ |- _ => invc H
    | _ : reunion_list Mt_rl _ _ _ |- _ => contradiction
  end.


Lemma Mt_mark_is_not_reducible1 : forall e t,
  ~ R_t (wctxt redn_term) e ## t.
Proof.
  intros e t H.
  assert (H100 : exists t', wctxt redn_term e ## t')
    by (induction H; eauto).
  destruct H100 as (t', Ht').
  repeat contradict_redn_simpl.
Qed.

Lemma Mt_mark_is_not_reducible2 : forall e n t,
  Rsteps (wctxt redn_term e) n ## t ->
  n = 0.
Proof.
  induction n; intros; simpl in *; auto.
  destruct H as (?, (H100, ?)).
  destruct (Mt_mark_is_not_reducible1 (t_step _ _ _ _ H100)).
Qed.

Lemma Mt_mark_is_not_dummy_reducible1 : forall e t,
  ~ R_t (wctxt dummy_redn_term) e ## t.
Proof.
  intros e t H.
  assert (H100 : exists t', wctxt dummy_redn_term e ## t')
    by (induction H; eauto).
  destruct H100 as (t', Ht').
  repeat contradict_redn_simpl.
Qed.

Lemma Mt_mark_is_not_dummy_reducible2 : forall e n t,
  Rsteps (wctxt dummy_redn_term e) n ## t ->
  n = 0.
Proof.
  induction n; intros; simpl in *; auto.
  destruct H as (?, (H100, ?)).
  destruct (Mt_mark_is_not_dummy_reducible1 (t_step _ _ _ _ H100)).
Qed.


Ltac contradict_redn := contradict_redn_simpl ||
  match goal with
    | H : R_t _ _ ## _ |- _ =>
      destruct (Mt_mark_is_not_reducible1 H)
    | H : Rsteps _ _ ## _ |- _ =>
        let Hnew := fresh in
        assert (Hnew := Mt_mark_is_not_reducible2 H);
        rewrite Hnew in *;
        rewrite <- H in *; clear H Hnew
    | H : Rsteps _ _ ## _ |- _ =>
        let Hnew := fresh in
        assert (Hnew := Mt_mark_is_not_dummy_reducible2 H);
        rewrite Hnew in *;
        rewrite <- H in *; clear H Hnew
  end.


(***** Inversion des hypothèses de réduction *****)

Lemma inv_dummy_wctxt_pair_one : forall e a b u,
  wctxt dummy_redn_term e (a,,b) u ->
  exists t,
  (u = (t,,b) /\ wctxt dummy_redn_term e a t) \/
  (u = (a,,t) /\ wctxt dummy_redn_term e b t).
Proof.
  intros; inv wctxt; repeat contradict_redn; eauto.
Qed.

Lemma inv_dummy_wctxt_pair : forall e n a b t,
  Rsteps (wctxt dummy_redn_term e) n (a,,b) t ->
  exists n1, exists n2, exists a', exists b',
  t = (a',,b') /\ n = n1 + n2 /\
  Rsteps (wctxt dummy_redn_term e) n1 a a' /\
  Rsteps (wctxt dummy_redn_term e) n2 b b'.
Proof.
  induction n; intros; simpl in *.
  subst.
  exists 0; exists 0; exists a; exists b; simpl; auto.
  destruct H as (y, (Hy, Hyt)).

  destruct (inv_dummy_wctxt_pair_one Hy) as (t', [(?, Ha)|(?,Hb)]); subst.

  destruct (IHn _ _ _ Hyt) as (n1, (n2, (t'', (b', (Ht, (Hn, (Htt', Hb'))))))).
  assert (Hat' := Rsteps_one (A:=term) Ha).
  assert (Hat'' := Rsteps_trans Hat' Htt').
  subst.
  exists (1+n1); exists n2; exists t''; exists b'; repeat split; simpl; auto with arith.

  destruct (IHn _ _ _ Hyt) as (n1, (n2, (t'', (b', (Ht, (Hn, (Htt', Hb'))))))).
  assert (Hbt' := Rsteps_one (A:=term) Hb).
  assert (Hbt'' := Rsteps_trans Hbt' Hb').
  subst.
  exists n1; exists (1+n2); exists t''; exists b'; repeat split; simpl; auto with arith.
Qed.

Lemma inv_wctxt_pair_one : forall e a b u,
  wctxt redn_term e (a,,b) u ->
  exists t,
  (u = (t,,b) /\ wctxt redn_term e a t) \/
  (u = (a,,t) /\ wctxt redn_term e b t).
Proof.
  intros; inv wctxt; repeat contradict_redn; eauto.
Qed.

Lemma inv_wctxt_pair : forall e n a b t,
  Rsteps (wctxt redn_term e) n (a,,b) t ->
  exists n1, exists n2, exists a', exists b',
  t = (a',,b') /\ n = n1 + n2 /\
  Rsteps (wctxt redn_term e) n1 a a' /\
  Rsteps (wctxt redn_term e) n2 b b'.
Proof.
  induction n; intros; simpl in *.
  subst.
  exists 0; exists 0; exists a; exists b; simpl; auto.
  destruct H as (y, (Hy, Hyt)).
  destruct (inv_wctxt_pair_one Hy) as (t', [(?, Ha)|(?,Hb)]); subst.

  destruct (IHn _ _ _ Hyt) as (n1, (n2, (t'', (b', (Ht, (Hn, (Htt', Hb'))))))).
  assert (Hat' := Rsteps_one (A:=term) Ha).
  assert (Hat'' := Rsteps_trans Hat' Htt').
  subst.
  exists (1+n1); exists n2; exists t''; exists b'; repeat split; simpl; auto with arith.

  destruct (IHn _ _ _ Hyt) as (n1, (n2, (t'', (b', (Ht, (Hn, (Htt', Hb'))))))).
  assert (Hbt' := Rsteps_one (A:=term) Hb).
  assert (Hbt'' := Rsteps_trans Hbt' Hb').
  subst.
  exists n1; exists (1+n2); exists t''; exists b'; repeat split; simpl; auto with arith.
Qed.


Ltac auto_typ :=
  match goal with
    | _ : ?sg – ?e ⊢ ?a : ?b |- ?sg – ⊢ ?e => apply typ_wf with a b; auto
    | |- ?sg – ⊢ (_; _ : Cst0 (oper_id Mark)) => apply Wf_cons_var with (Sprop Neg); repeat constructor
    | |- ?sg – ⊢ (_; _ : %%(Sprop _)) => apply Wf_cons_var with (Stype 0); repeat constructor
    | |- ?sg – ⊢ ?e => repeat constructor
  end.


Theorem Pe_wctxt_correct : forall e t t' u' T,
  closed O t ->
  gclosed t ->
  sg – e ⊢ t : T ->
  Pe e t t' ->
  wctxt dummy_redn_term e t' u' ->
  exists u, R_t (wctxt redn_term) e t u /\ Pe e u u'.
Proof.
  intros.
  revert t T H H0 H1 H2.
  induction H3; intros.

  (* rule *)
  destruct (Pe_redn_correct H0 H1 H2 H3 H)
    as (x, (Hx1, Hx2)).
  exists x; split; auto using WCtx_rule.

  (* cst1 *)
  invPe.

    (* Proj1 *)
    assert (H10 : exists u, R_t (wctxt redn_term) e t1 u /\ Pe e u t')
      by (inv_typ; eapply IHwctxt; auto_closed; eauto).
    destruct H10 as (u, (Hu1, Hu2)).
    unfold redn_term.
    exists (u..L); split; auto using t_wctxt_stable.

    (* Proj2 *)
    assert (H10 : exists u, R_t (wctxt redn_term) e t1 u /\ Pe e u t')
      by (inv_typ; eapply IHwctxt; auto_closed; eauto).
    destruct H10 as (uu, (Hu1, Hu2)).
    unfold redn_term.
    exists (uu..R); split; auto using t_wctxt_stable.

    (* App *)
    destruct (inv_dummy_wctxt_pair_one H3)
      as (t, [(Heq,Ht)|(Heq,Ht)]); subst.

    assert (H10 : exists x,
      R_t (wctxt redn_term) e (t1,,u) x /\ Pe e x (t,,u'))
      by (inv_typ; eapply (IHwctxt (t1,,u)); auto_closed; eauto).
    destruct H10 as (x, (Hx1, Hx2)).
    invc Hx2.
    unfold redn_term.
    exists (t0 @ u0); split; auto using t_wctxt_stable.

    assert (H10 : exists x,
      R_t (wctxt redn_term) e (t1,,u) x /\ Pe e x (t'0,,t))
      by (inv_typ; eapply (IHwctxt (t1,,u)); auto_closed; eauto).
    destruct H10 as (x, (Hx1, Hx2)).
    invc Hx2.
    unfold redn_term.
    exists (t0 @ u0); split; auto using t_wctxt_stable.

    (* Const *)
    (* a priori, il y aussi moyen de le faire sans contradiction *)
    inv_gclosed.
    destruct (H6 x); auto.

    (* Struct *)
    assert (H10 : exists u : term,
      R_t (wctxt redn_term) e (##,,t1) u /\ Pe e u t').
      inv_typ.
      apply IHwctxt with (Mark**b); auto_closed.

      (* <preuve de typage> *)
      apply Typ_pair with s2.
      repeat constructor; repeat auto_typ.
      auto.
      apply Typ_sum with (Sprop Neg) s2.
      repeat constructor; repeat auto_typ.
      auto.
      destruct s2; try destruct c; constructor; auto.
      (* </preuve de typage> *)

    destruct H10 as (u, (Hu1, Hu2)).
    destruct (inv_dummy_wctxt_pair_one H3) as (t, [(Heq,Ht)|(Heq,Ht)]);
      subst; invc Hu2.
    repeat contradict_redn.

    destruct (t_Rsteps Hu1) as (n, (Hn, H101)).
    destruct (inv_wctxt_pair H101) as (n1, (n2, (?, (t1', (?, (?, (?, Ht1))))))); subst.
    invc H2.
    exists {<l : T0 := t1'>}; split; auto.
    unfold redn_term; apply t_wctxt_stable.
    apply WCtx_cst.
    apply WCtx_pair_r.
    apply t_wctxt_stable2.
    hnf; apply Rsteps_clos_t with n2; auto.
    cut (n1 = 0); intros; subst; auto with arith.
    destruct (eq_nat_dec n1 0); auto.
    destruct n1; auto.
    simpl in *.
    destruct H5 as (?, (?, ?)).
    repeat contradict_redn.

    (* Field *)
    assert (H10 : exists u : term, R_t (wctxt redn_term) e t1 u /\ Pe e u t')
      by (inv_typ; eapply IHwctxt; auto_closed; eauto).
    destruct H10 as (u, (Hu1, Hu2)).
    unfold redn_term.
    exists (u..(x)); split; auto using t_wctxt_stable.

    (* Fixp *)
    (* La preuve est identique au cas Struct *)
    assert (H10 : exists u : term,
      R_t (wctxt redn_term) e (##,,t1) u /\ Pe e u t').
      inv_typ.
      apply IHwctxt with (Mark**b); auto_closed.

      (* <preuve de typage> *)
      apply Typ_pair with s2.
      repeat constructor; repeat auto_typ.
      auto.
      apply Typ_sum with (Sprop Neg) s2.
      repeat constructor; repeat auto_typ.
      auto.
      destruct s2; try destruct c; constructor; auto.
      (* </preuve de typage> *)

    destruct H10 as (u, (Hu1, Hu2)).
    destruct (inv_dummy_wctxt_pair_one H3) as (t, [(Heq,Ht)|(Heq,Ht)]);
      subst; invc Hu2.
    repeat contradict_redn.

    destruct (t_Rsteps Hu1) as (n, (Hn, H101)).
    destruct (inv_wctxt_pair H101) as (n1, (n2, (?, (t1', (?, (?, (?, Ht1))))))); subst.
    invc H2.
    exists Fix{T0,,t1'}; split; auto.
    unfold redn_term; apply t_wctxt_stable.
    apply WCtx_cst.
    apply WCtx_pair_r.
    apply t_wctxt_stable2.
    hnf; apply Rsteps_clos_t with n2; auto.
    cut (n1 = 0); intros; subst; auto with arith.
    destruct (eq_nat_dec n1 0); auto.
    destruct n1; auto.
    simpl in *.
    destruct H5 as (?, (?, ?)).
    repeat contradict_redn.

    (* MutCase *)
    assert (H10 : exists u,
      R_t (wctxt redn_term) e (c0,, ##,, br) u /\
      Pe e u t').
      inv_typ.
      apply IHwctxt with (a**Mark**b1); auto_closed.

      (* <preuve de typage> *)
      destruct (v6_all_pair s1 s5) as (smax, Hsmax).
      apply Typ_pair with smax.
      auto.
      apply Typ_pair with s5.
      repeat constructor; repeat auto_typ.
      auto.
      apply Typ_sum with (Sprop Neg) s5.
      repeat constructor; repeat auto_typ.
      auto.
      destruct s5; try destruct c; constructor; auto.
      apply Typ_sum with s1 s5; auto.
      apply Typ_sum with (Sprop Neg) s5; auto.
      repeat constructor; repeat auto_typ.
      destruct s5; try destruct c; constructor; auto.
      (* </preuve de typage> *)

    destruct H10 as (u, (Hu1, Hu2)).
    unfold redn_term.

    (* <passage laborieux> *)
    destruct (inv_dummy_wctxt_pair (Rsteps_one H3)) as (n1, (n2, (c'', (?, (?, (?, (Hc'', H100))))))).
    destruct (inv_dummy_wctxt_pair H100) as (n3, (n4, (?, (br'', (?, (?, (?, Hbr''))))))).
    clear H100; subst; repeat invPe.
    destruct (t_Rsteps Hu1) as (n5, (Hn5, Hu1_steps)).
    destruct (inv_wctxt_pair Hu1_steps) as (n6, (n7, (c0', (?, (?, (?, (Hc0', H100))))))).
    destruct (inv_wctxt_pair H100) as (n8, (n9, (?, (br''', (?, (?, (?, Hbr'''))))))).
    clear H100; subst; inv @eq.

    assert (H2 := Mt_mark_is_not_reducible2 H12).
    subst n8.
    rewrite <- H12 in Hu1_steps; clear H12.

    assert (H2 := Mt_mark_is_not_dummy_reducible2 H7).
    subst n3.
    rewrite <- H7 in IHwctxt |- *; clear H7.
    replace (n6 + (0 + n9)) with (n6 + n9) in *; auto with arith.
    replace (n1 + (0 + n4)) with (n1 + n4) in *; auto with arith.
    exists (Match c0' => pr with b — pats => br''' end); split; auto.
    assert (Hu1' : R_t (wctxt redn_term) e (c0,, pr,, br) (c0',, pr,, br''')).
      destruct (zerop n6); subst; simpl in *; subst; unfold redn_term in *.
      apply t_wctxt_stable; apply WCtx_pair_r; apply WCtx_pair_r; apply t_wctxt_stable2.
      hnf; apply Rsteps_clos_t with n9; auto.
      hnf; apply clos_trans_step with (c0',, pr,, br).
      apply rt_wctxt_stable; apply WCtx_pair_r; apply WCtx_pair_r; apply rt_wctxt_stable2.
      hnf; apply Rsteps_clos_rt with n9; auto.
      apply t_wctxt_stable; apply WCtx_pair_l; apply t_wctxt_stable2.
      hnf; apply Rsteps_clos_t with n6; auto.
    apply t_wctxt_stable; apply WCtx_cst; apply t_wctxt_stable2; auto.
    (* </passage laborieux> *)

    (* Cstr *)
    assert (H10 : exists u, R_t (wctxt redn_term) e (p,,a) u /\ Pe e u t')
      by (inv_typ; eapply IHwctxt; auto_closed; eauto).
    destruct H10 as (u, (Hu1, Hu2)).
    assert (H20 : exists p'', exists a'', t' = (p'',, a'')).
      clear - H3.
      inv wctxt.
      inv dummy_redn_term.
      assert (HH : red redn_term e (p',, a') t').
        apply rt_step, Ctx_rule; assumption.
      inversion HH using red_pair; intros.
      repeat esplit; reflexivity.
      repeat (inv reunion_list; [lazymatch goal with H : _ _ _ t' |- _ => inversion H; fail end| ]).
      contradiction.
      repeat esplit; reflexivity.
      repeat esplit; reflexivity.
    destruct H20 as [p'' [a'' ?]]; subst.
    assert (H20 : exists P, exists A, u = (P,, A)).
      clear - Hu1.
      remember (p,, a) as t.
      revert p a Heqt.
      induction Hu1; intros; subst.
      apply inv_wctxt_pair_one in H.
      destruct H as [t [[? H]|[? H]]]; subst; repeat esplit; reflexivity.
      destruct (IHHu1_1 _ _ (eq_refl _)) as [P [A ?]].
      eapply IHHu1_2.
      eassumption.
    destruct H20 as [P [A ?]]; subst.
    inversion Hu2; subst.
    exists (Cstr {C // P,, A}); split;
    hnf; try apply t_wctxt_stable; auto.

  (* lam_l *)
  invPe.
  inv wctxt.
  repeat contradict_redn.

  (* prd_l *)
  invPe.

  (* sum_l, sum_r *)
  invPe.
  invPe.

  (* pair_l *)
  invPe.
  assert (H10 : exists u, R_t (wctxt redn_term) e t0 u /\ Pe e u m')
    by (inv_typ; eapply IHwctxt; auto_closed; eauto).
  destruct H10 as (t10, (H10, H11)).
  unfold redn_term.
  exists (t10,,u); split;
  hnf; try apply t_wctxt_stable; auto.

  (* pair_r *)
  invPe.
  assert (H10 : exists t10, R_t (wctxt redn_term) e u t10 /\ Pe e t10 n')
    by (inv_typ; eapply IHwctxt; auto_closed; eauto).
  destruct H10 as (t10, (H10, H11)).
  unfold redn_term.
  exists (t0,,t10); split;
  hnf; try apply t_wctxt_stable; auto.
Qed.


Lemma Pe_t_wctxt_correct : forall e t t' u' T,
  closed O t ->
  gclosed t ->
  sg – e ⊢ t : T ->
  Pe e t t' ->
  R_t (wctxt dummy_redn_term) e t' u' ->
  exists u, R_t (wctxt redn_term) e t u /\ Pe e u u'.
Proof.
  intros.
  revert t H H0 H1 H2.
  induction H3 as [|t1' t2' t3']; intro t1; intros.
  eapply Pe_wctxt_correct; eauto.
  assert (H100 : exists t2, R_t (wctxt redn_term) e t1 t2 /\ Pe e t2 t2') by eauto.
  destruct H100 as (t2, (?, ?)).
  assert (sg – e ⊢ t2 : T)
    by (apply (@subjred_t_wctxt sg Hsg) with t1; auto).
  assert (closed 0 t2) by (unfold redn_term in *; eapply closed_stable_t_wctxt; eauto).
  assert (gclosed t2) by (unfold redn_term in *; eapply gclosed_stable_t_wctxt; eauto).
  assert (H100 : exists t3, R_t (wctxt redn_term) e t2 t3 /\ Pe e t3 t3') by eauto.
  destruct H100 as (t3, (?, ?)).
  exists t3; split; eauto.
  hnf; apply t_trans with t2; auto.
Qed.


(***** Normalisation des termes extraits *****)

Lemma Pe_sn_compat : forall e t T,
  closed 0 t -> gclosed t -> sg – e ⊢ t : T ->
  sn (R_t (wctxt redn_term)) e t ->
  forall t', Pe e t t' ->
  sn (R_t (wctxt dummy_redn_term)) e t'.
Proof.
  induction 4 as [t]; intros.
  constructor.
  intros u' Hu't'.
  destruct (Pe_t_wctxt_correct H H0 H1 H4 Hu't') as (u, (Hu, Huu')).
  unfold sn in H3.
  unfold redn_term in *.
  apply H3 with u; eauto.
  eapply closed_stable_t_wctxt; eauto.
  eapply gclosed_stable_t_wctxt; eauto.
  eapply subjred_t_wctxt; eauto.
  Existential 1 := Hsg.
Qed.

Theorem Pe_redn_sn : forall e t t' T,
  closed O t ->
  gclosed t ->
  sg – e ⊢ t : T ->
  Pe e t t' ->
  sn (wctxt dummy_redn_term) e t'.
Proof.
  intros.
  apply sn_Rt_R.
  eapply Pe_sn_compat; eauto.
  apply sn_R_Rt.
  eapply wctxt_is_strongly_normalizing; eauto.
  Existential 1 := Hsg.
Qed.


(***** Dual of Pe_redn_correct *****)

Lemma Pe_redn_correct_source : forall e t u t' T,
  closed O t ->
  gclosed t ->
  sg – e ⊢ t : T ->
  Pe e t t' ->
  redn_term e t u ->
  exists u', R_rt dummy_redn_term e t' u' /\ Pe e u u'.
Proof.
  intros e t u t' T Hct HCt Ht Het Hrt.
  inversion Hrt as [Irt|Jrt].
  (* delta *)
  inversion Irt as [n x v V Hv]; subst.
  inversion Het as [? ? T' Ht' HT| |? i| | | | | | | | | | | | | | | | ]; subst.
  exists ##; split.
  apply rt_refl; fail.
  apply Pe_Prop with T'; [ |eassumption].
  eapply subj_red_redn; eassumption.
  inversion Hct; subst; edestruct lt_n_O; eassumption.
  destruct Jrt as [Krt|Jrt].
  (* delta_glob *)
  inversion Krt as [d m x c Hx Hc]; subst.
  inversion Het as [? ? T' Ht' HT| | | | | | | | | | | | | | | | | | ]; subst.
  exists ##; split.
  apply rt_refl; fail.
  apply Pe_Prop with T'; [ |eassumption].
  eapply subj_red_redn; eassumption.
  inversion HCt as [ | | |? ? ? Hclos| | | | ]; subst.
  edestruct Hclos; reflexivity.
  destruct Jrt as [Krt|Jrt].
  (* proj1 *)
  inversion Krt as [t1 t2]; subst t u.
  rename t' into v'.
  inversion Het as [? ? T' Ht' HT| | | |? t t' He| | | | | | | | | | | | | | ]; subst v'; subst.
  exists ##; split.
  apply rt_refl; fail.
  apply Pe_Prop with T'; [ |eassumption].
  eapply subj_red_redn; eassumption.
  inversion He as [? ? T' Ht' HT| | | | | | | | | | | | | | | | | |? ? t1' ? t2' Ht1 Ht2]; subst.
  admit (* missing rule: ##..L -> ## *).
  exists t1'; split.
  apply rt_step; left; do 2 right; left; constructor; fail.
  assumption.
  destruct Jrt as [Krt|Jrt].
  (* proj2 *)
  inversion Krt as [t1 t2]; subst t u.
  rename t' into Z.
  inversion Het as [? ? T' Ht' HT| | | | |? t t' He| | | | | | | | | | | | | ]; subst Z; subst.
  exists ##; split.
  apply rt_refl; fail.
  apply Pe_Prop with T'; [ |eassumption].
  eapply subj_red_redn; eassumption.
  inversion He as [? ? T' Ht' HT| | | | | | | | | | | | | | | | | |? ? t1' ? t2' Ht1 Ht2]; subst.
  admit (* missing rule: ##..R -> ## *).
  exists t2'; split.
  apply rt_step; left; do 3 right; left; constructor; fail.
  assumption.
  destruct Jrt as [Krt|Jrt].
  (* projr *)
  inversion Krt as [x f Ts ts n Hxn]; subst.
  rename t' into Z.
  inversion Het as [? ? T' Ht' HT| | | | | | | | | | |? ? ? ? He| | | | | | | ]; subst Z; subst.
  exists ##; split.
  apply rt_refl; fail.
  apply Pe_Prop with T'; [ |eassumption].
  eapply subj_red_redn; eassumption.
  inversion He as [? ? T' Ht' HT| | | | | | | |? ? ts' ? ? Hts| | | | | | | | | | ]; subst.
  admit (* missing rule: ##..(x) -> ## *).
  exists ((ts' .^ n)..L); split.
  apply rt_step; left; do 4 right; left; constructor; assumption.
  constructor; apply Pe_proj; assumption.
  destruct Jrt as [Krt|Jrt].
  (* beta *)
  rename t into Z.
  inversion Krt as [x n U m]; subst Z; subst.
  rename t' into Z.
  inversion Het as [? ? T' Ht' HT| | | | | |? ? t' ? n' He Hen| | | | | | | | | | | | ]; subst Z; subst.
  exists ##; split.
  apply rt_refl; fail.
  apply Pe_Prop with T'; [ |eassumption].
  eapply subj_red_redn; eassumption.
  inversion He as [? ? T' Ht' HT| | | | | | | | | | | | | | | |? ? m' ? ? Hex| | ]; subst.
  admit (* missing rule: ## n' -> ## *).
  exists (m'[0:=n']); split.
  apply rt_step; left; do 5 right; left; constructor; fail.
  clear - Ht Hex Hen.
  InvTypHyp Ht; clear Ht; intros ? ? Hsign Hts Hus Happ HuT.
  inversion Hsign as [ | | |y ? ? a b| | | | | | | | ]; subst; clear Hsign.
  InvTypHyp Happ; clear Happ; intros A B s HA HB HAB Hpair.
  change (sg – e ⊢ A ** B <= (Forall y : a, b) ** a) in Hpair.
  assert (IA : sg – e ⊢ A <= Forall y : a, b) by (eapply inv_sum_l; eassumption).
  assert (IB : sg – e ⊢ B <= a) by (eapply inv_sum_r; eassumption).
  clear Hpair.
  inversion Hts as [ |? ? ? s' _ Hs| ]; subst.
  InvTypHyp Hs; clear Hs; intros ? ? ? Hfun Ha _ _.
  let x := eassumption in eapply Typ_conv in HA; [ |x|x].
  let x := eassumption in eapply Typ_conv in HB; [ |x|x].
  InvTypHyp HA; intros ? ? Hm HU Hle.
  assert (HaU : (e ⊦ a <= U)%pts) by exact (inv_prod_l Hle).
  clear HU; assert (HU := Hm); apply typ_wf in HU.
  inversion HU; subst.
  let x := eassumption in eapply Typ_conv in HB; [ |x|x].
  eapply Pe_sub with (d := mkdecl x U None); try eassumption; [ ].
  repeat constructor; fail.
  destruct Jrt as [Krt|Jrt].
  (* iota_fix *)
  inversion Krt as [n c cargs arg fx efix Hfx]; subst.
  rename t' into Z.
  inversion Het as [? ? T' Ht' HT| | | | | |? ? t' ? n' He Hen| | | | | | | | | | | | ]; subst Z; subst.
  exists ##; split.
  apply rt_refl; fail.
  apply Pe_Prop with T'; [ |eassumption].
  eapply subj_red_redn; eassumption.
  inversion He as [? ? T' Ht' HT| | | | | |? ? fx' ? arg' Hfx' Harg| | | | | | | | | | | | ]; subst t'; subst.
  admit (* missing rule: ## n' -> ## *).
  inversion Hfx' as [? ? T' Ht' HT| | | |? t t' Ie| | | | | | | | | | | | | | ]; subst fx'; subst.
  admit (* missing rule: ## n' -> ## *).
  assert (H20 : exists fx', t' = fx' .^ n /\ Pe e fx fx').
    (* inversion on Pe e (fx .^ n) t' *)
    clear - Ie.
    admit.
  destruct H20 as [fx' [? Hefx]]; subst t'.
  inversion Hfx as [ty fixs]; subst.
  inversion Hefx as [? ? T' Ht' HT| | | | | | | | | | | |? ? fixs' ? Hefixs| | | | | | ]; subst.
  admit (* missing rule: ## n' -> ## *).
  inversion Hen as [? ? T' Ht' HT| | | | | | | | | | | | | |? p p' a a' ? Hep Hea| | | | ]; subst n'; subst.
  exists ((((fixs' @ ##) @ Fix {(##,, fixs')} .^ n) ..L @ arg') @ ##); split.
  apply rt_step; do 5 right; left; repeat econstructor; fail.
  econstructor; [ |eassumption].
  econstructor; [ |eassumption].
  econstructor.
  eapply Pe_proj.
  econstructor; [ |eassumption].
  econstructor; [eassumption| ].
  constructor; fail.
  exists ((((fixs' @ Mt_mark) @ Fix {(##,, fixs')} .^ n) ..L @ arg') @ Cstr  {c // p',, a'}); split.
  apply rt_step; left; do 6 right; left; repeat econstructor; fail.
  repeat constructor; try eassumption; [ ].
  apply Pe_proj.
  repeat constructor; assumption.
  destruct Jrt as [Krt|Jrt].
  (* iota_case *)
  inversion Krt as [c pats n p a pred br b Hpatterns]; subst.
  inversion Het as [? ? T' Ht' HT| | | | | | | | | | | | |? ? c' pr ? br' ? ? Hec Hebr| | | | | ]; subst.
  exists ##; split.
  apply rt_refl; fail.
  apply Pe_Prop with T'; [ |eassumption].
  eapply subj_red_redn; eassumption.
  inversion Hec as [? ? T' Ht' HT| | | | | | | | | | | | | |? ? p' ? a' Hep Hea| | | | ]; subst.
  InvTypHyp Ht; clear Ht; intros v w Hsign Hv Hw Hcstr Hle.
  inversion Hsign as [ | | | | | | | | |inds is ? ? ? ? ? ? ? Hglob Hcase| | ]; subst; clear Hsign.
  inversion Hcase as [? ? ? ? tybr se Hfull Hb Hbranch]; subst; clear Hcase.
  destruct b.
  (* we are building a logical stuff by matching on a logical stuff *)
  admit (* missing rule: match ## -> ## *).
  (* we are building an informative stuff by matching on a logical stuff *)
  destruct Hb as [Hb|[Hsingle Hprop]].
  admit (* absurdity: c is a constructor of is, and is has 0 constructors *).
  assert (HH : n = 0 /\ pats = c::nil).
    (* because is has only one constructor *)
    admit.
  destruct HH; subst.
  exists (br'..L @ Mt_mark); split.
  apply rt_step; do 7 right; left; repeat econstructor; fail.
  econstructor.
  econstructor; eassumption.
  (* because the arguments of c are in Prop *)
  admit.
  exists ((br' .^ n)..L @ a'); split.
  apply rt_step; left; do 7 right; left; repeat econstructor; assumption.
  repeat econstructor; auto using Pe_proj; fail.
  destruct Jrt.
Qed.

End ExtractionProperties.

Hint Constructors Pe.
Hint Resolve @Pe_proj.
