(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Set Implicit Arguments.
Unset Strict Implicit.

(* theories*)
Require Export Bool.
Require Export Arith.
Require Export Peano_dec.
Require Export Compare_dec.
Require Export Wf_nat.
Require Export Relations.
Require Export Inclusion.
Require Export Transitive_Closure.
Require Export ExtractionInCoq.Library.Union.
Require Export Inverse_Image.
Require Export Lexicographic_Product.

(* tactics/contrib *)
Require Export Omega.

(* lib *)
Require Export MoreLogic.
Require Export MoreList.
Require Export Option.

Transparent sumor_rec.
Transparent sumbool_rec.
Transparent le_gt_dec.
Transparent le_lt_dec.
Transparent lt_eq_lt_dec.




Section MoreArith.

  Definition max_nat n m :=
    match le_gt_dec n m with
    | left _ => m
    | right _ => n
    end.

  Lemma least_upper_bound_max_nat :
   forall n m p, n <= p -> m <= p -> max_nat n m <= p.
Proof.
intros.
unfold max_nat in |- *.
elim (le_gt_dec n m); auto.
Qed.



  Lemma plus_minus_assoc : forall m n p, n <= m -> m - n + p = m + p - n.
Proof.
simple induction m; simpl in |- *.
intros.
replace n with 0; auto with arith.

simple destruct n0; simpl in |- *; auto with arith.
Qed.

  Lemma minus_pred : forall n m, n - S m = pred (n - m).
Proof.
simple induction n; intros.
simpl in |- *; auto.

case m; simpl in |- *; auto with arith.
Qed.

  Lemma minus_pred_pred : forall n m, pred n - m = pred (n - m).
Proof.
simple induction n.
simpl in |- *; auto.

simple destruct n0; simple destruct m; simpl in |- *; auto with arith.
Qed.

  Lemma minus_permute : forall m n p, m - n - p = m - p - n.
Proof.
simple induction m.
simpl in |- *; auto.

intros.
case n0; case p; simpl in |- *; intros; auto with arith.
repeat rewrite minus_pred.
rewrite H; auto.
Qed.

End MoreArith.


Section MoreWf.

  Variable A B C D : Set.
  Variable R2 : A * B -> A * B -> Prop.
  Variable R3 : A * (B * C) -> A * (B * C) -> Prop.
  Variable R4 : A * (B * (C * D)) -> A * (B * (C * D)) -> Prop.

  Lemma Acc2_rec :
   forall P : A -> B -> Set,
   (forall x x0,
    Acc R2 (x, x0) -> (forall y y0, R2 (y, y0) (x, x0) -> P y y0) -> P x x0) ->
   forall x x0, Acc R2 (x, x0) -> P x x0.
fix loop 5.
intros.
apply H; trivial.
intros.
apply loop.
exact H.

apply Acc_inv with (1 := H0) (2 := H1).
Qed.


  Lemma Acc3_rec :
   forall P : A -> B -> C -> Set,
   (forall x x0 x1,
    Acc R3 (x, (x0, x1)) ->
    (forall y y0 y1, R3 (y, (y0, y1)) (x, (x0, x1)) -> P y y0 y1) ->
    P x x0 x1) -> forall x x0 x1, Acc R3 (x, (x0, x1)) -> P x x0 x1.
fix loop 6.
intros.
apply H; trivial.
intros.
apply loop.
exact H.

apply Acc_inv with (1 := H0) (2 := H1).
Qed.


  Lemma Acc4_rec :
   forall P : A -> B -> C -> D -> Set,
   (forall x x0 x1 x2,
    Acc R4 (x, (x0, (x1, x2))) ->
    (forall y y0 y1 y2,
     R4 (y, (y0, (y1, y2))) (x, (x0, (x1, x2))) -> P y y0 y1 y2) ->
    P x x0 x1 x2) ->
   forall x x0 x1 x2, Acc R4 (x, (x0, (x1, x2))) -> P x x0 x1 x2.
fix loop 7.
intros.
apply H; trivial.
intros.
apply loop.
exact H.

apply Acc_inv with (1 := H0) (2 := H1).
Qed.

End MoreWf.


Section MoreRelations.

  Variable A : Type.
  Variable R : A -> A -> Prop.

  Definition in_dom_rel x := exists y : _, R x y.

  Lemma rst_preord : preorder _ (clos_refl_sym_trans _ R).
Proof.
intros.
split; auto.
exact (rst_trans _ R).
Qed.

  Lemma clos_refl_trans_ind_right :
   forall M (P : A -> Prop),
   P M ->
   (forall P0 N, R N P0 -> clos_refl_trans A R P0 M -> P P0 -> P N) ->
   forall a, clos_refl_trans A R a M -> P a.
intros.
generalize H H0.
elim H1; intros; auto.
apply H4 with y; auto.

apply H3; intros.
apply H5; intros; auto.
apply H7 with P0; auto.

apply H7 with P0; auto.
apply rt_trans with y; auto.
Qed.

Lemma step_clos_trans :
 forall x y,
 clos_refl_trans _ R x y ->
 forall z, clos_trans _ R y z -> clos_trans _ R x z.
simple induction 1; intros; auto.
apply t_trans with y0; auto.
Qed.

Lemma clos_trans_step :
 forall y z,
 clos_refl_trans _ R y z ->
 forall x, clos_trans _ R x y -> clos_trans _ R x z.
Proof.
simple induction 1; intros; auto.
apply t_trans with x; auto.
Qed.


  Fixpoint Rsteps n : relation A :=
    match n with
      | O => fun x y => x = y
      | S n => fun x z => exists y, R x y /\ Rsteps n y z
    end.

  Lemma Rsteps_clos_rt : forall n x y,
    Rsteps n x y -> clos_refl_trans A R x y.
  Proof.
    induction n; intros x y H; simpl in *.
    subst; auto.
    destruct H as (z, (Hz1, Hz2)).
    apply rt_trans with z; auto.
  Qed.

  Lemma Rsteps_clos_t : forall n x y,
    n >= 1 -> Rsteps n x y -> clos_trans A R x y.
  Proof.
    intros n x y Hn Hxy; destruct n.
    inversion Hn.
    simpl in *.
    destruct Hxy as (z, (Hxz, Hzy)).
    apply clos_trans_step with z.
    eapply Rsteps_clos_rt; eauto.
    eauto.
  Qed.

  Lemma Rsteps_one : forall x y, R x y -> Rsteps 1 x y.
  Proof.
    intros; simpl; eauto.
  Qed.

  Lemma Rsteps_trans : forall m n x y z,
    Rsteps m x y -> Rsteps n y z -> Rsteps (m+n) x z.
  Proof.
    induction m; intros n x y z Hxy Hyz; simpl in *; subst; auto.
    destruct Hxy as (x', (Hxx', Hx'y)).
    exists x'; split; eauto.
  Qed.

  Lemma rt_Rsteps : forall x y,
    clos_refl_trans _ R x y -> exists n, Rsteps n x y.
  Proof.
    induction 1.
    exists 1; simpl; eauto.
    exists 0; simpl; eauto.
    destruct IHclos_refl_trans1 as (n1, Hxy).
    destruct IHclos_refl_trans2 as (n2, Hyz).
    exists (n1+n2); eapply Rsteps_trans; eauto.
  Qed.

  Lemma t_Rsteps : forall x y,
    clos_trans _ R x y -> exists n, n >= 1 /\ Rsteps n x y.
  Proof.
    induction 1.
    exists 1; split; simpl; eauto.
    destruct IHclos_trans1 as (n1, (Hn1, Hxy)).
    destruct IHclos_trans2 as (n2, (Hn2, Hyz)).
    exists (n1+n2); split; simpl; auto with arith.
    eapply Rsteps_trans; eauto.
  Qed.

End MoreRelations.

(*
Section NonEmptySet.

  Record ne_set: Type := {
    Carrier: Set;
    Elt: Carrier
  }.

End NonEmptySet.
*)


  Hint Resolve left_sym right_sym sp_swap sp_noswap: core.

