(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Import Arith.
Require Import String.
Require Import MoreList.
Require Import MlInteger.

Open Scope string_scope.

Set Implicit Arguments.
Unset Strict Implicit.

Unset Elimination Schemes.

Record ml_string := mk_str {
  string_value : String.string
}.

Definition nil_string := mk_str "".
Definition dummy_string := mk_str "_".

Section string_of_int.

  Import Ascii.
  Import BinPos BinNat.

  Fixpoint string_of_positive x :=
    match x with
      | xI p => String "1"%char (string_of_positive p)
      | xO p => String "0"%char (string_of_positive p)
      | xH => EmptyString
    end.

  Definition string_of_N x :=
    match x with
      | N0 => "0"
      | Npos p => String "1"%char (string_of_positive p)
    end.

  Definition string_of_int x := mk_str (string_of_N (int_value x)).

End string_of_int.

Definition app_str x y := mk_str (append (string_value x) (string_value y)).

Definition eq_string : forall s1 s2 : ml_string, {s1 = s2} + {s1 <> s2}.
Proof.
  intros [s1] [s2].
  refine (
    match string_dec s1 s2 with
      | left H => left _ _
      | right H => right _ _
    end
  ); congruence.
Defined.

Lemma app_str_inj_r : forall x s1 s2, app_str x s1 = app_str x s2 -> s1 = s2.
Proof.
  unfold app_str; intros [x] [s1] [s2] Heq; simpl in *; f_equal.
  injection Heq; clear Heq; revert s1 s2.
  induction x; simpl; intros s1 s2 Heq.
  assumption.
  injection Heq; clear Heq; intro Heq.
  auto; fail.
Qed.

Lemma app_nil_neutral_r : forall s, app_str s nil_string = s.
Proof.
  unfold app_str, nil_string.
  intros [s]; simpl in *; f_equal.
  induction s; simpl.
  reflexivity.
  f_equal; assumption.
Qed.

Lemma string_of_int_inj :
  forall i1 i2, string_of_int i1 = string_of_int i2 -> i1 = i2.
Proof.
  unfold string_of_int.
  intros [i1] [i2] Heq; simpl in *.
  f_equal; injection Heq; clear Heq; intro Heq.
  destruct i1 as [ |i1]; destruct i2 as [ |i2]; simpl in *.
  reflexivity.
  discriminate.
  discriminate.
  f_equal; injection Heq; clear Heq; intro Heq.
  revert i2 Heq.
  induction i1 as [i1|i1| ]; destruct i2 as [i2|i2| ]; intro Heq; simpl in *; try discriminate.
  f_equal; apply IHi1; congruence.
  f_equal; apply IHi1; congruence.
  reflexivity.
Qed.

Lemma string_of_int_not_nil : forall i, string_of_int i <> nil_string.
Proof.
  unfold string_of_int, nil_string, string_of_N.
  intros [i] Ha; simpl in *.
  injection Ha; clear Ha; intro Ha.
  destruct i; discriminate.
Qed.

(* printing string... *)
Definition print_endline (x : ml_string) := tt.

  (* Proving the set of strings is infinite *)
  Definition var_nat x i :=
    match i with
      | O => x
      | S k => app_str x (string_of_int (int_of_nat k))
    end.

  Lemma inj_var_nat : forall x m n, var_nat x m = var_nat x n -> m = n.
  Proof.
    intros x m [| n]; simpl in |- *; intros.
    generalize H.
    case m; trivial; simpl in |- *; intros.
    generalize H0.
    pattern x at 2 in |- *.
    elim app_nil_neutral_r with (s := x).
    intros.
    elim string_of_int_not_nil with (1 := app_str_inj_r H1).

    generalize H.
    case m; simpl in |- *; intros.
    generalize H0.
    pattern x at 1 in |- *.
    elim app_nil_neutral_r with (s := x).
    intros.
    elim string_of_int_not_nil with (1 := app_str_inj_r (sym_equal H1)).

    generalize (int_injection (string_of_int_inj (app_str_inj_r H0))).
    auto with arith.
  Qed.

  (* list of strings *)
  Definition ml_strings := list ml_string.

  Inductive ord_insert : ml_strings -> ml_strings -> Prop :=
    oi_intro : forall x n l1 l2, insert x n l1 l2 -> ord_insert l1 l2.


  Lemma wf_oi : well_founded ord_insert.
  Proof.
    cut (forall n l, length l = n -> Acc ord_insert l).
    red in |- *; intros.
    apply H with (length a); auto.

    simple induction n.
    simple destruct l; intros.
    apply Acc_intro; intros.
    inversion_clear H0.
    inversion_clear H1.

    discriminate H.

    simple destruct l; simpl in |- *; intros.
    discriminate H0.

    injection H0; intros.
    apply Acc_intro; intros.
    inversion_clear H2.
    apply H.
    cut (S (length y) = length (m :: l0)); intros.
    simpl in H2.
    injection H2; auto.
    elim H1; auto.

    elim H3; auto.
    intros.
    simpl in |- *.
    elim H4; auto.
  Qed.


  Lemma rmv :
    forall x l,
      {l1 : _ | exists n : _, insert (x:ml_string) n l1 l} + {~ In x l}.
  Proof.
    fix rmv 2.
    intros x [| y l1].
    right; red in |- *; intro in_nil; inversion in_nil.

    case (eq_string x y).
    intros eq_x_y.
    left; exists l1; rewrite eq_x_y; exists 0; repeat constructor.

    intros diff_x_y.
    case (rmv x l1).
    intros (v, rmv_v).
    left; exists (y :: v); case rmv_v; intros n ins_n; exists (S n);
      repeat constructor; trivial.

    intros not_found; red in not_found.
    right; red in |- *; intros not_in_l; generalize diff_x_y;
      inversion_clear not_in_l; auto.
  Qed.

  (*Realizer Fix rmv {rmv [x:?; l:?]: ? :=
    Cases l of
       nil => (inright ?)
     | (cons y l1) => Cases (eq_string x y) of
           left => (inleft ? l1)
         | right => Cases (rmv x l1) of
                      (inleft v) => (inleft ? (cons ? y v))
                    | inright => (inright ?)
                    end
         end
     end}.
  *)


  Lemma find_free : forall s l n, {m : _ | n <= m &  ~ In (var_nat s m) l}.
  Proof.
    intros s l.
    elim l using well_founded_induction with (1 := wf_oi).
    clear l; intros l ff_rec n.
    case (rmv (var_nat s n) l).
    intros (l1, rmv_n).
    case ff_rec with l1 (S n).
    case rmv_n; intros m ins_m; apply oi_intro with (var_nat s n) m; auto.

    intros m lt_m m_free.
    exists m; auto with arith.
    case rmv_n; intros n' ins_n'.
    red in |- *; intros in_l; apply m_free.
    assert (m_in_ext := insert_in_inv ins_n' in_l).
    inversion m_in_ext; auto.
    rename H into eq_vn_vm.
    absurd (m = n).
    red in |- *; intros; omega.
    symmetry; apply inj_var_nat with (1 := eq_vn_vm).

    intros not_in_l; exists n; auto.
  Qed.
  (*Realizer [s]<nat->nat>rec ffv :: :: { ord_insert }
    [l:ml_strings][n]Cases (rmv (var_nat s n) l) of
        (inleft l1) => (ffv l1 (S n))
      | inright => n
      end.
  *)

  Lemma find_free_var :
    ml_string -> forall l : ml_strings, {x : _ | ~ In x l}.
  Proof.
    (*
    Realizer [s,l](var_nat s (find_free s l O)).
    *)
    intros s l.
    elim (find_free s l 0); intros.
    split with (var_nat s x); trivial.
  Qed.
