(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export PTS_spec.

Module Type IMetatheory
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes).

  Include IPTS_spec P MTermes MEnv.


Section IMetatheory.

  Context {MPTSO_spec : PTSO_spec}.

  Definition le_type := Rule subrule.
  Definition ge_type e := transp _ (le_type e).
  Definition le_sort s1 s2 := forall e, le_type e %%s1 %%s2.
  Definition le_env (e f : env) := rel_in_list (rel_typ le_type) e f.
  Hint Unfold le_sort: core.

  Definition le_type_lift := Rule_lifts subrule.
  Definition le_type_subst := Rule_substs subrule.
  Definition le_type_stable := Rule_stable subrule.

  Definition le_type_refl e : forall t, le_type e t t :=
    preord_refl _ _ (preord subrule e).
  Definition le_type_trans e :
    forall x y z, le_type e x y -> le_type e y z -> le_type e x z :=
      preord_trans _ _ (preord subrule e).
  Hint Resolve le_type_refl: core.


  (* Typing judgements: *)

  Inductive typ : env -> term -> term -> Prop :=
    | Typ_srt : forall e s1 s2, wf e -> axiom s1 s2 -> typ e %%s1 %%s2
    | Typ_rel : forall e v t, wf e -> type_of_rel e v t -> typ e #v t
    | Typ_cst0 :
        forall e c u,
        wf e -> wf (nil; x_ : u) -> in_sign0 c u -> typ e (Cst0 c) u
    | Typ_cst1 :
        forall e c m t u,
        wf (e; x_ : u) ->
        wf (e; x_ : t) ->
        typ e m t -> in_sign1 c e m t u -> typ e (Cst1 c m) u
    | Typ_lam :
        forall e x a t u s,
        typ (e; x : a) t u ->
        typ e (Forall x : a, u) %%s ->
        typ e (Fun x : a => t) (Forall x : a, u)
    | Typ_prd :
        forall e x a b s1 s2 s3,
        typ e a %%s1 ->
        typ (e; x : a) b %%s2 ->
        rule s1 s2 s3 -> typ e (Forall x : a, b) %%s3
    | Typ_sum :
        forall e a b s1 s2 s3,
        typ e a %%s1 -> typ e b %%s2 -> prd s1 s2 s3 -> typ e (a ** b) %%s3
    | Typ_pair :
        forall e m t a b s,
        typ e m a ->
        typ e t b -> typ e (a ** b) %%s -> typ e (m,,t) (a ** b)
    | Typ_conv :
        forall e t U V s,
        typ e t U -> le_type e U V -> typ e V %%s -> typ e t V
    | Typ_conv_srt :
        forall e t U s, typ e t U -> le_type e U %%s -> typ e t %%s
  with wf : env -> Prop :=
    | Wf_nil : wf nil
    | Wf_cons_var : forall e x T s, wf e -> typ e T %%s -> wf (e; x : T)
    | Wf_cons_def :
        forall e x M T s,
        wf e -> typ e M T -> typ e T %%s -> wf (e; x := M : T).

  Hint Resolve Typ_srt Typ_rel Typ_cst0 Wf_nil: core.

  (* Syntax for context and judgements *)

  Notation "e |- m : t" := (typ e m t) (at level 50, m at next level).
  Notation "|-  e" := (wf e) (at level 50).
  Notation "e |- t <= u" := (le_type e t u) (at level 50, t at next level).


  Scheme typ_mind := Minimality for typ Sort Prop
  with wf_mind := Minimality for wf Sort Prop.


  (* Inversion rules *)

  Theorem typ_wf : forall e t T, e |- t : T -> |- e.
  Proof.
    simple induction 1; intros; auto.
  Qed.

  Lemma inv_wf : forall (e : env) d, |- (e; d) -> |- e.
  Proof.
    intros.
    inversion_clear H; auto.
  Qed.

  Lemma inv_wf_sort :
   forall (e : env) d, |- (e; d) -> exists s : _, e |- TYPE d : %%s.
  Proof.
    intros.
    inversion_clear H; exists s; auto.
  Qed.


  (* Generation lemmas *)

  Definition inv_pred_wf e (P : Prop) :=
    match e with
    | nil => P
    | mkdecl x t None :: f => forall s, f |- t : %%s -> P
    | mkdecl x t (Some m) :: f => forall s, f |- m : t -> f |- t : %%s -> P
    end.

  Lemma inversion_wf : forall P e, inv_pred_wf e P -> |- e -> P.
  Proof.
    intros P e invwf j.
    generalize P invwf; clear invwf P.
    elim j; simpl in |- *; intros; eauto.
  Qed.

  Ltac InvWfHyp H :=
    unfold add_var, add_def in H; inversion H using inversion_wf;
      unfold inv_pred_wf in |- *.


  Definition inv_pred_typ e M T (P : Prop) :=
    match M with
    | %%s1 => forall s2, axiom s1 s2 -> e |- %%s2 <= T -> P
    | #n => forall d, item_rel e n d -> e |- ^(S n)(TYPE d) <= T -> P
    | Cst0 c =>
        forall u, |- (nil; x_ : u) -> in_sign0 c u -> e |- u <= T -> P
    | Cst1 c m =>
        forall t u,
        in_sign1 c e m t u ->
        |- (e; x_ : t) -> |- (e; x_ : u) -> e |- m : t -> e |- u <= T -> P
    | Fun x : a => t =>
        forall u s,
        (e; x : a) |- t : u ->
        e |- Forall x : a, u : %%s -> e |- Forall x : a, u <= T -> P
    | Forall x : a, b =>
        forall s1 s2 s3,
        e |- a : %%s1 ->
        (e; x : a) |- b : %%s2 -> rule s1 s2 s3 -> e |- %%s3 <= T -> P
    | a ** b =>
        forall s1 s2 s3,
        e |- a : %%s1 ->
        e |- b : %%s2 -> prd s1 s2 s3 -> e |- %%s3 <= T -> P
    | (m,,t) =>
        forall a b s,
        e |- m : a ->
        e |- t : b -> e |- a ** b : %%s -> e |- a ** b <= T -> P
    end.


  Theorem inv_pred_typ_conv :
   forall e M U V,
   e |- U <= V -> forall P, inv_pred_typ e M V P -> inv_pred_typ e M U P.
  Proof.
    intros e M U V le_UV.
    cut (forall x, e |- x <= U -> e |- x <= V).
    case M; simpl in |- *; intros.
    apply H0 with s2; auto.

    apply H0 with d; auto.

    apply H0 with u; auto.

    apply H0 with t0 u; auto.

    apply H0 with s1 s2 s3; auto.

    apply H0 with u s; auto.

    apply H0 with s1 s2 s3; auto.

    apply H0 with a b s; auto.

    intros.
    apply le_type_trans with U; auto.
  Qed.


  Theorem inversion_lemma :
    forall P e M T, inv_pred_typ e M T P -> e |- M : T -> P.
  Proof.
    intros P e M T inv_P j.
    generalize P inv_P; clear inv_P P.
    elim j; simpl in |- *; intros.
    apply inv_P with s2; trivial.

    inversion_clear H0.
    apply inv_P with x; trivial.
    elim H1; trivial.

    apply inv_P with u; trivial.

    apply inv_P with t u; trivial.

    apply inv_P with u s; auto.

    apply inv_P with s1 s2 s3; auto.

    apply inv_P with s1 s2 s3; auto.

    apply inv_P with a b s; auto.

    apply H0.
    apply inv_pred_typ_conv with V; auto.

    apply H0.
    apply inv_pred_typ_conv with (%%s); auto.
  Qed.


  Ltac InvTypHyp H :=
    inversion H using inversion_lemma; unfold inv_pred_typ in |- *.


  Ltac Inversion_typ T :=
    generalize T; intro INVTYP; InvTypHyp INVTYP || InvWfHyp INVTYP;
     clear INVTYP; intros.


  Theorem wf_sort :
    forall e n d,
      item_rel e n d ->
      |- e -> forall f, trunc (S n) e f -> exists s : _, f |- TYPE d : %%s.
  Proof.
    simple induction 1; intros.
    apply inv_wf_sort.
    replace f with l; trivial.
    inversion_clear H1.
    inversion_clear H2; trivial.

    inversion_clear H3.
    apply H1; trivial.
    apply (inv_wf _ _ H2).
  Qed.

  Theorem wf_definition :
    forall n e x m t,
      item_rel e n (mkdecl x t (Some m)) ->
      |- e -> forall f, trunc (S n) e f -> f |- m : t.
  Proof.
    simple induction 1; intros.
    Inversion_typ H0.
    replace f with l; trivial.
    inversion_clear H1.
    inversion_clear H4; trivial.

    inversion_clear H3.
    apply H1; trivial.
    apply (inv_wf _ _ H2).
  Qed.




  (* Free variables lemma *)
  Lemma typ_clos : forall e m t, e |- m : t -> closed (length e) m.
  Proof.
    simple induction 1; intros; auto.
    constructor.
    inversion_clear H1.
    elim H3; simpl in |- *; intros; auto with arith.

    constructor; trivial.
    inversion_clear H3; trivial.
  Qed.

  Lemma ins_clos :
    forall e f d, |- f -> ins_in_env e d (length f) (f ++ e) (f ++ d :: e).
  Proof.
    simple induction 1; simpl in |- *; intros; auto.
    pattern (mkdecl x T None) at 2 in |- *.
    replace (mkdecl x T None) with (liftd_rec 1 (mkdecl x T None) (length e0));
      auto.
    unfold liftd_rec in |- *.
    simpl in |- *.
    rewrite closed_lift_rec; trivial.
    apply typ_clos with (1 := H2).

    pattern (mkdecl x T (Some M)) at 2 in |- *.
    replace (mkdecl x T (Some M)) with
      (liftd_rec 1 (mkdecl x T (Some M)) (length e0)); auto.
    unfold liftd_rec in |- *.
    simpl in |- *.
    rewrite closed_lift_rec.
    rewrite closed_lift_rec; trivial.
    apply typ_clos with (1 := H2).

    apply typ_clos with (1 := H3).
  Qed.



  (* Thinning lemma *)
  Theorem typ_weak :
    forall (g : env) d e t T,
      |- (g; d) ->
        e |- t : T -> forall f k, ins_in_env g d k e f -> f |- ^(1//k)t : ^(1//k)T.
  Proof.
    intros g d e t T wfgd j.
    elim j using
      typ_mind with (P0 := fun e => forall f k, ins_in_env g d k e f -> |- f);
      intros.
    simpl in |- *.
    constructor; auto.
    apply H0 with k; auto.

    apply lift_split; intros; constructor; auto.
    apply H0 with k; auto.

    apply ins_type_of_rel_lt with d e0 g; auto.

    apply H0 with k; auto.

    simpl in |- *.
    inversion_clear H1.
    split with x.
    rewrite H4.
    unfold lift in |- *.
    rewrite simpl_lift_rec; simpl in |- *; auto with arith.

    apply ins_item_ge with (1 := H2); auto.

    simpl in |- *; rewrite closed_lift_rec.
    apply Typ_cst0; trivial.
    apply H0 with k; trivial.

    apply closed_weak with 0; auto with arith.
    Inversion_typ H1.
    apply typ_clos with (1 := H5).

    simpl in |- *.
    apply Typ_cst1 with (^(1//k)t0); eauto.
    apply in_sign_lift with (1 := H6); trivial.

    simpl in |- *.
    apply Typ_lam with s; auto.
    exact (H2 _ _ H3).

    simpl in |- *.
    simpl in H0, H2.
    apply Typ_prd with s1 s2; auto.

    simpl in |- *.
    simpl in H0, H2.
    apply Typ_sum with s1 s2; auto.

    simpl in |- *.
    simpl in H4.
    apply Typ_pair with s; auto.

    simpl in H3.
    apply Typ_conv with (^(1//k)U) s; auto.
    apply le_type_lift with g e0 d; auto.

    simpl in |- *.
    apply Typ_conv_srt with (^(1//k)U); auto.
    fold (^(1//k)%%s) in |- *.
    apply le_type_lift with g e0 d; auto.

    generalize wfgd.
    inversion_clear H; auto.

    simpl in H2.
    unfold add_var in H3.
    generalize wfgd.
    inversion_clear H3; simpl in |- *; intros; auto.
    unfold liftd_rec in |- *; simpl in |- *.
    fold (g0; x : ^(1//n)T0) in |- *.
    apply Wf_cons_var with s; auto.
    apply H0 with (1 := H4).

    simpl in H4.
    unfold add_def in H5.
    generalize wfgd.
    inversion_clear H5; simpl in |- *; intros; auto.
    unfold liftd_rec in |- *; simpl in |- *.
    fold (g0; x := ^(1//n)M : ^(1//n)T0) in |- *.
    apply Wf_cons_def with s; auto.
    apply H0 with (1 := H6).
  Qed.


  Theorem thinning :
    forall (e f : env) n t T,
      trunc n f e -> e |- t : T -> |- f -> f |- ^(n)t : ^(n)T.
  Proof.
    simple induction 1; intros.
    repeat rewrite lift0; auto.

    rewrite simpl_lift.
    pattern (^(S k)T) in |- *.
    rewrite simpl_lift.
    unfold lift at 1 3 in |- *.
    apply typ_weak with e0 x e0; auto.
    apply H1; auto.
    apply inv_wf with x; trivial.
  Qed.


  (* Substitution lemma *)

  Theorem typ_sub :
    forall e g t d u U,
      g |- t : TYPE d ->
        e |- u : U -> forall n f, sub_in_env g t d n e f -> f |- u[n:=t] : U[n:=t].
  Proof.
    intros e g t d u U jt j.
    elim j using
      typ_mind with (P0 := fun e => forall n f, sub_in_env g t d n e f -> |- f);
      intros.
    simpl in |- *; eauto.

    cut (|- f); intros; eauto.
    generalize (refl_equal v).
    pattern v at 1, (#v[n:=t]) in |- *.
    apply subst_split; intros.
    apply Typ_rel; auto.
    apply nth_sub_item_inf with e0 g d; auto.

    inversion_clear H1.
    rewrite H5.
    elim H4.
    rewrite simpl_subst; auto.
    apply thinning with g; auto.
    apply sub_env_trunc with (1 := H2).

    elim inj_item_rel with (d1 := d) (2 := H6); auto.
    elim H4.
    apply sub_decl_eq_def with (1 := H2).

    apply Typ_rel; auto.
    inversion_clear H1.
    rewrite H6.
    elim H5.
    rewrite simpl_subst; auto.
    split with x; auto.
    apply nth_sub_sup with (1 := H2); auto.
    rewrite H5; trivial.

    simpl in |- *; rewrite closed_subst_rec.
    apply Typ_cst0; trivial.
    apply H0 with n; trivial.

    apply closed_weak with 0; auto with arith.
    Inversion_typ H1.
    apply typ_clos with (1 := H5).

    simpl in |- *.
    apply Typ_cst1 with (t0[n:=t]); eauto.
    apply in_sign_subst with (1 := H6); trivial.

    simpl in |- *.
    simpl in H2.
    apply Typ_lam with s; auto.

    simpl in |- *.
    simpl in H0, H2.
    apply Typ_prd with s1 s2; auto.

    simpl in |- *.
    simpl in H0, H2.
    apply Typ_sum with s1 s2; auto.

    simpl in |- *.
    simpl in H4.
    apply Typ_pair with s; auto.

    simpl in H3.
    apply Typ_conv with (U0[n:=t]) s; auto.
    apply le_type_subst with (2 := H4); auto.

    simpl in |- *.
    apply Typ_conv_srt with (U0[n:=t]); auto.
    fold (%%s[n:=t]) in |- *.
    apply le_type_subst with (2 := H2); auto.

    generalize jt.
    inversion_clear H; auto.

    simpl in H2.
    unfold add_var in H3.
    generalize H.
    inversion_clear H3; intros; auto.
    unfold substd_rec in |- *; simpl in |- *.
    fold (g0; x : T[n0:=t]) in |- *.
    apply Wf_cons_var with s; auto.
    apply H0 with (1 := H4).

    simpl in H4.
    unfold add_def in H5.
    generalize H.
    inversion_clear H5; simpl in |- *; intros; auto.
    unfold substd_rec in |- *; simpl in |- *.
    fold (g0; x := M[n0:=t] : T[n0:=t]) in |- *.
    apply Wf_cons_def with s; auto.
    apply H0 with (1 := H6).
  Qed.


  Theorem substitution :
    forall e x d t u U,
      (e; x : t) |- u : U -> e |- d : t -> e |- u[0:=d] : U[0:=d].
  Proof.
    intros.
    apply typ_sub with (e; x : t) e (mkdecl x t None); auto.
    unfold add_var in |- *; auto.
    constructor; compute in |- *; auto.
  Qed.

  Lemma substitution1 :
    forall e x y d t m u U,
      ((e; x : t); y : m) |- u : U ->
        e |- d : t -> (e; y : m[0:=d]) |- u[1:=d] : U[1:=d].
  Proof.
    intros.
    apply typ_sub with ((e; x : t); y : m) e (mkdecl x t None); auto.
    apply sub_in_env_var.
    unfold add_var in |- *.
    constructor; compute in |- *; auto.
  Qed.

  Lemma sub_in_env_wf: forall g e f d t i,
    g |- t : TYPE d -> |- e ->
      sub_in_env g t d i e f ->
      |- f.
  Proof.
    induction 3.
    eapply typ_wf; eauto.
    unfold substd_rec; simpl.
    destruct u as (x, a, [b|]); inversion H0; subst.
    change (|- (g0; x := b[n:=t] : a[n:=t])).
      econstructor.
      apply IHsub_in_env; auto.
      eapply typ_sub; eauto.
      instantiate (1:=s).
      change (g0 |- a[n:=t] : %%s[n:=t]).
      eapply typ_sub; eauto.
    change (|- (g0; x : a[n:=t])).
      econstructor.
      apply IHsub_in_env; auto.
      instantiate (1:=s).
      change (g0 |- a[n:=t] : %%s[n:=t]).
      eapply typ_sub; eauto.
  Qed.


  (* Well-formed types: sorts and terms typed by a sort *)
  Inductive wf_type (e : env) : term -> Prop :=
    | Wft_typed : forall T s, e |- T : %%s -> wf_type e T
    | Wft_top : forall s, wf_type e %%s.

  Hint Resolve Wft_top: core.

  Lemma var_wft : forall e t, |- (e; x_ : t) -> wf_type e t.
  Proof.
    intros.
    Inversion_typ H.
    left with s; trivial.
  Qed.


  (* Signature closed under subtyping *)
  Definition in_sign_sat c e m t u :=
    exists2 t' : _,
      e |- t <= t' /\ |- (e; x_ : t') & in_sign1 c e m t' u /\ |- (e; x_ : u).
(*
  Lemma in_sign_inj: (c:oper)(e:env)(x:name)(m,t,u:term)
      [~ e |- m : t ~]->[~ |- e; x_:u ~]
        ->(in_sign1 c e m t u)->(in_sign_sat c e m t u).
Proof.
Intros.
Split with t; Auto.
Save.
*)

  (* Completion of the PTS under subtyping *)
  Definition complete (rl : sort -> sort -> sort -> Prop) s1 s2 s3 :=
    exists2 x1 : _,
      le_sort s1 x1 & (exists2 x2 : _, le_sort s2 x2 & rl x1 x2 s3).

  Lemma inject_sat :
    forall (R : sort -> sort -> sort -> Prop) s1 s2 s3,
      R s1 s2 s3 -> complete R s1 s2 s3.
  Proof.
    red in |- *; intros.
    split with s1; auto.
    split with s2; auto.
  Qed.

  Hint Resolve inject_sat: core.

  Definition rl_sat := complete rule.
  Definition pr_sat := complete prd.
  Hint Unfold rl_sat pr_sat: core.


  (* Introduction rules *)
  Lemma wf_var : forall e x t s, e |- t : %%s -> |- (e; x : t).
  Proof.
    intros.
    apply Wf_cons_var with s; auto.
    apply typ_wf with (1 := H).
  Qed.

  Lemma wf_def :
    forall e x m t s, e |- m : t -> e |- t : %%s -> |- (e; x := m : t).
  Proof.
    intros.
    apply Wf_cons_def with s; auto.
    apply typ_wf with (1 := H).
  Qed.

  Lemma inst_var :
    forall e x m t, |- (e; x : t) -> e |- m : t -> |- (e; x := m : t).
  Proof.
    intros.
    Inversion_typ H.
    apply wf_def with s; trivial.
  Qed.

  Lemma forget_def : forall e x m t, |- (e; x := m : t) -> |- (e; x : t).
  Proof.
    intros.
    Inversion_typ H.
    apply wf_var with s; trivial.
  Qed.



  Theorem typ_conv_wf :
    forall e t T U, e |- t : T -> wf_type e U -> e |- T <= U -> e |- t : U.
  Proof.
    simple induction 2; intros.
    apply Typ_conv with T s; auto.

    apply Typ_conv_srt with T; auto.
  Qed.



  Lemma typ_cst1 :
    forall e c m t u,
      in_sign1 c e m t u ->
      |- (e; x_ := m : t) -> |- (e; x_ : u) -> e |- Cst1 c m : u.
  Proof.
    intros.
    Inversion_typ H0.
    apply Typ_cst1 with t; trivial.
    apply wf_var with s; trivial.
  Qed.

  Lemma typ_cst1_sat :
    forall e c m t u, in_sign_sat c e m t u -> e |- m : t -> e |- Cst1 c m : u.
  Proof.
    intros e c m t u (t', (lett', wft'), (ins, wfu)) j.
    apply Typ_cst1 with t'; trivial.
    apply typ_conv_wf with t; auto.
    apply var_wft; trivial.
  Qed.


  Lemma Typ_prd' :
    forall e x a b s1 s2 s3,
      e |- a : %%s1 ->
        (e |- a : %%s1 -> (e; x : a) |- b : %%s2) ->
        rule s1 s2 s3 -> e |- Forall x : a, b : %%s3.
  Proof.
    intros.
    apply Typ_prd with s1 s2; auto.
  Qed.

  Lemma typ_prd :
    forall e x a b s1 s2 s3,
      e |- a : %%s1 ->
        (e; x : a) |- b : %%s2 -> rl_sat s1 s2 s3 -> e |- Forall x : a, b : %%s3.
  Proof.
    intros.
    red in H1.
    red in H1.
    decompose record H1.
    apply Typ_prd with x0 x1; trivial.
    apply Typ_conv_srt with (%%s1); auto.

    apply Typ_conv_srt with (%%s2); auto.
  Qed.


  Lemma typ_sum :
    forall e a b s1 s2 s3,
      e |- a : %%s1 -> e |- b : %%s2 -> pr_sat s1 s2 s3 -> e |- a ** b : %%s3.
  Proof.
    intros e a b s1 s2 s3 ja jb (s1', le1, (s2', le2, pr)).
    apply Typ_sum with s1' s2'; trivial.
    apply Typ_conv_srt with (%%s1); auto.

    apply Typ_conv_srt with (%%s2); auto.
  Qed.


  Lemma inst_pair :
    forall e u v a b,
      e |- u : a -> e |- v : b ->
        |- (e; x_ : a ** b) -> |- (e; x_ := (u,,v) : a ** b).
  Proof.
    intros.
    Inversion_typ H1.
    apply wf_def with s; auto.
    apply Typ_pair with s; auto.
  Qed.

  Lemma inst_triple :
    forall e u v w a b c,
      e |- u : a -> e |- v : b -> e |- w : c ->
        |- (e; x_ : a ** (b ** c)) -> |- (e; x_ := (u,,(v,,w)) : a ** (b ** c)).
  Proof.
    intros.
    apply inst_pair; auto.
    Inversion_typ H2.
    Inversion_typ H3.
    apply Typ_pair with s2; auto.
  Qed.


  (* Other corollaries of the thinning lemma *)

  Theorem wf_type_thinning :
    forall e f n t, wf_type e t -> trunc n f e -> |- f -> wf_type f ^(n)t.
  Proof.
    simple destruct 1; intros.
    constructor 1 with s.
    fold (^(n//0)%%s) (^(n)%%s) in |- *.
    apply thinning with e; auto.

    unfold lift, lift_rec in |- *.
    auto.
  Qed.


  Lemma wf_var_lift :
    forall e x y t u,
      |- (e; x : t) -> |- (e; y : u) -> |- ((e; x : t); y : ^(1)u).
  Proof.
    intros.
    generalize wf_var.
    Inversion_typ H0.
    apply H2 with s.
    replace (%%s) with (^(1)%%s); auto.
    apply thinning with e; auto.
  Qed.







  Lemma typ_thin_app_wk :
    forall e f m t,
      |- e -> f |- m : t -> (f ++ e) |- m : ^(length e//length f)t.
  Proof.
    simple induction 1; intros.
    rewrite <- app_nil_end.
    rewrite lift_rec0.
    trivial.

    simpl in |- *.
    fold (1 + length e0) in |- *.
    rewrite <- simpl_lift_rec with (i := length f); auto with arith.
    replace m with (^(1//length f)m).
    apply typ_weak with e0 (mkdecl x T None) (f ++ e0); auto.
    fold (e0; x : T) in |- *.
    apply wf_var with s; trivial.

    unfold add_var in |- *.
    apply ins_clos.
    apply (typ_wf _ _ _ H3).

    apply closed_lift_rec.
    apply typ_clos with (1 := H3).

    simpl in |- *.
    fold (1 + length e0) in |- *.
    rewrite <- simpl_lift_rec with (i := length f); auto with arith.
    replace m with (^(1//length f)m).
    apply typ_weak with e0 (mkdecl x T (Some M)) (f ++ e0); auto.
    fold (e0; x := M : T) in |- *.
    apply wf_def with s; trivial.

    unfold add_def in |- *.
    apply ins_clos.
    apply (typ_wf _ _ _ H4).

    apply closed_lift_rec.
    apply typ_clos with (1 := H4).
  Qed.


  Lemma wf_thin_app : forall e f, |- e -> |- f -> |- (f ++ e).
  Proof.
    simple destruct 2; intros.
    simpl in |- *; trivial.

    simpl in |- *.
    fold (e0 ++ e; x : T) in |- *.
    apply wf_var with s; auto.
    fold (^(length e//length e0)%%s) in |- *.
    apply typ_thin_app_wk; trivial.

    simpl in |- *.
    fold (e0 ++ e; x := M : T) in |- *.
    apply wf_def with s.
    replace T with (^(length e//length e0)T).
    apply typ_thin_app_wk; trivial.

    apply closed_lift_rec.
    apply typ_clos with (1 := H3).

    fold (^(length e//length e0)%%s) in |- *.
    apply typ_thin_app_wk; trivial.
  Qed.

  (* 3 applications of wf_thin_app *)
  Lemma wf_thin1 : forall e x t, |- (( ); x : t) -> |- e -> |- (e; x : t).
  Proof.
    intros.
    apply (wf_thin_app _ _ H0 H).
  Qed.

  Lemma thin_par_wf : forall e d, |- (( ); d) -> |- e -> |- (e; d).
  Proof.
    intros.
    apply (wf_thin_app _ _ H0 H).
  Qed.

  Lemma thin_par_wf2 :
    forall e x y t u,
      |- ((( ); x : t); y : u) -> |- e -> |- ((e; x : t); y : u).
  Proof.
    intros.
    apply (wf_thin_app _ _ H0 H).
  Qed.




  (* Other corollaries of the substitution lemma *)

  Theorem substitution_def :
    forall e x d t u U, (e; x := d : t) |- u : U -> e |- u[0:=d] : U[0:=d].
  Proof.
    intros.
    apply typ_sub with (e; x := d : t) e (mkdecl x t (Some d)); simpl in |- *;
      auto.
    Inversion_typ (typ_wf _ _ _ H); auto.

    unfold add_def in |- *; auto.
    constructor; compute in |- *; auto.
  Qed.


  Lemma subst_wf :
    forall e x y m a b,
      e |- m : a -> |- ((e; x : a); y : b) -> |- (e; y : b[0:=m]).
  Proof.
    intros.
    Inversion_typ H0.
    apply wf_var with s.
    apply (substitution _ _ _ _ _ _ H1 H).
  Qed.

  Lemma subst_wf2 :
    forall e x y z m a b c,
      e |- m : a ->
        |- (((e; x : a); y : b); z : c) -> |- ((e; y : b[0:=m]); z : c[1:=m]).
  Proof.
    intros.
    Inversion_typ H0.
    apply wf_var with s.
    fold (%%s[1:=m]) in |- *.
    apply substitution1 with x a; trivial.
  Qed.


  Theorem wf_type_subst :
    forall e x d t T, wf_type (e; x : t) T -> e |- d : t -> wf_type e T[0:=d].
  Proof.
    simple destruct 1; simpl in |- *; intros; auto.
    constructor 1 with s.
    fold (%%s[0:=d]) in |- *.
    apply substitution with (1 := H0); auto.
  Qed.

  Theorem wf_type_subst_def :
    forall e x d t T, wf_type (e; x := d : t) T -> wf_type e T[0:=d].
  Proof.
    simple destruct 1; simpl in |- *; intros; auto.
    constructor 1 with s.
    fold (%%s[0:=d]) in |- *.
    apply substitution_def with (1 := H0).
  Qed.



  Lemma typ_replace :
    forall (e : env) x s ts m t d,
      (e; d) |- s : ^(1//0)ts ->
        (e; x : ts) |- m : t -> (e; d) |- replace s m 0 : replace s t 0.
  Proof.
    intros.
    unfold replace in |- *.
    apply
      typ_sub
      with (d :: e; x : ^(1//0)ts) (d :: e) (liftd_rec 1 (mkdecl x ts None) 0);
        trivial.
    apply typ_weak with e d (e; x : ts); trivial.
    apply typ_wf with (1 := H).

    apply (Ins_tl _ _ _ _ _ (mkdecl x ts None) (Ins_hd e d)).

    unfold add_var, liftd_rec in |- *; simpl in |- *.
    constructor; compute in |- *; auto.
  Qed.



  (* Type correctness: the right term of a judgement is a well-formed type *)

  Lemma typ_rel_wf : forall e n t, type_of_rel e n t -> |- e -> wf_type e t.
  Proof.
    intros e n t (d, eqd, itm).
    rewrite eqd.
    elim itm; intros.
    apply wf_type_thinning with l; auto.
    inversion_clear H; simpl in |- *.
    constructor 1 with s; auto.

    constructor 1 with s; auto.

    rewrite simpl_lift.
    apply wf_type_thinning with l; auto.
    apply H0.
    exact (inv_wf _ _ H1).
  Qed.


  Theorem type_correctness : forall e t T, e |- t : T -> wf_type e T.
  Proof.
    simple induction 1; intros; auto.
    apply typ_rel_wf with v; auto.

    apply var_wft.
    apply wf_thin1; auto.

    apply var_wft; trivial.

    left with s; auto.

    left with s; auto.

    left with s; auto.
  Qed.


  Lemma typ_ty_clos : forall e m t, e |- m : t -> closed (length e) t.
  Proof.
    intros.
    generalize (type_correctness _ _ _ H); intros.
    inversion_clear H0; auto.
    apply typ_clos with (1 := H1).
  Qed.

  Lemma typ_thin_app :
    forall e f m t, |- e -> f |- m : t -> (f ++ e) |- m : t.
  Proof.
    intros.
    generalize (typ_thin_app_wk _ _ _ _ H H0).
    rewrite closed_lift_rec; trivial.
    apply typ_ty_clos with (1 := H0).
  Qed.

  (* 3 applications of typ_thin_app *)
  Lemma thin_par_def0 : forall e m t, ( ) |- m : t -> |- e -> e |- m : t.
  Proof.
    intros.
    apply (typ_thin_app _ _ _ _ H0 H).
  Qed.

  Lemma thin_par_def :
    forall e x p m t, (( ); x : p) |- m : t -> |- e -> (e; x : p) |- m : t.
  Proof.
    intros.
    apply (typ_thin_app _ _ _ _ H0 H).
  Qed.

  Lemma thin_par_def2 :
    forall e x y p q m t,
      ((( ); x : p); y : q) |- m : t -> |- e -> ((e; x : p); y : q) |- m : t.
  Proof.
    intros.
    apply (typ_thin_app _ _ _ _ H0 H).
  Qed.

  Section ContextReduction.

    Variable stab : env -> decl -> decl -> Prop.

    Hypothesis
      incl_rel_typ : forall e o1 o2, stab e o1 o2 -> e |- TYPE o1 <= TYPE o2.

    Hypothesis
      stable_in_sign1 :
      forall c f e m t u,
        rel_in_list stab f e -> in_sign1 c e m t u -> in_sign1 c f m t u.

    Hypothesis stable_le : R_stable stab le_type.

    Lemma reduction_env_rec :
      forall e t T,
        e |- t : T -> forall f, rel_in_list stab f e -> |- f -> f |- t : T.
    Proof.
      intros e t T j.
      elim j using
        typ_mind
        with
          (P0 := fun e : env =>
            forall l f e0 f0,
              split e0 e l ->
                rel_in_list stab f0 e0 -> |- f0 -> split f0 f l -> |- f);
       intros.
      auto.

      inversion_clear H1.
      rewrite H4.
      elim rel_list_item with (1 := H5) (2 := H2); intros.
      apply Typ_rel; auto.
      split with x; auto.

      elim item_trunc with (1 := H5); intros.
      elim H1 with (1 := H6).
      intros (d, stbl, itm) tr.
      apply typ_conv_wf with (^(S v)(TYPE d)).
      apply Typ_rel; auto.
      split with d; trivial.

      apply wf_type_thinning with x0; auto.
      elim wf_sort with (1 := H5) (3 := H6); intros; trivial.
      constructor 1 with x1; trivial.

      apply iter_R_lift with x0; [ exact le_type_lift | assumption | ].
      apply incl_rel_typ; trivial.

      apply Typ_cst0; trivial.

      apply Typ_cst1 with t0; auto.
      apply H0 with (l := (nil; x_ : u)) (2 := H6); unfold add_var in |- *; auto.

      apply H2 with (l := (nil; x_ : t0)) (2 := H6); unfold add_var in |- *; auto.

      apply stable_in_sign1 with (1 := H6); trivial.

      cut (f |- Forall x : a, u : %%s); intros; auto.
      apply Typ_lam with s; auto.
      apply H0.
      unfold add_var in |- *; auto.

      Inversion_typ H5.
      apply typ_wf with (1 := H7).

      apply Typ_prd with s1 s2; auto.
      apply H2.
      unfold add_var in |- *; auto.

      apply wf_var with s1; auto.

      apply typ_sum with s1 s2; auto.

      apply Typ_pair with s; auto.

      apply Typ_conv with U s; auto.
      apply stable_le with e0; auto.

      apply Typ_conv_srt with U; auto.
      apply stable_le with e0; auto.

      generalize H2 H1.
      inversion_clear H; intro spl_nil; inversion_clear spl_nil; trivial.

      unfold add_var in H3.
      generalize H6 H5; clear H5 H6.
      inversion_clear H3.
      intro spl_nil; inversion_clear spl_nil; trivial.

      intro spl_cs; inversion_clear spl_cs; intros.
      fold (l1; x : T0) in |- *.
      apply wf_var with s.
      apply H2.
      apply rel_split with (1 := H4) (2 := H3) (3 := H5).

      apply H0 with (l := l2) (2 := H4); trivial.

      unfold add_def in H5.
      generalize H8 H7; clear H7 H8.
      inversion_clear H5.
      intros spl_nil; inversion_clear spl_nil; trivial.

      intro spl_cs; inversion_clear spl_cs; intros.
      fold (l1; x := M : T0) in |- *.
      cut (|- l1); intros.
      cut (rel_in_list stab l1 e0); intros.
      apply wf_def with s; auto.

      apply rel_split with (1 := H6) (2 := H5) (3 := H7).

      apply H0 with (l := l2) (2 := H6); trivial.
    Qed.

  End ContextReduction.

  Theorem subtype_env :
    forall e f t T,
      e |- t : T -> rel_in_list (rel_typ le_type) f e -> |- f -> f |- t : T.
  Proof.
    intros.
    apply reduction_env_rec with (rel_typ le_type) e; eauto.
    simple induction 1; simpl in |- *; auto.
    exact in_sign_stable.
    apply Rule_stable.
  Qed.
  (* Ce lemme est le seul a utiliser la stabilite de la relation de sous-typage.
   * Sert-il ?
   * il peut servir a` renommer les variables
   *)

End IMetatheory.

(* The following belong to the inside of the section... *)

Hint Resolve @Wft_top: core.
Hint Resolve @Typ_srt @Typ_rel @Typ_cst0 @Wf_nil: core.

Hint Unfold rl_sat pr_sat: core.

(* Syntax for context and judgements *)

Delimit Scope pts_scope with pts.

Notation "e ⊦ m : t" := (typ e m t) (at level 50, m at next level) : pts_scope.
Notation "⊦  e" := (wf e) (at level 50) : pts_scope.
Notation "e ⊦ t <= u" := (le_type e t u) (at level 50, t at next level) : pts_scope.

Ltac InvWfHyp H :=
  unfold add_var, add_def in H; inversion H using inversion_wf;
    unfold inv_pred_wf in |- *.

Ltac InvTypHyp H :=
  inversion H using inversion_lemma; unfold inv_pred_typ in |- *.

Ltac Inversion_typ T :=
  generalize T; intro INVTYP; InvTypHyp INVTYP || InvWfHyp INVTYP;
    clear INVTYP; intros.

End IMetatheory.
