(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export Ctxt.
Require Export ParSubst.

Module Type IParCtxt
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MRules : IRules P MTermes MEnv)
  (Import MCtxt : ICtxt P MTermes MEnv MRules).

  Include IParSubst P MTermes.

Section ParallelContextualClosure.

  Variable R : red_rule.


  Inductive par_ctxt : red_rule :=
    | Pctx_rule : forall e t u, R e t u -> par_ctxt e t u
    | Pctx_srt : forall e s, par_ctxt e %%s %%s
    | Pctx_rel : forall e n, par_ctxt e #n #n
    | Pctx_cst0 : forall e c, par_ctxt e (Cst0 c) (Cst0 c)
    | Pctx_cst1 :
        forall e c t t', par_ctxt e t t' -> par_ctxt e (Cst1 c t) (Cst1 c t')
    | Pctx_lam :
        forall e x m m' n n',
        par_ctxt e m m' ->
        par_ctxt (e; x : m) n n' ->
        par_ctxt e (Fun x : m => n) (Fun x : m' => n')
    | Pctx_prd :
        forall e x m m' n n',
        par_ctxt e m m' ->
        par_ctxt (e; x : m) n n' ->
        par_ctxt e (Forall x : m, n) (Forall x : m', n')
    | Pctx_sum :
        forall e m m' n n',
        par_ctxt e m m' -> par_ctxt e n n' -> par_ctxt e (m ** n) (m' ** n')
    | Pctx_pair :
        forall e m m' n n',
        par_ctxt e m m' -> par_ctxt e n n' -> par_ctxt e (m,,n) (m',,n').

  Hint Resolve Pctx_rule Pctx_srt Pctx_rel Pctx_cst0 Pctx_cst1 Pctx_lam
    Pctx_prd Pctx_sum Pctx_pair: core.

  Lemma par_refl : forall t e, par_ctxt e t t.
Proof.
simple induction t; intros; auto.
Qed.
  Hint Resolve par_refl: core.

  Lemma incl_ctxt_par : forall e u v, ctxt R e u v -> par_ctxt e u v.
Proof.
simple induction 1; intros; auto.
Qed.

  Lemma incl_wctxt_par : forall e u v, wctxt R e u v -> par_ctxt e u v.
Proof.
simple induction 1; intros; auto.
Qed.

  Lemma incl_par_red : forall e u v, par_ctxt e u v -> red R e u v.
Proof.
simple induction 1; intros; auto.
apply red_trans with (Fun x : m => n'); auto.
Qed.


  Section LiftAndSubst.

  Hypothesis Rlift : R_lift R.

  Lemma par_ctxt_lift_rec :
   forall d g e x y,
   par_ctxt e x y ->
   forall k f, ins_in_env g d k e f -> par_ctxt f ^(1//k)x ^(1//k)y.
Proof.
simple induction 1; simpl in |- *; intros; auto.
constructor 1.
apply Rlift with g e0 d; trivial.
Qed.


  Lemma par_ctxt_liftn :
   forall (e f : env) n x y,
   trunc n f e -> par_ctxt e x y -> par_ctxt f ^(n)x ^(n)y.
Proof.
simple induction 1; intros.
repeat rewrite lift0; trivial.

rewrite simpl_lift.
unfold lift at 1 in |- *.
rewrite simpl_lift.
unfold lift at 2 in |- *.
apply par_ctxt_lift_rec with x0 e0 e0; auto.
Qed.


  Hypothesis Rsubst : R_subst R.


  Lemma par_ctxt_subst_rec :
   forall s d g e x y,
   par_ctxt e x y ->
   forall k f, sub_in_env g s d k e f -> par_ctxt f x[k:=s] y[k:=s].
Proof.
simple induction 1; simpl in |- *; intros; auto.
constructor.
apply Rsubst with (2 := H1); trivial.
Qed.

  Lemma par_ctxt_subst_l :
   forall e x y,
   par_ctxt e x y -> forall t k f, trunc k f e -> par_ctxt f t[k:=x] t[k:=y].
Proof.
simple induction t; intros; simpl in |- *; auto.
simpl in |- *.
elim (lt_eq_lt_dec k n); intros; auto.
elim a; intros; auto.
apply par_ctxt_liftn with e; auto.
Qed.


  Lemma par_ctxt_subst_l2 :
   forall e x y x' y',
   par_ctxt e x y ->
   par_ctxt e x' y' ->
   forall t k f,
   trunc k f e -> par_ctxt f (t[S k:=x'])[k:=x] (t[S k:=y'])[k:=y].
Proof.
simple induction t; intros; try (simpl in |- *; auto; fail).
unfold subst_rec at 2 4 in |- *.
elim (lt_eq_lt_dec (S k) n); intros.
elim a; intros.
apply par_ctxt_subst_l with e; auto.

repeat rewrite simpl_subst; auto.
apply par_ctxt_liftn with e; auto.

apply par_ctxt_subst_l with e; auto.
Qed.

  Lemma par_ctxt_psubst :
   forall e ts1 ts2,
   for_all2 (par_ctxt e) ts1 ts2 ->
   forall t k f,
   trunc k f e -> par_ctxt f (psubst_rec ts1 t k) (psubst_rec ts2 t k).
  simple induction t; simpl in |- *; auto.
intros.
elim (le_gt_dec k n); auto.
intros.
apply par_ctxt_liftn with (1 := H0); trivial.

generalize (n - k).
elim H; intros; auto.
case n0; simpl in |- *; auto.
Qed.

  End LiftAndSubst.


  Lemma par_ctxt_stable :
   forall stab_rel, R_stable stab_rel R -> R_stable stab_rel par_ctxt.
Proof.
red in |- *.
simple induction 2; intros; auto.
eauto.
Qed.


End ParallelContextualClosure.

  Hint Resolve Pctx_rule Pctx_srt Pctx_rel Pctx_cst0 Pctx_cst1 Pctx_lam
    Pctx_prd Pctx_sum Pctx_pair: core.
  Hint Resolve par_refl incl_ctxt_par incl_wctxt_par: core.

End IParCtxt.
