(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export Env.

Module Type IRules
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes).


Section Operators.

  Variable R R' : red_rule.

  (* union of two rules *)
  Definition reunion : red_rule := fun e => union _ (R e) (R' e).

  (* Reunion of a list of reduction rules *)
  Inductive Rules : Type :=
    | Mt_rl : Rules
    | Cs_rl : red_rule -> Rules -> Rules.

  Inductive For_all (P : red_rule -> Prop) : Rules -> Prop :=
    | Fa_mt : For_all P Mt_rl
    | Fa_cs : forall r l, P r -> For_all P l -> For_all P (Cs_rl r l).

  Hint Resolve Fa_mt Fa_cs: core.


  Fixpoint reunion_list (l : Rules) : red_rule :=
    match l with
    | Mt_rl => fun _ _ _ => False
    | Cs_rl r l' => fun e => union _ (r e) (reunion_list l' e)
    end.


  Hint Unfold reunion reunion_list: core.

  (* reduction in any context *)
  Inductive ctxt : red_rule :=
    | Ctx_rule : forall e t u, R e t u -> ctxt e t u
    | Ctx_cst : forall e c t t', ctxt e t t' -> ctxt e (Cst1 c t) (Cst1 c t')
    | Ctx_lam_l :
        forall e x m m' n,
        ctxt e m m' -> ctxt e (Fun x : m => n) (Fun x : m' => n)
    | Ctx_lam_r :
        forall e x m n n',
        ctxt (e; x : m) n n' -> ctxt e (Fun x : m => n) (Fun x : m => n')
    | Ctx_prd_l :
        forall e x m m' n,
        ctxt e m m' -> ctxt e (Forall x : m, n) (Forall x : m', n)
    | Ctx_prd_r :
        forall e x m n n',
        ctxt (e; x : m) n n' -> ctxt e (Forall x : m, n) (Forall x : m, n')
    | Ctx_sum_l : forall e t t' n, ctxt e t t' -> ctxt e (t ** n) (t' ** n)
    | Ctx_sum_r : forall e t n n', ctxt e n n' -> ctxt e (t ** n) (t ** n')
    | Ctx_pair_l : forall e m m' n, ctxt e m m' -> ctxt e (m,,n) (m',,n)
    | Ctx_pair_r : forall e m n n', ctxt e n n' -> ctxt e (m,,n) (m,,n').

  Hint Resolve Ctx_rule Ctx_cst Ctx_lam_l Ctx_lam_r Ctx_prd_l Ctx_prd_r
    Ctx_sum_l Ctx_sum_r Ctx_pair_l: core.
  Hint Resolve Ctx_pair_r: core.


  (* reduction in weak contexts (quantifier-free) *)
  Inductive wctxt : red_rule :=
    | WCtx_rule : forall e t u, R e t u -> wctxt e t u
    | WCtx_cst : forall e c t t', wctxt e t t' -> wctxt e (Cst1 c t) (Cst1 c t')
    | WCtx_lam_l :
        forall e x m m' n,
        wctxt e m m' -> wctxt e (Fun x : m => n) (Fun x : m' => n)
    | WCtx_prd_l :
        forall e x m m' n,
        wctxt e m m' -> wctxt e (Forall x : m, n) (Forall x : m', n)
    | WCtx_sum_l : forall e t t' n, wctxt e t t' -> wctxt e (t ** n) (t' ** n)
    | WCtx_sum_r : forall e t n n', wctxt e n n' -> wctxt e (t ** n) (t ** n')
    | WCtx_pair_l : forall e m m' n, wctxt e m m' -> wctxt e (m,,n) (m',,n)
    | WCtx_pair_r : forall e m n n', wctxt e n n' -> wctxt e (m,,n) (m,,n').

  Hint Constructors wctxt.

  Lemma wctxt_ctxt : forall e t u, wctxt e t u -> ctxt e t u.
  Proof.
    induction 1; intuition.
  Qed.

  Section LiftProperties.

    Hypothesis R_lifts : R_lift R.
    Hypothesis R_lifts' : R_lift R'.

    Lemma reunion_lift : R_lift reunion.
Proof.
red in |- *; simple destruct 1; eauto.
Qed.

    Lemma reunion_list_lift :
     forall l, For_all R_lift l -> R_lift (reunion_list l).
Proof.
simple induction 1; simpl in |- *; intros.
red in |- *; intros.
auto.

red in |- *; simple destruct 1; eauto.
Qed.

    Lemma ctxt_lift : R_lift ctxt.
Proof.
red in |- *.
simple induction 1; simpl in |- *; intros; auto.
eauto.
Qed.

    Lemma wctxt_lift : R_lift wctxt.
Proof.
red in |- *.
simple induction 1; simpl in |- *; intros; auto.
eauto.
Qed.

  End LiftProperties.


  Section SubstProperties.

    Hypothesis R_substs : R_subst R.
    Hypothesis R_substs' : R_subst R'.

    Lemma reunion_subst : R_subst reunion.
Proof.
red in |- *.
simple destruct 1; eauto.
Qed.

    Lemma reunion_list_subst :
     forall l, For_all R_subst l -> R_subst (reunion_list l).
Proof.
simple induction 1; simpl in |- *; intros.
red in |- *; intros.
auto.

red in |- *; simple destruct 1; eauto.
Qed.

    Lemma ctxt_subst : R_subst ctxt.
Proof.
red in |- *.
simple induction 1; simpl in |- *; intros; auto.
eauto.
Qed.

    Lemma wctxt_subst : R_subst wctxt.
Proof.
red in |- *.
simple induction 1; simpl in |- *; intros; auto.
eauto.
Qed.


  End SubstProperties.


  Section SubstWeakProperties.

    Hypothesis R_substs : R_subst_weak R.
    Hypothesis R_substs' : R_subst_weak R'.


    Lemma reunion_subst_weak : R_subst_weak reunion.
Proof.
red in |- *.
simple destruct 1; intros.
elim (R_substs _ _ _ _ _ _ H0 _ _ H1); intros; auto.
apply rt_trans with y; trivial.

elim (R_substs' _ _ _ _ _ _ H0 _ _ H1); intros; auto.
apply rt_trans with y; trivial.
Qed.


    Lemma ctxt_subst_weak : R_subst_weak ctxt.
Proof.
red in |- *.
simple induction 1; simpl in |- *; intros; auto.
elim (R_substs _ _ _ _ _ _ H0 _ _ H1); intros; auto; prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := sub_in_env_var _ _ _ _ _ x _ m0 H2); intros; auto;
 try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := sub_in_env_var _ _ _ _ _ x _ m0 H2); intros; auto;
 try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.
Qed.

    Lemma wctxt_subst_weak : R_subst_weak wctxt.
Proof.
red in |- *.
simple induction 1; simpl in |- *; intros; auto.
elim (R_substs _ _ _ _ _ _ H0 _ _ H1); intros; auto; prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.

elim H1 with (1 := H2); intros; auto; try prolog [ rt_trans ] 5.
Qed.


  End SubstWeakProperties.


  Section StableProperties.

    Variable stab_rel : env -> decl -> decl -> Prop.
    Hypothesis stable : R_stable stab_rel R.
    Hypothesis stable' : R_stable stab_rel R'.

    Lemma reunion_stable : R_stable stab_rel reunion.
Proof.
red in |- *.
simple destruct 1; eauto.
Qed.

  Lemma reunion_list_stable :
   forall R l, For_all (R_stable R) l -> R_stable R (reunion_list l).
Proof.
simple induction 1; simpl in |- *; intros.
red in |- *; intros.
auto.

red in |- *; simple destruct 1; eauto.
Qed.

    Lemma ctxt_stable : R_stable stab_rel ctxt.
Proof.
red in |- *.
simple induction 1; intros; auto.
eauto.
Qed.

    Lemma wctxt_stable : R_stable stab_rel wctxt.
Proof.
red in |- *.
simple induction 1; intros; auto.
eauto.
Qed.

  End StableProperties.


End Operators.

  Hint Unfold reunion reunion_list: core.
  Hint Resolve Fa_mt Fa_cs: core.
  Hint Constructors ctxt wctxt: core.

  Hint Resolve reunion_lift reunion_subst reunion_subst_weak reunion_stable
    reunion_list_lift reunion_list_subst reunion_list_stable
    wctxt_lift wctxt_subst wctxt_subst_weak wctxt_stable
    ctxt_lift ctxt_subst ctxt_subst_weak ctxt_stable: core.

End IRules.
