(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export Mmain.

Module Type ITermes (Import P : PTSPARAMS).

  Inductive term : Set :=
    | Srt : sort -> term
    | Rel : nat -> term
    | Cst0 : oper -> term
    | Cst1 : oper -> term -> term
    | Prd : name -> term -> term -> term
    | Lam : name -> term -> term -> term
    | Sum : term -> term -> term
    | Pair0 : term -> term -> term.

  Instance eq_dec_term : has_eq_dec term.
  Proof.
    hnf.
    unfold decide in |- *.
    decide equality; apply eq_dec.
  Defined.


  Lemma is_a_sort : forall t, decide (exists s : _, t = Srt s).
  Proof.
    (*Realizer [t]Cases t of
       (Srt s) => left
       | _ => right
       end.*)
    intros [s| n| c| c u| x a b| x a m| a b| u v];
      try (right; red in |- *; intros (s, eq_srt); discriminate).
    left; exists s; trivial.
  Defined.

  Lemma is_pair_dec :
    forall t,
      {p : term * term | let (u, v) := p in t = Pair0 u v} +
      {(forall u v, t <> Pair0 u v)}.
  Proof.
  (*Realizer [t]
     Cases t of
     (Pair u v) => (inleft ? (u,v))
     | _ => (inright ?)
     end.*)
    intros [s| n| c| c u| x a b| x a m| a b| u v];
      try (right; red in |- *; intros u' v' eq_pair; discriminate).
    left; exists (u, v); trivial.
  Defined.

  Inductive is_sum : term -> Set :=
    Is_sum : forall a b, is_sum (Sum a b).
  Hint Resolve Is_sum: core.

  Lemma is_sum_dec : forall a, is_sum a + {is_sum a -> False}.
  Proof.
  (*Realizer [a]
     Cases a of
     (Sum a b) => (inleft ? (Is_sum a b))
     | _ => (inright is_sum)
     end.*)
    intros [s| n| c| c u| x a b| x a m| a b| u v];
      try (right; intros eq_sum; inversion eq_sum; fail).
    left; auto.
  Defined.

  Inductive is_prd : term -> Set :=
    Is_prd : forall x a b, is_prd (Prd x a b).
  Hint Resolve Is_prd: core.

  Lemma is_prd_dec : forall a, is_prd a + {is_prd a -> False}.
  Proof.
  (*Realizer [a]
     Cases a of
     (Prd x a b) => (inleft ? (Is_prd x a b))
     | _ => (inright is_prd)
     end.*)
    intros [s| n| c| c u| x a b| x a m| a b| u v];
      try (right; intros eq_prd; inversion eq_prd; fail).
    left; auto.
  Defined.


  Fixpoint lift_rec (n : nat) (t : term) (k : nat) {struct t} : term :=
    match t with
    | Srt s => Srt s
    | Rel i =>
        Rel match le_gt_dec k i with
            | left _ => n + i
            | right _ => i
            end
    | Cst0 c => Cst0 c
    | Cst1 c a => Cst1 c (lift_rec n a k)
    | Lam x a m => Lam x (lift_rec n a k) (lift_rec n m (S k))
    | Prd x a b => Prd x (lift_rec n a k) (lift_rec n b (S k))
    | Sum a b => Sum (lift_rec n a k) (lift_rec n b k)
    | Pair0 u v => Pair0 (lift_rec n u k) (lift_rec n v k)
    end.

  Definition lift n t := lift_rec n t 0.


  Lemma lift_split :
    forall n k p (P : term -> Prop),
      (p > n -> P (Rel n)) ->
      (p <= n -> P (Rel (k + n))) -> P (lift_rec k (Rel n) p).
  Proof.
    simpl in |- *; intros.
    elim (le_gt_dec p n); auto.
  Qed.

  Lemma lift_ref_ge :
    forall k n p, p <= n -> lift_rec k (Rel n) p = Rel (k + n).
  Proof.
    intros.
    apply lift_split; intros; auto.
    absurd (p <= n); omega.
  Qed.


  Lemma lift_ref_lt : forall k n p, p > n -> lift_rec k (Rel n) p = Rel n.
  Proof.
    intros.
    apply lift_split; intros; auto.
    absurd (p <= n); omega.
  Qed.




  Lemma lift_rec0 : forall M k, lift_rec 0 M k = M.
  Proof.
    simple induction M; intros.
    reflexivity.

    apply lift_split; intros; auto.

    reflexivity.

    simpl in |- *; rewrite H; auto.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.
  Qed.



  Lemma lift0 : forall M, lift 0 M = M.
  Proof fun M => lift_rec0 M 0.
  Hint Rewrite lift0 : core.


  Lemma simpl_lift_rec :
    forall n p M k i,
      i <= k + n ->
      k <= i -> lift_rec p (lift_rec n M k) i = lift_rec (p + n) M k.
  Proof.
    simple induction M; intros.
    reflexivity.

    apply lift_split; intros.
    rewrite lift_ref_lt.
    rewrite lift_ref_lt; auto.

    omega.

    rewrite lift_ref_ge.
    rewrite lift_ref_ge; auto.
    rewrite plus_assoc; auto.

    omega.

    reflexivity.

    simpl in |- *; rewrite H; auto.

    simpl in |- *; rewrite H; auto; rewrite H0; simpl in |- *; auto with arith.

    simpl in |- *; rewrite H; auto; rewrite H0; simpl in |- *; auto with arith.

    simpl in |- *; rewrite H; auto; rewrite H0; auto.

    simpl in |- *; rewrite H; auto; rewrite H0; simpl in |- *; auto.
  Qed.



  Lemma simpl_lift : forall M n, lift (S n) M = lift 1 (lift n M).
  Proof.
    intros; unfold lift in |- *.
    rewrite simpl_lift_rec; auto with arith.
  Qed.


  Lemma permute_lift_rec :
    forall n p M {k i},
      i <= k ->
      lift_rec p (lift_rec n M k) i = lift_rec n (lift_rec p M i) (p + k).
  Proof.
    simple induction M; intros.
    reflexivity.

    apply lift_split; intros.
    apply lift_split; intros.
    rewrite lift_ref_lt; auto.
    omega.

    rewrite lift_ref_lt; auto.
    omega.

    rewrite lift_ref_ge; try omega.
    rewrite lift_ref_ge; try omega.
    rewrite lift_ref_ge; try omega.
    rewrite plus_permute; auto.

    reflexivity.

    simpl in |- *; rewrite H; auto.

    simpl in |- *; rewrite plus_n_Sm.
    rewrite H; auto; rewrite H0; auto with arith.

    simpl in |- *; rewrite plus_n_Sm.
    rewrite H; auto; rewrite H0; auto with arith.

    simpl in |- *; rewrite H; auto; rewrite H0; auto.

    simpl in |- *; rewrite H; auto; rewrite H0; auto.
  Qed.

  Lemma permute_lift_rec1 :
    forall n M {k i},
      i <= k ->
      lift_rec 1 (lift_rec n M k) i = lift_rec n (lift_rec 1 M i) (S k).
  Proof fun n => permute_lift_rec n 1.

  Lemma permute_lift :
    forall M n p k, lift n (lift_rec p M k) = lift_rec p (lift n M) (n + k).
  Proof fun M n p k => permute_lift_rec p n M (le_O_n k).






  Fixpoint subst_rec (ts t : term) (k : nat) {struct t} : term :=
    match t with
      | Srt s => Srt s
      | Rel i =>
        match lt_eq_lt_dec k i with
          | inleft (left _) => Rel (pred i)
          | inleft (right _) => lift k ts
          | inright _ => Rel i
        end
      | Cst0 c => Cst0 c
      | Cst1 c a => Cst1 c (subst_rec ts a k)
      | Lam x a m => Lam x (subst_rec ts a k) (subst_rec ts m (S k))
      | Prd x a b => Prd x (subst_rec ts a k) (subst_rec ts b (S k))
      | Sum a b => Sum (subst_rec ts a k) (subst_rec ts b k)
      | Pair0 u v => Pair0 (subst_rec ts u k) (subst_rec ts v k)
    end.


  Lemma subst_split :
    forall s n k (P : _ -> _ -> Prop),
      (k > n -> P n (Rel n)) ->
      P k (lift k s) ->
      (forall q, k <= q -> P (S q) (Rel q)) -> P n (subst_rec s (Rel n) k).
  Proof.
    simpl in |- *; intros.
    elim (lt_eq_lt_dec k n); intros.
    elim a.
    case n; simpl in |- *; intros; auto with arith.
    inversion_clear a0.

    simple destruct 1; auto.

    auto.
  Qed.


  Lemma subst_ref_lt : forall s n k, k > n -> subst_rec s (Rel n) k = Rel n.
  Proof.
    intros s n k.
    pattern n, (subst_rec s (Rel n) k) in |- *.
    apply subst_split; intros; auto.
    apply False_ind.
    omega.

    apply False_ind.
    omega.
  Qed.

  Lemma subst_ref_eq : forall s k, subst_rec s (Rel k) k = lift k s.
  Proof.
    intros.
    generalize (refl_equal k).
    pattern k at 1 3, (subst_rec s (Rel k) k) in |- *.
    apply subst_split; intros; auto.
    apply False_ind; omega.

    apply False_ind; omega.
  Qed.


  Lemma subst_ref_gt :
    forall s k n, k < n -> subst_rec s (Rel n) k = Rel (pred n).
  Proof.
    intros s k n.
    pattern n, (subst_rec s (Rel n) k) in |- *.
    apply subst_split; intros; auto.
    apply False_ind; omega.

    apply False_ind; omega.
  Qed.




  Lemma simpl_subst_rec :
    forall s n M p k,
      k <= p ->
      p <= k + n -> subst_rec s (lift_rec (S n) M k) p = lift_rec n M k.
  Proof.
    simple induction M; intros.
    reflexivity.

    apply lift_split; intros.
    rewrite lift_ref_lt; auto.
    rewrite subst_ref_lt; auto.
    omega.

    rewrite lift_ref_ge; auto.
    unfold plus at 1 in |- *; fold plus in |- *.
    rewrite subst_ref_gt; auto.
    omega.

    reflexivity.

    simpl in |- *; rewrite H; auto.

    simpl in |- *; rewrite H; auto; rewrite H0; simpl in |- *; auto with arith.

    simpl in |- *; rewrite H; auto; rewrite H0; simpl in |- *; auto with arith.

    simpl in |- *; rewrite H; auto; rewrite H0; simpl in |- *; auto.

    simpl in |- *; rewrite H; auto; rewrite H0; simpl in |- *; auto.
  Qed.


  Lemma simpl_subst :
    forall s M n p, p <= n -> subst_rec s (lift (S n) M) p = lift n M.
  Proof.
    intros; unfold lift in |- *.
    apply simpl_subst_rec; auto with arith.
  Qed.

  Lemma simpl_sl : forall m t, subst_rec m (lift 1 t) 0 = t.
  Proof.
    intros.
    rewrite simpl_subst; auto.
    apply lift0.
  Qed.



  Lemma commut_lift_subst_rec :
    forall n s M p k,
      k <= p ->
      lift_rec n (subst_rec s M p) k = subst_rec s (lift_rec n M k) (n + p).
  Proof.
    simple induction M; intros.
    reflexivity.

    apply lift_split; intros.
    rewrite subst_ref_lt; try omega.
    rewrite subst_ref_lt; try omega.
    rewrite lift_ref_lt; auto.

    pattern n0, (subst_rec s (Rel n0) p) in |- *.
    apply subst_split; intros.
    rewrite lift_ref_ge; auto.
    rewrite subst_ref_lt; auto with arith.

    rewrite subst_ref_eq.
    unfold lift in |- *.
    apply simpl_lift_rec; auto with arith.

    rewrite lift_ref_ge; try omega.
    rewrite subst_ref_gt; try omega.
    rewrite <- plus_n_Sm; simpl in |- *; auto.

    reflexivity.

    simpl in |- *; rewrite H; auto.

    simpl in |- *; rewrite plus_n_Sm.
    rewrite H; auto; rewrite H0; auto with arith.

    simpl in |- *; rewrite plus_n_Sm.
    rewrite H; auto; rewrite H0; auto with arith.

    simpl in |- *; rewrite H; auto; rewrite H0; auto.

    simpl in |- *; rewrite H; auto; rewrite H0; auto.
  Qed.

  Lemma commut_lift_subst_rec1 :
    forall s M p k,
      k <= p ->
      lift_rec 1 (subst_rec s M p) k = subst_rec s (lift_rec 1 M k) (S p).
  Proof commut_lift_subst_rec 1.

  Lemma commut_lift_subst :
    forall s M n k, subst_rec s (lift n M) (n + k) = lift n (subst_rec s M k).
  Proof.
    intros; unfold lift in |- *.
    rewrite commut_lift_subst_rec; auto with arith.
  Qed.

  Lemma commut_lift_subst0 :
    forall s M n, subst_rec s (lift n M) n = lift n (subst_rec s M 0).
  Proof.
    intros; unfold lift in |- *.
    rewrite commut_lift_subst_rec; auto.
  Qed.

  Lemma distr_lift_subst_rec :
    forall n s M p k,
      lift_rec n (subst_rec s M p) (p + k) =
      subst_rec (lift_rec n s k) (lift_rec n M (S (p + k))) p.
  Proof.
    simple induction M; intros.
    reflexivity.

    pattern n0, (subst_rec s (Rel n0) p) in |- *.
    apply subst_split; intros.
    rewrite lift_ref_lt; try omega.
    rewrite lift_ref_lt; try omega.
    rewrite subst_ref_lt; auto.

    rewrite lift_ref_lt; try omega.
    rewrite subst_ref_eq.
    rewrite permute_lift; auto.

    apply lift_split; intros.
    rewrite lift_ref_lt; try omega.
    rewrite subst_ref_gt; auto; try omega.

    rewrite lift_ref_ge; try omega.
    rewrite subst_ref_gt; try omega.
    rewrite <- plus_n_Sm; simpl in |- *; auto.

    reflexivity.

    simpl in |- *; rewrite H; reflexivity.

    simpl in |- *; unfold plus at 2 in |- *; fold (S p + k) in |- *.
    rewrite H; rewrite H0; reflexivity.

    simpl in |- *; unfold plus at 2 in |- *; fold (S p + k) in |- *.
    rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.
  Qed.


  Lemma distr_lift_subst :
    forall M N n k,
      lift_rec n (subst_rec N M 0) k =
      subst_rec (lift_rec n N k) (lift_rec n M (S k)) 0.
  Proof fun M N n => distr_lift_subst_rec n N M 0.

  Lemma distr_lift_subst1 :
    forall n k s t,
      lift_rec n (subst_rec s t 1) (S k) =
      subst_rec (lift_rec n s k) (lift_rec n t (S (S k))) 1.
  Proof.
    intros.
    fold (1 + k) in |- *.
    apply distr_lift_subst_rec.
  Qed.


  Lemma distr_subst_rec :
    forall sl sh n M p,
      subst_rec sh (subst_rec sl M p) (p + n) =
      subst_rec (subst_rec sh sl n) (subst_rec sh M (S (p + n))) p.
  Proof.
    simple induction M; intros.
    reflexivity.

    pattern n0, (subst_rec sl (Rel n0) p) in |- *.
    apply subst_split; intros.
    rewrite subst_ref_lt; try omega.
    rewrite subst_ref_lt; try omega.
    rewrite subst_ref_lt; auto.

    rewrite subst_ref_lt; try omega.
    rewrite subst_ref_eq.
    apply commut_lift_subst.

    pattern q, (subst_rec sh (Rel q) (p + n)) in |- *.
    apply subst_split; intros.
    rewrite subst_ref_lt; try omega.
    rewrite subst_ref_gt; auto; try omega.

    rewrite subst_ref_eq.
    rewrite simpl_subst; auto; try omega.

    rewrite subst_ref_gt; try omega.
    rewrite subst_ref_gt; simpl in |- *; try omega.
    reflexivity.

    reflexivity.

    simpl in |- *; unfold plus at 2 in |- *; fold (S p + n) in |- *.
    rewrite H; reflexivity.

    simpl in |- *; unfold plus at 2 in |- *; fold (S p + n) in |- *.
    rewrite H; rewrite H0; reflexivity.

    simpl in |- *; unfold plus at 2 in |- *; fold (S p + n) in |- *.
    rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.
  Qed.


  Lemma distr_subst :
    forall sl sh M n,
      subst_rec sh (subst_rec sl M 0) n =
      subst_rec (subst_rec sh sl n) (subst_rec sh M (S n)) 0.
  Proof fun sl sh M n => distr_subst_rec sl sh n M 0.

  Lemma distr_subst1 :
    forall k s1 s2 t,
      subst_rec s1 (subst_rec s2 t 1) (S k) =
      subst_rec (subst_rec s1 s2 k) (subst_rec s1 t (S (S k))) 1.
  Proof.
    intros.
    fold (1 + k) in |- *.
    apply distr_subst_rec.
  Qed.



  Definition replace s t k := subst_rec s (lift_rec 1 t (S k)) k.

  Lemma simpl_replace :
    forall k m t b,
      subst_rec m (replace t b k) k = subst_rec (subst_rec m t 0) b k.
  Proof.
    unfold replace in |- *; intros.
    pattern k at 3 in |- *.
    replace k with (k + 0); auto.
    rewrite distr_subst_rec.
    rewrite simpl_subst_rec; auto.
    rewrite lift_rec0; trivial.

    omega.
  Qed.

  Lemma simpl_replace_lift : forall s m, replace s (lift 1 m) 0 = lift 1 m.
  Proof.
    unfold replace in |- *; intros.
    unfold lift in |- *.
    rewrite simpl_lift_rec; auto.
    simpl in |- *.
    rewrite simpl_subst_rec; auto with arith.
  Qed.


  Lemma replace0 : forall t k, replace (Rel 0) t k = t.
  Proof.
    unfold replace in |- *.
    simple induction t; intros.
    reflexivity.

    apply lift_split; intros.
    generalize H.
    pattern n, (subst_rec (Rel 0) (Rel n) k) in |- *.
    apply (subst_split (Rel 0)); intros.
    reflexivity.

    unfold lift in |- *.
    simpl in |- *.
    rewrite <- plus_n_O; reflexivity.

    apply False_ind; omega.

    rewrite subst_ref_gt; simpl in |- *; auto.

    reflexivity.

    simpl in |- *; rewrite H; auto.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.

    simpl in |- *; rewrite H; rewrite H0; reflexivity.
  Qed.

  Lemma comm_subst :
    forall a b m,
      subst_rec a (subst_rec b m 1) 0 = subst_rec b (subst_rec (lift 1 a) m 0) 0.
  Proof.
    intros.
    rewrite distr_subst.
    rewrite simpl_sl; trivial.
  Qed.

  Lemma subst_id :
    forall a m,
      subst_rec (Rel 0) (subst_rec (lift 1 a) (lift_rec 1 m 2) 1) 0 =
      subst_rec a m 1.
  Proof.
    intros.
    unfold lift in |- *.
    replace 2 with (S (1 + 0)); auto.
    rewrite <- distr_lift_subst_rec.
    simpl in |- *.
    fold (replace (Rel 0) (subst_rec a m 1) 0) in |- *.
    apply replace0.
  Qed.


  Inductive closed : nat -> term -> Prop :=
  | Cl_srt : forall k s, closed k (Srt s)
  | Cl_rel : forall k n, n < k -> closed k (Rel n)
  | Cl_cst0 : forall k c, closed k (Cst0 c)
  | Cl_cst1 : forall k c m, closed k m -> closed k (Cst1 c m)
  | Cl_abs :
    forall k x a m, closed k a -> closed (S k) m -> closed k (Lam x a m)
  | Cl_prd :
    forall k x a m, closed k a -> closed (S k) m -> closed k (Prd x a m)
  | Cl_sum : forall k a b, closed k a -> closed k b -> closed k (Sum a b)
  | Cl_pair : forall k m a, closed k m -> closed k a -> closed k (Pair0 m a).

  Hint Constructors closed.

  Lemma closed_dec : forall n t, decide (closed n t).
  Proof.
    fix 2; intros n t.
    refine (
      match t as t0 return decide (closed n t0) with
        | Srt s => left _ _
        | Rel k => if le_lt_dec n k then right _ _ else left _ _
        | Cst0 c => left _ _
        | Cst1 c m => if closed_dec n m then left _ _ else right _ _
        | Lam x a b =>
          if closed_dec n a
            then if closed_dec (S n) b then left _ _ else right _ _
            else right _ _
        | Prd x a b =>
          if closed_dec n a
            then if closed_dec (S n) b then left _ _ else right _ _
            else right _ _
        | Sum a b =>
          if closed_dec n a
            then if closed_dec n b then left _ _ else right _ _
            else right _ _
        | Pair0 a b =>
          if closed_dec n a
            then if closed_dec n b then left _ _ else right _ _
            else right _ _
      end
    ); try ((constructor || contradict _H; inversion _H); auto; fail).
    intro H; inversion H; elimtype False; omega.
  Defined.

  Lemma closed_weak : forall n k t, n <= k -> closed n t -> closed k t.
  Proof.
    intros n k t lenk clst.
    generalize k lenk.
    elim clst; intros; constructor; auto with arith.
    omega.
  Qed.

  Lemma closed_lift_rec : forall n t k, closed k t -> lift_rec n t k = t.
  Proof.
    simple induction 1; intros.
    reflexivity.

    rewrite lift_ref_lt; auto.

    reflexivity.

    simpl in |- *; rewrite H1; trivial.

    simpl in |- *; rewrite H1; rewrite H3; trivial.

    simpl in |- *; rewrite H1; rewrite H3; trivial.

    simpl in |- *; rewrite H1; rewrite H3; trivial.

    simpl in |- *; rewrite H1; rewrite H3; trivial.
  Qed.


  Lemma closed_subst_rec : forall u t k, closed k t -> subst_rec u t k = t.
  Proof.
    simple induction 1; intros.
    reflexivity.

    rewrite subst_ref_lt; auto.

    reflexivity.

    simpl in |- *; rewrite H1; trivial.

    simpl in |- *; rewrite H1; rewrite H3; trivial.

    simpl in |- *; rewrite H1; rewrite H3; trivial.

    simpl in |- *; rewrite H1; rewrite H3; trivial.

    simpl in |- *; rewrite H1; rewrite H3; trivial.
  Qed.


  Record cterm (n : nat) : Set :=  {ct :> term; clost : closed n ct}.
  Implicit Arguments ct [ n ].
  Implicit Arguments clost [ n ].
  Hint Resolve @clost: core.

  Definition mk_cterm n t :=
    match closed_dec n t as x return if x then cterm n else ~closed n t with
      | left H => Build_cterm n t H
      | right H => H
    end
  .

  Lemma closed0_lift : forall (c : cterm 0) n k, lift_rec n c k = c.
  Proof.
    intros.
    apply closed_lift_rec.
    apply closed_weak with 0; auto with arith.
  Qed.

  Lemma closed1_lift : forall (c : cterm 1) n k, lift_rec n c (S k) = c.
  Proof.
    intros.
    apply closed_lift_rec.
    apply closed_weak with 1; auto with arith.
  Qed.

  Lemma closed2_lift : forall (c : cterm 2) n k, lift_rec n c (S (S k)) = c.
  Proof.
    intros.
    apply closed_lift_rec.
    apply closed_weak with 2; auto with arith.
  Qed.

  Lemma closed0_subst : forall (c : cterm 0) s k, subst_rec s c k = c.
  Proof.
    intros.
    apply closed_subst_rec.
    apply closed_weak with 0; auto with arith.
  Qed.

  Lemma id_closed1 : forall t : cterm 1, subst_rec (Rel 1) t 0 = lift 1 t.
  Proof.
    unfold lift in |- *; intros.
    pattern (t:term) at 2 in |- *.
    elim replace0 with (t:term) 0.
    unfold replace in |- *.
    rewrite closed1_lift.
    rewrite distr_lift_subst.
    rewrite closed1_lift.
    trivial.
  Qed.


  Lemma closed1_subst : forall (c : cterm 1) s k, subst_rec s c (S k) = c.
  Proof.
    intros.
    apply closed_subst_rec.
    apply closed_weak with 1; auto with arith.
  Qed.

  Lemma closed2_subst : forall (c : cterm 2) s k, subst_rec s c (S (S k)) = c.
  Proof.
    intros.
    apply closed_subst_rec.
    apply closed_weak with 2; auto with arith.
  Qed.

  Hint Resolve closed0_lift closed1_lift closed2_lift closed0_subst
    closed1_subst closed2_subst: core.

  Notation "x ,, y" := (Pair0 x y) (at level 100, right associativity).
  Notation "'Forall' x : t1 , t2" := (Prd x t1 t2) (at level 48, x at level 9).
  Notation "'Fun' x : t => m" := (Lam x t m) (at level 48, x at level 9).
  Notation "x ** y" := (Sum x y) (at level 45, right associativity).
  (* { n := t2 } t1 at level 0 because of sumor and sumbool *)
  Notation "t1 [ n := t2 ]" := (subst_rec t2 t1 n)
    (format "t1 [ n := t2 ]", at level 9).
  Notation "^ ( n // k ) t" := (lift_rec n t k)
    (format "^ ( n // k ) t", at level 8).
  Notation "^ ( n ) t" := (lift n t) (format "^ ( n ) t", at level 8).
  Notation "# n" := (Rel n) (at level 2).
  Notation "%% s" := (Srt s) (at level 2).

End ITermes.
