(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

open Kernel
open CicCore
open Libkernel
open Term_printer
open Parser

(*> code noyau *)
let exec_failed() = raise Exec;;

let exec_infer env t =
  match (infer_type env t) with
    Some {coq_VAL=u;coq_TYP=tt} ->
      print_string "Type infere: ";
      print_term [] tt;
      print_newline();
      env
  | _ -> print_endline "Erreur: Infer."; exec_failed()
;;

let exec_check env trm typ =
  match (check_type env trm typ) with
    true ->
      print_endline "Correct.";
      env
  | false -> print_endline "Erreur: Check."; exec_failed()
;;

let exec_decl env x (_,p) a ob =
  let si = Constant {c_par=p;c_decl={dNAME=x;dTYPE=a;dBODY=ob}} in
  if !Clflags.debug then begin
    Printf.printf "{\n  c_par = ";
    print_term [] p;
    let param = find_free_var "param%" [] in
    Printf.printf ";\n  c_decl = {\n    NAME = %s;\n    TYPE = " x;
    print_term [param] a;
    Printf.printf ";\n    BODY = ";
    (match ob with
      | Some ob -> print_term [param] ob;
      | None -> Printf.printf "None");
    Printf.printf ";\n  }\n}\n";
  end;
  match enter_decl env si with
    Some nenv ->
      print_endline (x^(if ob=None then " admis." else " defini."));
      nenv
  | _ -> exec_failed()
;;

let print_one_ind one_ind =
  Printf.printf "{\n  i_name = %s;\n  i_param = " one_ind.i_name;
  print_term [] one_ind.i_param;
  Printf.printf ";\n  i_arg = ";
  let param = find_free_var "param%" [] in
  print_term [param] one_ind.i_arg;
  Printf.printf ";\n  i_ckind = ";
  print_sort one_ind.i_ckind;
  Printf.printf ";\n  i_kind = ";
  print_sort one_ind.i_kind;
  Printf.printf ";\n  i_constrs = ";
  begin match one_ind.i_constrs with
    | Some xs ->
      Printf.printf "[";
      List.iter (fun c ->
        Printf.printf "{\n    c_name = %s;\n    c_arg = " c.c_name;
        let param = find_free_var "param%" [] in
        let arg = find_free_var "arg%" [param] in
        print_term [param] c.c_arg;
        Printf.printf ";\n    c_inst = ";
        print_term [arg; param] c.c_inst;
        Printf.printf ";\n}\n") xs;
      Printf.printf "]"
    | None -> Printf.printf "None"
  end;
  Printf.printf ";\n}\n"

let exec_inductive env ip =
  if !Clflags.debug then List.iter print_one_ind ip;
  match enter_decl env (MutInductive ip) with
    Some nenv ->
      print_endline
 	(string_conc "," (map i_name ip)^" defini(s) inductivement.");
      nenv
  | _ -> exec_failed()
;;

let exec_quit env =
  print_endline "Au revoir..."; exit 0;;

let exec_delete env =
  match env with
    [] -> print_endline "Erreur: environement deja vide."; exec_failed()
  | _::nenv ->
      print_endline "Suppression OK.";
      nenv
;;


let exec_list env =
  let rec it_env = function
      [] -> ()
    | si::l ->
	it_env l;
	print_endline (string_conc " " (names_of si)) in
  it_env env;
  env
;;


let exec_extraction env x =
  match search_const env x with
    | Some {c_decl = {dBODY = Some t}} ->
        Printf.printf "Extraction de %s:\n" x;
        print_term [] (extract_term env [] t);
        print_newline ();
        env
    | _ ->
        Printf.printf "Erreur: %s n'est pas une constante\n%!" x;
        exec_failed ();
;;

let exec_print env x =
  match search_const env x with
    | Some {c_decl = {dBODY = Some t}} ->
        Printf.printf "%s:\n" x;
        print_term [] t;
        print_newline ();
        env
    | _ ->
        Printf.printf "Erreur: %s n'est pas une constante\n%!" x;
        exec_failed ();
;;


let _ENV = ref (nil_genv: sigma);;

let exec_command raw_com =
  let com = raw_com !_ENV in
  let newenv =
    match com with
    | Check(trm, Some typ)   -> exec_check !_ENV trm typ
    | Check(trm, None)       -> exec_infer !_ENV trm
    | Definition ((x,p,t),b) -> exec_decl !_ENV x p t b
    | Inductive ip           -> exec_inductive !_ENV ip
    | Quit                   -> exec_quit !_ENV
    | Delete                 -> exec_delete !_ENV
    | List                   -> exec_list !_ENV
    | Extraction x           -> exec_extraction !_ENV x
    | Print x                -> exec_print !_ENV x
  in _ENV := newenv
;;


(* arret a la premiere erreur *)
let handle_batch_error pos loc err =
  print_endline ("Erreur"^error_string pos (Some loc) err);
  exec_failed()
;;

let run_fail_abort = run parse_cmd exec_command handle_batch_error;;

let use_file f =
  let ic = open_in f in
  try
    let n = run_fail_abort (Stream.of_channel ic) in
    close_in ic;
    n
  with e ->
    close_in ic;
    if e<>Exec then print_endline ("Erreur"^error_string 0 None e);
    exec_failed()
;;

let use_string s =
  run_fail_abort (Stream.of_string s)
;;



(* boucle toplevel:
 *   affiche un prompt,
 *   redemarre apres une erreur
 *)
let prompt() = print_string "Bcoq < "; flush stdout;;

let rec parse_top_com po =
  prompt();
  try parse_cmd po
  with err ->
    let s = "Erreur"^error_string 0 None err in
    match skip_til_dot po with
      Some() -> print_endline s; parse_top_com po
    | None -> None
;;

let handle_top pos loc err =
  match err with
    Sys.Break -> print_newline(); raise Sys.Break
  | Exec -> print_endline "Error while executing toplevel phrase."
  | Failure _ -> print_endline ("Erreur"^error_string pos None err)
  | _ -> raise err
;;

let go() =
  let _ = run parse_top_com exec_command handle_top (Stream.of_channel stdin)
  in print_newline()
;;

let prelude_file = "bcoq/prelude.v";;
let prelude() =
  if Sys.file_exists prelude_file then
    try let _ = use_file prelude_file in ()
    with Exec -> print_endline "Le prelude contient une erreur."
  else print_endline "Prelude introuvable."
;;

(*
if Filename.basename (Sys.argv.(0)) <> "ocaml" then Run.loop();;
*)
