(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

(* camlp4o pa_extend.cmo *)

open Kernel
open CicCore
open Libkernel

let conv_sort gs =
  match sort_of_gen gs with
    (Some s) -> s
  | None -> failwith "sorte incorrecte"
;;


(* Partially applied constants... *)
let mk_APP(a,b) =
  match a with
    OP1(App, PAIR(REL ("@"|"!" as s), appl))
      -> OP1(App, PAIR(REL s, PAIR(appl, b)))
  | _ -> OP1(App, PAIR(a,b));;

let rec dest_APP stk = function
    PAIR(al,an) -> dest_APP (an::stk) al
  | REL x -> (x,stk)
  | _ -> failwith "dest_APP"
;;

let mk_CST x =
  OP1(App,PAIR(REL "@", REL x))
;;
let mk_exp_CST x =
  OP1(App,PAIR(REL "!", REL x))
;;

let dest_CST = function
    REL x -> x (* unused... *)
  | OP1(App,PAIR(REL ("@"|"!"), REL x)) -> x
  | _ -> raise (Stream.Error "( term : ... )")
;;


let rec term_of_expr env ctx = function
    SRT gs -> Srt (conv_sort gs)
  | REL x -> Rel(nat_of_int (int_of_string x))
  | OP1 (App, PAIR(REL ("@"|"!" as exp), appl)) ->
      let (x,args) = dest_APP [] appl in
      globalize1 (exp="!") (env,ctx) x (List.map (term_of_expr env ctx) args)
  | OP1 (App, PAIR(hd, args)) ->
      mk_app (term_of_expr env ctx hd) (term_of_expr env ctx args)
  | OP0 c -> Cst0 c
  | OP1 (c,t) -> Cst1 (c, term_of_expr env ctx t) 
  | LAM (x,tt,t) ->
      Lam (x, term_of_expr env ctx tt, term_of_expr env (x::ctx) t)
  | PRD (x,tt,u) ->
      Prd (x,term_of_expr env ctx tt, term_of_expr env (x::ctx) u)
  | PAIR (u,v) -> Pair0 (term_of_expr env ctx u, term_of_expr env ctx v)
  | SUM (u,v) -> Sum (term_of_expr env ctx u, term_of_expr env ctx v)
;;

let conv_term env t = term_of_expr env [] t;;
let conv_term1 env x t = term_of_expr env [x] t;;




let check_infer e oetyp env =
  let trm = term_of_expr env [] e in
  let otyp =
    match oetyp with
      Some ety -> Some (term_of_expr env [] ety)
    | None -> None
  in Check(trm,otyp)
;;

let conv_spec env (x,(pn,epty),etyp) =
  let pty = term_of_expr env [] epty in
  let typ = term_of_expr env [pn] etyp in
  (x,(pn,pty),typ)
;;

let check_axiom sp env =
  let csp = conv_spec env sp in
  Definition (csp, None)
;;

let check_def sp body env =
  let (_,(pn,_),_) as csp = conv_spec env sp in
  let bd = term_of_expr env [pn] body in
  Definition (csp, Some bd)
;;

(* Checking inductive declarations *)
let split_constr_ty env ctx t =
  let rec split stk = function
      PRD("_",ty,rng) -> split (term_of_expr env ctx ty::stk) rng
    | t -> (List.fold_left (fun t ty -> Sum(ty,t)) (Cst0 Lmark) stk,
	    term_of_expr env ("_"::ctx) t)
  in match t with
    PRD(x,(OP1(Record _,_) as arg),t) ->
	(term_of_expr env ctx arg, term_of_expr env (x::ctx) t)
  | _ -> split [] t
;;

let conv_constr env inm pn (c,tye) =
  match split_constr_ty env [pn] tye with
    (args, Cst1(MutInd(i,l), Pair0(_, Pair0(inst,Cst0 Nil_mark))))
    when i=inm -> {c_name=c; c_arg=args; c_inst=inst}
  | _ -> failwith "bad constructor type"
;;

let conv_one_ind env (i,(pn,ptye),ar,osc,cl) =
  let pty = term_of_expr env [] ptye in
  let (args,s) = split_arity (term_of_expr env [pn] ar) in
  let sc = match osc with
    Some sc -> conv_sort sc
  | None -> s in
  ({i_name=i;i_param=pty;i_arg=args;i_ckind=sc;i_kind=s;i_constrs=None},
       (pn,cl))
;;

let conv_ind_pack nenv ip =
  let mk_is (is,(pn,cl)) =
    let ccl = map (conv_constr nenv (i_name is) pn) cl in
    define_constr is ccl
  in (map mk_is ip)
;;

let check_inductive indp env =
  let abs_ip = map (conv_one_ind env) indp in
  let nenv = (MutInductive (map fst abs_ip))::env in
  Inductive (conv_ind_pack nenv abs_ip)
;;


let mk_one = OP0 Lmark;;
let mk_unit = OP0 Nil_mark;;

let rec eproj e = function
    0 -> OP1(Proj1, e)
  | n -> eproj (OP1(Proj2, e)) (n-1)
;;



(* *)
module Gram = Grammar.GMake(Lexer.L);;

let ident  = Gram.Entry.create "ident";;
let lident = Gram.Entry.create "lident";;
let kind   = Gram.Entry.create "kind";;
let expr   = Gram.Entry.create "expr";;
let term   = Gram.Entry.create "term";;

let one_ind = Gram.Entry.create "one_ind";;
let command = Gram.Entry.create "command";;
let command_opt = Gram.Entry.create "command_opt";;


GEXTEND Gram
  GLOBAL: ident lident kind expr term one_ind command command_opt;

  ident:
    [ [ x = LIDENT -> x
      |	"_"        -> "_" ] ];

  lident:
    [ [ x = ident; ","; l = lident -> x::l
      | x = ident -> [x] ] ];
(*
  lident0:
    [ [ l = lident -> l
      | -> Nil ] ];
*)

  (* Sorts *)
  univ:
    [ [ "("; n = INT; ")" -> (nat_of_int (int_of_string n))
      | -> O ] ];

  kind:
    [ [ "Prop" -> Gprop
      | "Set"  -> Gset
      |	"Type"; u = univ -> Gtype u
      |	"Typeset"; u = univ -> Gtypeset u ] ];


  (* Records *)
  record_arity:
    [ [ (x,ty) = binder; ";"; (fl,ra) = record_arity
          -> (x::fl, SUM(ty,PRD(x,ty,ra)))
      | (x,ty) = binder -> ([x], SUM(ty,PRD(x,ty,mk_one)))
      |	-> ([], mk_one) ] ];

  record_struct:
    [ [ (x,ty,v) = definition; ";"; (fl,ar,vls) = record_struct
          -> (x::fl, PRD(x,ty,ar), PAIR(v,vls))
      | (x,ty,v) = definition
          -> ([x], PRD(x,ty,mk_one), PAIR(v, mk_unit))
      | -> ([], mk_one, mk_unit) ] ];

  (* Inductive types *)
  mark_level:
    [ [ ";"; l = INT -> nat_of_int (int_of_string l)
      | -> O ] ];


  (* Cases branches *)
  branch:
    [ [ x = ident; "=>"; e = expr -> (x,e) ] ];
  branchs1:
    [ [ (p,b) = branch; "|"; (pats,brs) = branchs1
        -> (p::pats, PAIR(b,brs))
      | (p,b) = branch -> ([p], PAIR(b, mk_unit)) ] ];
  branchs:
    [ [ p = branchs1 -> p
      | -> ([], mk_unit) ] ];


  (* Fixpoints *)
  fix_def:
    [ [ (pn,pty) = param; nty = expr; ":>"; ty = expr; ":="; bd = expr
 	-> (PRD(pn,pty,ty), PRD(pn,pty,nty), LAM(pn,pty,bd)) ] ];
  fixbodies:
    [ [ (ty,nty,bd) = fix_def; "with"; (ar,nar,bds) = fixbodies
 	-> (SUM(ty,ar),SUM(nty,nar),PAIR(bd,bds))
      | (ty,nty,bd) = fix_def
 	-> (SUM(ty, mk_one), SUM(nty, mk_one),
	    PAIR(bd, mk_unit)) ] ];
  fixpoint:
    [ [ m = ident; ","; f = ident; "|"; (ar,nar,bd) = fixbodies
	  -> PAIR(ar, LAM(m, OP0 Mark, LAM(f,nar,bd))) ] ];


  (* Binders *)
  definition:
    [ [ x = ident; ":"; ty = expr; ":="; v = expr -> (x,ty,v)
      | x = ident; ":="; v = expr; ":"; ty = expr -> (x,ty,v) ] ];
  binder:
    [ [ x = ident; ":"; e = expr -> (x,e) ] ];
  mbinder:
    [ [ lv = lident; ":"; e = expr -> (lv,e) ] ];
  param:
    [ [ "["; b = binder; "]" -> b
      | -> ("_", mk_one) ] ];

  (* Terms *) 
  expr:
    [ LEFTA
      [ e1 = expr; e2 = expr -> mk_APP(e1,e2) ]

    | "default"
      RIGHTA
      [ "["; (lid,typ) = mbinder; "]"; trm = expr
	  -> List.fold_right (fun x t -> LAM (x,typ,t)) lid trm
      | "let"; (x,typ) = binder; ":="; arg = expr; "in"; trm = expr
	  -> mk_APP (LAM (x,typ,trm), arg) ]

    | RIGHTA
      [ e1 = expr; "->"; e2 = expr -> PRD("_",e1,e2) ]

    | RIGHTA
      [ e1 = expr; "*"; e2 = expr -> SUM(e1,e2) ]

    | "atom"
      [	op = oper1; e = expr -> OP1(op,e)
      |	x = ident -> mk_CST x
      |	"!"; x = ident -> mk_exp_CST x
      | s = kind -> SRT s
      |	"Mark" -> OP0 Mark
      |	"%" -> OP0 Mt_mark
      |	"Lmark" -> OP0 Lmark
      |	"~" -> OP0 Nil_mark
      |	"+"; i = INT; e = expr -> OP1(Nil_mark, PRD(i,e,OP0 Nil_mark))
      |	"+"; e = expr -> OP1(Mark,e)
      |	"-"; e = expr -> OP1(Lmark,e)
      |	"#"; i = INT -> REL i
      |	e = expr; "^"; n = INT -> eproj e (int_of_string n)
      | e = expr; "^"; x = ident -> OP1(Field x, e)
      |	e = expr; "^"; "^"; INT "1" -> OP1(Proj1, e)
      |	e = expr; "^"; "^"; INT "2" -> OP1(Proj2, e)
      | "{<"; (flds, ra) = record_arity; ">}" -> OP1(Record flds, ra)
      |	"{"; (flds, pa, vls) = record_struct; "}"
 	  -> OP1(Struct flds, PAIR(pa,vls))
      |	"Fix"; "{"; p = fixpoint; "}" -> OP1(Fixp, p)
      |	"<"; q = expr; ">"; "Cases"; e = expr; "of"; OPT [ "|" ];
	(pats,brs) = branchs; "end"
	  -> OP1(MutCase (true, pats), PAIR(e, PAIR(q, brs)))
      |	"("; e1 = expr; ","; e2 = expr; ")" -> PAIR(e1,e2) 
      |	"("; v1 = expr; ","; v2 = expr; ","; lid = lident; ":"; f = end_product
          -> f ((dest_CST v1)::(dest_CST v2)::lid)
      | "("; v1 = expr; ","; v2 = expr; ":"; f = end_product
          -> f [dest_CST v1; dest_CST v2]
      | "("; v = expr; ":"; f = end_product -> f [dest_CST v]
      | "("; e = expr; ")" -> e ] ];

  end_product:
    [ [ typ = expr; ")"; rng = expr LEVEL "default"
          -> (fun lid -> List.fold_right (fun x t -> PRD (x,typ,t)) lid rng) ] ];

  oper1:
    [ [	"Cst"; "{"; x = ident; "}" -> Const x
      | "Ind"; "{"; i = ident; l = mark_level; "}" -> MutInd(i, l)
      | "Cstr"; "{"; c = ident; "}" -> MutConstr c ] ];


  (* The entry point for terms *)
  term:
    [ [ t = expr LEVEL "default" -> t ] ];



  (* Toplevel inductive declarations *)
  constructor:
    [ [ c = ident; ":"; ty = term -> (c, ty) ] ];
  constructors1:
    [ [ c = constructor; "|"; lc = constructors -> c::lc
      | c = constructor -> [c] ] ];
  constructors:
    [ [ lc = constructors1 -> lc
      | -> [] ] ];

  inductive_tok:
    [ [ OPT [ "Mutual" ]; "Inductive" ] ];
  constr_kind:
    [ [ "/"; s = kind -> Some s
      | -> None ] ];
  ind_defs:
    [ [ is = one_ind; "with"; li = ind_defs -> is::li
      | is = one_ind -> [is] ] ];
  one_ind:
    [ [ i = ident; par = param; ":"; ar = term;
 	osc = constr_kind; ":="; cl = constructors -> (i,par,ar,osc,cl) ] ];
	


  spec:
    [ [ x = ident; par = param; ":" ; typ = term -> (x,par,typ) ] ];

  typ_cstr:
    [ [ "::"; typ = term -> Some typ
      | -> None ] ];

  const_tok:
    [ [ "Definition" | "Lemma" ] ];
  def_sep:
    [ [ ":=" | "."; "Proof" ] ];

  command:
    [ [ "Infer"; trm = term; "." -> check_infer trm None
      | "Check"; trm = term ; otyp = typ_cstr; "." -> check_infer trm otyp
      | "Extraction"; x = ident; "." -> (fun _ -> Extraction x)
      | "Print"; x = ident; "." -> (fun _ -> Print x)
      | "Axiom"; sp = spec; "." -> check_axiom sp
      | const_tok; sp = spec; def_sep; body = term; "." -> check_def sp body 
      | inductive_tok; indp = ind_defs; "." -> check_inductive indp
      | "Quit"; "." -> (fun _ -> Quit)
      | "Delete"; "." -> (fun _ -> Delete)
      | "List"; "." -> (fun _ -> List) ] ];

  command_opt:
    [ [ c = command -> Some (loc,c)
      | EOI -> None ] ];

END;;



let parse_cmd pobj = Gram.Entry.parse command_opt pobj;;


let run parse_fun exec_fun err_handler cs =
  let ts = Gram.parsable cs in
  let rec parse_exec_com cnt =
    let pos = Stream.count cs in
    match parse_fun ts with
      Some(loc,rawcom) ->
	let ncnt =
	  try
	    let () = exec_fun rawcom in
	    (cnt+1)
	  with err -> let () = err_handler pos loc err in cnt
	in parse_exec_com ncnt
    | None -> cnt
  in parse_exec_com 0
;;



let loc_string pos = function
    Some loc ->
      let b = Ploc.first_pos loc in
      let e = Ploc.last_pos loc in
      " en ("^string_of_int (b-pos)^","^string_of_int (e-pos)^")"
  | None -> ""
;;

exception Exec;;

let rec error_string pos oloc = function
    Ploc.Exc(loc,e) -> error_string pos (Some loc) e
  | Failure s -> loc_string pos oloc^": "^s
  | Token.Error s -> " de lexeme"^loc_string pos oloc^": "^s
  | Stream.Error s -> " de syntaxe"^loc_string pos oloc^": "^s
  | Exec -> " pendant l'execution de la commande"^loc_string pos oloc^"."
  | e -> raise e
;;

(* toplevel *)
let skip_til_dot =
  let rec skip_tok = parser
      [< '("",".") >] -> Some()
    | [< '("EOI","") >] -> None
    | [< '_ ; strm >] -> skip_tok strm in
  let e = Gram.Entry.of_parser "skipping" skip_tok in
  Gram.Entry.parse e
;;
