(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

open Kernel
open CicCore
open Libkernel

let rec var_indep x = function
   SRT _ -> true
 | REL y -> x<>y
 | OP0 _ -> true
 | OP1 (c,t) -> (var_indep x t)
 | LAM (y,tt,t) -> (var_indep x tt) & (x=y or (var_indep x t))
 | PRD (y,tt,u) -> (var_indep x tt) & (x=y or (var_indep x u))
 | PAIR (t,u) -> (var_indep x t) & (var_indep x u)
 | SUM (t,u) -> (var_indep x t) & (var_indep x u)
;;


let expr_of_term ctx t = 
 let rec exp_of_trm ctx = function
    Srt s -> SRT (gen_of_sort s)
  | Rel n -> (match list_item ctx n with
                Some x -> REL x
              | None -> assert false)
  | Cst0 c -> OP0 c
  | Cst1 (c,t) -> OP1(c, exp_of_trm ctx t)
  | Lam (x,tt,t)-> let v = find_free_var x ctx in
                 LAM (v, exp_of_trm ctx tt, exp_of_trm (v::ctx) t)
  | Prd (x,tt,u)-> let v = find_free_var x ctx in
                 PRD (v, exp_of_trm ctx tt, exp_of_trm (v::ctx) u)
  | Pair0 (u,v)-> PAIR (exp_of_trm ctx u, exp_of_trm ctx v)
  | Sum (u,v)-> SUM (exp_of_trm ctx u, exp_of_trm ctx v) in
 exp_of_trm ctx t
;;


(*> affichage *)

let string_of_sort = function
  Gprop      -> "Prop"
| Gset       -> "Set"
| Gtype n    -> "Type(" ^ (Big_int.string_of_big_int (int_of_nat n)) ^ ")"
| Gtypeset n -> "Typeset(" ^ (Big_int.string_of_big_int (int_of_nat n)) ^ ")"
;;

let rec string_of_lid = function
    [] -> ""
  | [x] -> x
  | x::l -> x^","^string_of_lid l
;;

let string_of_op = function
    Mark -> "Mark"
  | Mt_mark -> "#"
  | Lmark -> "Lmark"
  | Nil_mark -> "~"
  | Cons_mark -> "::"
  | Proj1 -> "P1 "
  | Proj2 -> "P2 "
  | Const x -> "Cst{"^x^"}"
  | MutInd (i,l) -> "Ind{"^i^";"^(Big_int.string_of_big_int (int_of_nat l))^"}"
  | MutConstr c -> "Cstr{"^c^"}"
  | MutCase (_, p) -> "Case{"^string_conc "|" p^"}"
  | Fixp -> "Fix"
  | Struct flds -> "Struct{"^string_of_lid flds^"}"
  | _ -> failwith "Not a printable Op"
;;



let rec string_of_expr = function
    SRT s -> string_of_sort s
  | REL x -> x
  | OP0 c -> string_of_op c
  | OP1 (c,arg) -> print_op1 c arg
  | LAM (x,tt,t) -> "["^x^":"^(string_of_expr tt)^"]"^(string_of_expr t)
  | PRD (x,tt,u) -> if var_indep x u
                     then (string_of_arrow tt)^"->"^(string_of_expr u)
                     else "("^x^":"^(string_of_expr tt)^")"^(string_of_expr u)
  | PAIR (u,v) -> "("^(string_of_app u)^","^(string_of_app v)^")"
  | SUM (u,v) -> (string_of_sum u)^"*"^(string_of_sum v)

and print_op1 op arg =
  match (op, arg) with
    (Const c, OP0 Nil_mark) -> c
  | (MutInd(i,O), PAIR(OP0 Nil_mark, PAIR(OP0 Nil_mark, OP0 Nil_mark)))
      -> i
  | (App, PAIR(u,v)) -> "("^(string_of_app u)^" "^(string_of_expr v)^")"
  | (Field x, t) -> (string_of_field t)^"^"^x
  | (Record flds, t) -> "{"^string_of_lid flds^":"^string_of_expr t^"}"
  | (Struct flds, PAIR(ar,v)) ->
      "{"^string_of_lid flds^":"^string_of_expr ar^":="^string_of_expr v^"}"
  | _ -> (string_of_op op)^(string_of_expr arg)

and string_of_field = function
    (SRT _ | REL _ | OP0 _ | OP1(Record _,_) | OP1(Struct _,_) |
    OP1(Fixp,_)) as t
    -> string_of_expr t
  | t -> "("^string_of_expr t^")"

and string_of_sum = function
    (LAM _ | PRD _ | SUM _) as t -> "("^(string_of_expr t)^")"
  | t -> string_of_expr t

and string_of_app = function
    OP1(App, PAIR(u,v)) -> (string_of_app u)^" "^(string_of_expr v)
  | t -> string_of_expr t

and string_of_arrow = function
    (LAM _ | PRD _) as t -> "("^(string_of_expr t)^")"
  | t -> string_of_expr t
;;

let print_sort s = print_string (string_of_sort (gen_of_sort s));;
let print_term ctx t=print_string (string_of_expr (expr_of_term ctx t));;

