(**************************************************************************)
(*                       The Extracted Coq subsytem                       *)
(*                                                                        *)
(*                            Stéphane Glondu                             *)
(*                                                                        *)
(*                           © 2007-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

open Kernel
open Names
open Term

let rec lazy_assoc key = function
  | [] -> None
  | (lazy k, v)::xs -> if k = key then Some v else lazy_assoc key xs

let int_of_bool x = if x then 1 else 0

let string_of_char_list xs =
  let res = String.create (List.length xs) in
  let i = ref 0 in
  let put c = res.[!i] <- c; incr i in
  List.iter put xs; res

let char_list_of_string s =
  let rec aux i accu =
    if i < 0 then accu
    else aux (i-1) (s.[i]::accu)
  in aux (String.length s - 1) []

let get_construct m x = match Coqlib.find_reference "Xcoq" m x with
  | Libnames.ConstructRef cp -> cp
  | _ -> raise Not_found

let get_ind m x = match Coqlib.find_reference "Xcoq" m x with
  | Libnames.IndRef i -> i
  | _ -> raise Not_found

let get_constant m x = match Coqlib.find_reference "Xcoq" m x with
  | Libnames.ConstRef c -> c
  | _ -> raise Not_found

type 'a constr = {
  interp : Term.constr -> 'a option;
  reify : 'a -> Term.constr;
  constr : Term.constr Lazy.t;
}

type 'a desc = (constructor Lazy.t * ((Term.constr -> Obj.t option) -> Term.constr list -> 'a option)) list

let interp_desc (desc : 'a desc) c =
  let rec aux_app c args = match kind_of_term c with
    | App (f, a) -> aux_app f (Array.to_list a @ args)
    | Construct cp -> aux_cons cp args
    | _ -> None
  and aux_cons cp args = match lazy_assoc cp desc with
    | Some f ->
      let ip, j = cp in
      let nparam, _ = Inductiveops.inductive_nargs (Global.env ()) ip in
      let rec aux args n =
        if n <= 0 then f (fun c -> Obj.magic (aux_app c [])) args
        else (match args with [] -> None | _::xs -> aux xs (n-1))
      in aux args nparam
    | _ -> None
  in aux_app c []

let coq_Init_Datatypes = ["Coq"; "Init"; "Datatypes"]

(* type nat = O | S of nat *)

let nat =
  let get = get_construct coq_Init_Datatypes in
  let cO = lazy (get "O") and cS = lazy (get "S") in {
    interp =
      interp_desc [
        cO, (fun _ -> function [] -> Some O | _ -> None);
        cS, (fun interp -> function
          | [p] ->
            begin match interp p with
              | Some p -> Some (S (Obj.obj p))
              | _ -> None
            end
          | _ -> None
        );
      ];
    reify =
      begin
        let rec reify = function
          | O -> mkConstruct (Lazy.force cO)
          | S p -> mkApp (mkConstruct (Lazy.force cS), [| reify p |])
        in reify
      end;
    constr = lazy (mkInd (get_ind coq_Init_Datatypes "nat"));
  }

(* type bool = true | false *)

let bool =
  let get = get_construct coq_Init_Datatypes in
  let ctrue = lazy (get "true") and cfalse = lazy (get "false") in {
    interp =
      interp_desc [
        ctrue, (fun _ -> function [] -> Some true | _ -> None);
        cfalse, (fun _ -> function [] -> Some false | _ -> None);
      ];
    reify =
      begin fun x ->
        if x then mkConstruct (Lazy.force ctrue)
        else mkConstruct (Lazy.force cfalse)
      end;
    constr = lazy (mkInd (get_ind coq_Init_Datatypes "bool"))
  }

(* type ascii = Ascii of bool * bool * bool * bool * bool * bool * bool * bool *)

let ascii =
  let mp = ["Coq"; "Strings"; "Ascii"] in
  let get = get_construct mp in
  let cAscii = lazy (get "Ascii") in {
    interp =
      interp_desc [
        cAscii, (
          let rec aux res = function
            | [] -> Some (char_of_int res)
            | b::bs -> match bool.interp b with
            | Some b -> aux ((res lsl 1) lor (int_of_bool b)) bs
            | None -> None
          in fun _ bs -> if List.length bs = 8 then aux 0 (List.rev bs) else None
        );
      ];
    reify =
      begin
        let rec reify res c n =
          if n = 0 then
            mkApp (mkConstruct (Lazy.force cAscii), Array.of_list (List.rev res))
          else
            let b = bool.reify (c land 1 <> 0) in
            reify (b::res) (c lsr 1) (n-1)
        in fun c -> reify [] (int_of_char c) 8
      end;
    constr = lazy (mkInd (get_ind mp "ascii"));
  }

(* type string = EmptyString | String of ascii * string *)

let string =
  let mp = ["Coq"; "Strings"; "String"] in
  let get = get_construct mp in
  let cEmptyString = lazy (get "EmptyString") and cString = lazy (get "String") in {
    interp =
      begin
        let aux = interp_desc [
          cEmptyString, (fun _ -> function [] -> Some [] | _ -> None);
          cString, (fun interp -> function
            | [x; xs] ->
              begin match ascii.interp x, interp xs with
                | Some x, Some xs -> Some (x::(Obj.obj xs))
                | _, _ -> None
              end
            | _ -> None
          )
        ] in
        fun c -> match aux c with
          | Some xs -> Some (string_of_char_list xs)
          | None -> None
      end;
    reify =
      begin
        let rec aux = function
          | [] -> mkConstruct (Lazy.force cEmptyString)
          | c::cs -> mkApp (mkConstruct (Lazy.force cString), [| c; aux cs |])
        in fun s ->
          let xs = char_list_of_string s in
          aux (List.map ascii.reify xs)
      end;
    constr = lazy (mkInd (get_ind mp "string"));
  }

let list a =
  let clist = lazy (mkInd (get_ind coq_Init_Datatypes "list")) in
  let get = get_construct coq_Init_Datatypes in
  let cnil = lazy (get "nil") and ccons = lazy (get "cons") in {
    interp =
      interp_desc [
        cnil, (fun _ -> function [] -> Some [] | _ -> None);
        ccons, (fun interp -> function
          | [x; xs] ->
            begin match a.interp x, interp xs with
              | Some x, Some xs -> Some (x::(Obj.obj xs))
              | _, _ -> None
            end
          | _ -> None
        )
      ];
    reify =
      begin
        let rec reify = function
          | [] -> mkApp (mkConstruct (Lazy.force cnil), [| Lazy.force a.constr |])
          | x::xs -> mkApp (mkConstruct (Lazy.force ccons), [| Lazy.force a.constr; a.reify x; reify xs |])
        in reify
      end;
    constr = lazy (mkApp (Lazy.force clist, [| Lazy.force a.constr |]));
  }

let option a =
  let coption = lazy (mkInd (get_ind coq_Init_Datatypes "option")) in
  let get = get_construct coq_Init_Datatypes in
  let cNone = lazy (get "None") and cSome = lazy (get "Some") in {
    interp =
      interp_desc [
        cNone, (fun _ -> function [] -> Some None | _ -> None);
        cSome, (fun interp -> function
          | [x] ->
            begin match a.interp x with
              | Some x -> Some (Some x)
              | _ -> None
            end
          | _ -> None
        )
      ];
    reify =
      begin
        let rec reify = function
          | None -> mkApp (mkConstruct (Lazy.force cNone), [| Lazy.force a.constr |])
          | Some x -> mkApp (mkConstruct (Lazy.force cSome), [| Lazy.force a.constr; a.reify x |])
        in reify
      end;
    constr = lazy (mkApp (Lazy.force coption, [| Lazy.force a.constr |]));
  }

let (>>=) x f = match x with Some y -> f y | None -> None

open Kernel
open CicCore

let ml_string =
  let mp = ["ExtractionInCoq"; "Library"; "MlString"] in
  let get = get_construct mp in
  let cmk_str = lazy (get "mk_str") in {
    interp =
      interp_desc [
        cmk_str, (fun _ -> function
          | [s] -> (string.interp s >>= fun s -> Some s)
          | _ -> None);
      ];
    reify =
      (fun s -> mkApp (mkConstruct (Lazy.force cmk_str), [| string.reify s |]));
    constr = lazy (mkInd (get_ind mp "ml_string"));
  }

let list_string = list ml_string

let oper =
  let mp = ["ExtractionInCoq"; "CicMeta"; "CicCore"; "CicCore"] in
  let get = get_construct mp in
  let cMark = lazy (get "Mark") and
      cMt_mark = lazy (get "Mt_mark") and
      cLmark = lazy (get "Lmark") and
      cNil_mark = lazy (get "Nil_mark") and
      cCons_mark = lazy (get "Cons_mark") and
      cApp = lazy (get "App") and
      cProj1 = lazy (get "Proj1") and
      cProj2 = lazy (get "Proj2") and
      cStruct = lazy (get "Struct") and
      cRecord = lazy (get "Record") and
      cField = lazy (get "Field") and
      cConst = lazy (get "Const") and
      cMutInd = lazy (get "MutInd") and
      cMutConstr = lazy (get "MutConstr") and
      cMutCase = lazy (get "MutCase") and
      cFixp = lazy (get "Fixp")
  in {
    interp =
      interp_desc [
        cMark, (fun _ -> function [] -> Some Mark | _ -> None);
        cMt_mark, (fun _ -> function [] -> Some Mt_mark | _ -> None);
        cLmark, (fun _ -> function [] -> Some Lmark | _ -> None);
        cNil_mark, (fun _ -> function [] -> Some Nil_mark | _ -> None);
        cCons_mark, (fun _ -> function [] -> Some Cons_mark | _ -> None);
        cApp, (fun _ -> function [] -> Some App | _ -> None);
        cProj1, (fun _ -> function [] -> Some Proj1 | _ -> None);
        cProj2, (fun _ -> function [] -> Some Proj2 | _ -> None);
        cStruct, (fun _ -> function
          | [s] -> list_string.interp s >>= fun s -> Some (Struct s)
          | _ -> None);
        cRecord, (fun _ -> function
          | [s] -> list_string.interp s >>= fun s -> Some (Record s)
          | _ -> None);
        cField, (fun _ -> function
          | [s] -> ml_string.interp s >>= fun s -> Some (Field s)
          | _ -> None);
        cConst, (fun _ -> function
          | [s] -> ml_string.interp s >>= fun s -> Some (Const s)
          | _ -> None);
        cMutInd, (fun _ -> function
          | [i; n] ->
            ml_string.interp i >>= fun i ->
            nat.interp n >>= fun n ->
            Some (MutInd (i, n))
          | _ -> None);
        cMutConstr, (fun _ -> function
          | [s] -> ml_string.interp s >>= fun s -> Some (MutConstr s)
          | _ -> None);
        cMutCase, (fun _ -> function
          | [b; p] ->
            bool.interp b >>= fun b ->
            list_string.interp p >>= fun p ->
            Some (MutCase (b, p))
              | _ -> None);
        cFixp, (fun _ -> function [] -> Some Fixp | _ -> None);
      ];
    reify =
      begin function
        | Mark -> mkConstruct (Lazy.force cMark)
        | Mt_mark -> mkConstruct (Lazy.force cMt_mark)
        | Lmark -> mkConstruct (Lazy.force cLmark)
        | Nil_mark -> mkConstruct (Lazy.force cNil_mark)
        | Cons_mark -> mkConstruct (Lazy.force cCons_mark)
        | App -> mkConstruct (Lazy.force cApp)
        | Proj1 -> mkConstruct (Lazy.force cProj1)
        | Proj2 -> mkConstruct (Lazy.force cProj2)
        | Struct xs -> mkApp (mkConstruct (Lazy.force cStruct), [| list_string.reify xs |])
        | Record xs -> mkApp (mkConstruct (Lazy.force cRecord), [| list_string.reify xs |])
        | Field x -> mkApp (mkConstruct (Lazy.force cField), [| ml_string.reify x |])
        | Const x -> mkApp (mkConstruct (Lazy.force cConst), [| ml_string.reify x |])
        | MutInd (x, n) -> mkApp (mkConstruct (Lazy.force cMutInd), [| ml_string.reify x; nat.reify n |])
        | MutConstr x -> mkApp (mkConstruct (Lazy.force cMutConstr), [| ml_string.reify x |])
        | MutCase (b, p) -> mkApp (mkConstruct (Lazy.force cMutCase), [| bool.reify b; list_string.reify p |])
        | Fixp -> mkConstruct (Lazy.force cFixp)
      end;
    constr = lazy (mkInd (get_ind mp "cci_op"));
  }

let calc =
  let mp = ["ExtractionInCoq"; "CicMeta"; "SortV6"] in
  let get = get_construct mp in
  let cPos = lazy (get "Pos") and cNeg = lazy (get "Neg") in {
    interp =
      interp_desc [
        cPos, (fun _ -> function [] -> Some Pos | _ -> None);
        cNeg, (fun _ -> function [] -> Some Neg | _ -> None);
      ];
    reify =
      begin function
        | Pos -> mkConstruct (Lazy.force cPos)
        | Neg -> mkConstruct (Lazy.force cNeg)
      end;
    constr = lazy (mkInd (get_ind mp "calc"));
  }

let sort =
  let mp = ["ExtractionInCoq"; "CicMeta"; "SortV6"] in
  let get = get_construct mp in
  let cSprop = lazy (get "Sprop") and cStype = lazy (get "Stype") in {
    interp =
      interp_desc [
        cSprop, (fun _ -> function
          | [c] -> calc.interp c >>= fun c -> Some (Sprop c)
          | _ -> None);
        cStype, (fun _ -> function
          | [n] -> nat.interp n >>= fun n -> Some (Stype n)
          | _ -> None);
      ];
    reify =
      begin function
        | Sprop c -> mkApp (mkConstruct (Lazy.force cSprop), [| calc.reify c |])
        | Stype n -> mkApp (mkConstruct (Lazy.force cStype), [| nat.reify n |])
      end;
    constr = lazy (mkInd (get_ind mp "srt_v6"));
  }

let term =
  let mp = ["ExtractionInCoq"; "CicMeta"; "CicCore"; "CicCore"] in
  let get = get_construct mp in
  let cSrt = lazy (get "Srt") and
      cRel = lazy (get "Rel") and
      cCst0 = lazy (get "Cst0") and
      cCst1 = lazy (get "Cst1") and
      cPrd = lazy (get "Prd") and
      cLam = lazy (get "Lam") and
      cSum = lazy (get "Sum") and
      cPair0 = lazy (get "Pair0")
  in {
    interp =
      interp_desc [
        cSrt, (fun _ -> function
          | [s] -> sort.interp s >>= fun s -> Some (Srt s)
          | _ -> None);
        cRel, (fun _ -> function
          | [n] -> nat.interp n >>= fun n -> Some (Rel n)
          | _ -> None);
        cCst0, (fun _ -> function
          | [o] -> oper.interp o >>= fun o -> Some (Cst0 o)
          | _ -> None);
        cCst1, (fun interp -> function
          | [o; t] ->
            oper.interp o >>= fun o ->
            interp t >>= fun t ->
            Some (Cst1 (o, Obj.obj t))
          | _ -> None);
        cPrd, (fun interp -> function
          | [x; t; u] ->
            ml_string.interp x >>= fun x ->
            interp t >>= fun t ->
            interp u >>= fun u ->
            Some (Prd (x, Obj.obj t, Obj.obj u))
          | _ -> None);
        cLam, (fun interp -> function
          | [x; t; u] ->
            ml_string.interp x >>= fun x ->
            interp t >>= fun t ->
            interp u >>= fun u ->
            Some (Lam (x, Obj.obj t, Obj.obj u))
          | _ -> None);
        cSum, (fun interp -> function
          | [t; u] ->
            interp t >>= fun t ->
            interp u >>= fun u ->
            Some (Sum (Obj.obj t, Obj.obj u))
          | _ -> None);
        cPair0, (fun interp -> function
          | [t; u] ->
            interp t >>= fun t ->
            interp u >>= fun u ->
            Some (Pair0 (Obj.obj t, Obj.obj u))
          | _ -> None);
      ];
    reify =
      begin
        let rec reify = function
          | Srt s -> mkApp (mkConstruct (Lazy.force cSrt), [| sort.reify s |])
          | Rel n -> mkApp (mkConstruct (Lazy.force cRel), [| nat.reify n |])
          | Cst0 o -> mkApp (mkConstruct (Lazy.force cCst0), [| oper.reify o |])
          | Cst1 (o, t) -> mkApp (mkConstruct (Lazy.force cCst1), [| oper.reify o; reify t |])
          | Prd (x, t, u) -> mkApp (mkConstruct (Lazy.force cPrd), [| ml_string.reify x; reify t; reify u |])
          | Lam (x, t, u) -> mkApp (mkConstruct (Lazy.force cLam), [| ml_string.reify x; reify t; reify u |])
          | Sum (t, u) -> mkApp (mkConstruct (Lazy.force cSum), [| reify t; reify u |])
          | Pair0 (t, u) -> mkApp (mkConstruct (Lazy.force cPair0), [| reify t; reify u |])
        in reify
      end;
    constr = lazy (mkInd (get_ind mp "term"));
  }

let cterm =
  let mp = ["ExtractionInCoq"; "CicMeta"; "CicCore"; "CicCore"] in
  let cmk_cterm = lazy (mkConst (get_constant mp "mk_cterm")) in
  let ccterm = lazy (mkInd (get_ind mp "cterm")) in
  let cBuild_cterm = lazy (get_construct mp "Build_cterm") in
  fun n ->
    let n' = lazy (nat.reify n) in
    let cctermn = lazy (mkApp (Lazy.force ccterm, [| Lazy.force n' |])) in
    {
      interp =
        interp_desc [
          cBuild_cterm, (fun _ -> function
            | [ct; _] -> term.interp ct >>= fun t -> Some t
            | _ -> None);
      ];
      reify =
        begin fun t ->
          mkCast (
            mkApp (Lazy.force cmk_cterm, [| Lazy.force n'; term.reify t |]),
            VMcast,
            Lazy.force cctermn
          )
        end;
      constr = cctermn;
    }

let cdecl =
  let mp = ["ExtractionInCoq"; "CicMeta"; "CicCore"; "CicCore"] in
  let ccdecl = lazy (mkInd (get_ind mp "cdecl")) in
  let cBuild_cdecl = lazy (get_construct mp "Build_cdecl") in
  fun n ->
    let n' = lazy (nat.reify n) in
    let ctermn = cterm n in
    let ctermn_option = option ctermn in
    let ccdecln = lazy (mkApp (Lazy.force ccdecl, [| Lazy.force n' |])) in
    {
      interp =
        interp_desc [
          cBuild_cdecl, (fun _ -> function
            | [n; t; b] ->
              ml_string.interp n >>= fun n ->
              ctermn.interp t >>= fun t ->
              ctermn_option.interp b >>= fun b ->
              Some {
                dNAME = n;
                dTYPE = t;
                dBODY = b;
              }
            | _ -> None);
      ];
      reify =
        begin fun d ->
          mkApp (
            mkConstruct (Lazy.force cBuild_cdecl),
            [|
              Lazy.force n';
              ml_string.reify d.dNAME;
              ctermn.reify d.dTYPE;
              ctermn_option.reify d.dBODY;
            |]
          )
        end;
      constr = ccdecln;
    }

let ml_string =
  let mp = ["ExtractionInCoq"; "Library"; "MlString"] in
  let cmk_str = lazy (get_construct mp "mk_str") in
  {
    interp =
      interp_desc [
        cmk_str, (fun _ -> function
          | [s] -> string.interp s >>= fun s -> Some s
          | _ -> None);
      ];
    reify =
      begin fun s ->
        mkApp (
          mkConstruct (Lazy.force cmk_str),
          [| string.reify s |]
        )
      end;
    constr = lazy (mkInd (get_ind mp "ml_string"));
  }

let cterm0 = cterm O
let cterm1 = cterm (S O)
let cterm2 = cterm (S (S O))

let constructor =
  let mp = ["ExtractionInCoq"; "CicMeta"; "Signature"] in
  let cmk_cstr = lazy (get_construct mp "mk_cstr") in
  {
    interp =
      interp_desc [
        cmk_cstr, (fun _ -> function
          | [n; a; i] ->
            ml_string.interp n >>= fun n ->
            cterm1.interp a >>= fun a ->
            cterm2.interp i >>= fun i ->
            Some { c_name = n; c_arg = a; c_inst = i }
          | _ -> None);
      ];
    reify =
      begin fun c ->
        mkApp (
          mkConstruct (Lazy.force cmk_cstr),
          [|
            ml_string.reify c.c_name;
            cterm1.reify c.c_arg;
            cterm2.reify c.c_inst;
          |]
        )
      end;
    constr = lazy (mkInd (get_ind mp "constructor"));
  }

let constructor_list_option = option (list constructor)

let one_ind =
  let mp = ["ExtractionInCoq"; "CicMeta"; "Signature"] in
  let cmk_one_ind = lazy (get_construct mp "mk_one_ind") in
  {
    interp =
      interp_desc [
        cmk_one_ind, (fun _ -> function
          | [n; p; a; ck; k; c] ->
            ml_string.interp n >>= fun n ->
            cterm0.interp p >>= fun p ->
            cterm1.interp a >>= fun a ->
            sort.interp ck >>= fun ck ->
            sort.interp k >>= fun k ->
            constructor_list_option.interp c >>= fun c ->
            Some {
              i_name = n;
              i_param = p;
              i_arg = a;
              i_ckind = ck;
              i_kind = k;
              i_constrs = c;
            }
          | _ -> None);
      ];
    reify =
      begin fun i ->
        mkApp (
          mkConstruct (Lazy.force cmk_one_ind),
          [|
            ml_string.reify i.i_name;
            cterm0.reify i.i_param;
            cterm1.reify i.i_arg;
            sort.reify i.i_ckind;
            sort.reify i.i_kind;
            constructor_list_option.reify i.i_constrs;
          |]
        )
      end;
    constr = lazy (mkInd (get_ind mp "one_ind"));
  }

let cdecl1 = cdecl (S O)

let constant_obj =
  let mp = ["ExtractionInCoq"; "CicMeta"; "Signature"] in
  let cmk_cobj = lazy (get_construct mp "mk_cobj") in
  {
    interp =
      interp_desc [
        cmk_cobj, (fun _ -> function
          | [p; d] ->
            cterm0.interp p >>= fun p ->
            cdecl1.interp d >>= fun d ->
            Some {
              c_par = p;
              c_decl = d;
            }
          | _ -> None);
      ];
    reify =
      begin fun co ->
        mkApp (
          mkConstruct (Lazy.force cmk_cobj),
          [|
            cterm0.reify co.c_par;
            cdecl1.reify co.c_decl;
          |]
        )
      end;
    constr = lazy (mkInd (get_ind mp "constant_obj"));
  }

let ind_pack = list one_ind

let sig_item =
  let mp = ["ExtractionInCoq"; "CicMeta"; "Signature"] in
  let cMutInductive = lazy (get_construct mp "MutInductive") in
  let cConstant = lazy (get_construct mp "Constant") in
  {
    interp =
      interp_desc [
        cMutInductive, (fun _ -> function
          | [i] ->
            ind_pack.interp i >>= fun i ->
            Some (MutInductive i)
          | _ -> None);
        cConstant, (fun _ -> function
          | [co] ->
            constant_obj.interp co >>= fun co ->
            Some (Constant co)
          | _ -> None);
      ];
    reify =
      begin function
        | MutInductive i ->
          mkApp (
            mkConstruct (Lazy.force cMutInductive),
            [| ind_pack.reify i |]
          )
        | Constant co ->
          mkApp (
            mkConstruct (Lazy.force cConstant),
            [| constant_obj.reify co |]
          )
      end;
    constr = lazy (mkInd (get_ind mp "sig_item"));
  }
